/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */



#ifndef IUPACALPHABET_H_
#define IUPACALPHABET_H_

#include <stdlib.h>
#include <string.h>



char *get_encoding(char); // converts a symbol into 1,2,3,4... symbols that it represents
short int get_degeneracy(char); // returns the number of letters coded by a symbol
char *get_symbols(short int); // returns all symbols that have a specified degeneracy
char get_complement(char); // returns the complement of a symbol
char get_struct_complement(char); // returns the structural complement of a symbol => uppercase if lowercase and viceversa
int get_intersection(char,char); //returns the number of symbols in common
double get_relation_score(char, char); // returns the relationship score for encoding of the two symbols: 1 if identic, 1/degeneracy if one is encoded by the other, 0 otherwise
char *increase_degeneracy(char); // returns all possible new values for an increase in degeneracy (works for degen 1,2,3)
char *decrease_degeneracy(char); // returns all possible new values for a decrease in degeneracy (works for degen 2 and 4)

#endif /* IUPACALPHABET_H_ */
