/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "IUPACdecoder.h"


void duplicate(char ***oligos, int index, int times)
{
	if(times <= 0 || index <= 0)
		return;
	int i,j;
	for(i=0;i<times; i++)
		for(j=0;j<index;j++)
			strcpy((*oligos)[(i+1)*index+j], (*oligos)[j]);
}

void build_oligos(char *motif, int pos, char ***oligos, int index)
{
	if(pos == strlen(motif))
		return;
	int i,j;
	short int degen = get_degeneracy(motif[pos]);
	char *encoding = get_encoding(motif[pos]);
	duplicate(oligos, index, degen - 1);
	for(i=0;i<degen;i++)
		for(j=0;j<index;j++)
			(*oligos)[i*index + j][pos] = encoding[i];
	build_oligos(motif, pos+1, oligos, index*degen);
}


void init_decode_IUPAC(char *degen_oligo, char ***oligos, int **instantiations_counts, int *instantiations_length)
{
	int oligo_size = strlen(degen_oligo);
	int i;
	unsigned num_inst = 1;
	for(i=0;i<oligo_size;i++)
		num_inst *= get_degeneracy(degen_oligo[i]);

	MALLOC(*oligos, sizeof(char *)*num_inst);
	for(i=0;i<num_inst;i++)
	{
		MALLOC((*oligos)[i], (oligo_size + 1) * sizeof(char));
		(*oligos)[i][oligo_size] = '\0';
	}
	MALLOC(*instantiations_counts, sizeof(int)*num_inst);
	*instantiations_length = num_inst;
//	printf("\n OLIGO: %s \t\t instantiations_length: %i\n", degen_oligo, num_inst);
}	

void free_decode_IUPAC(char ***oligos, int num_instances, int **instantiations_counts)
{
	int i;
	for(i=0;i<num_instances;i++)
		FREE((*oligos)[i]);
	FREE(*oligos);
	FREE(*instantiations_counts);
}

// works with the full IUPAC alphabet
// num_inst is passed by reference so that on exit it is initialized with the correct value
void decode_IUPAC(char *degen_oligo, char ***instantiations)
{
	int pos = 0, index = 1;
	build_oligos(degen_oligo, pos, instantiations, index);
}


/* TESTING routines */

void test_build_oligos(char *degen_oligo)
{
	int i;
//	char *degen_oligo = "RASTN";
	int num_instances = 1;
	for(i=0;i<strlen(degen_oligo);i++)
		num_instances *= get_degeneracy((int)degen_oligo[i]);
	printf("degen_oligo: %s\nnum_instances: %i\n", degen_oligo, num_instances);

	char **oligos;
	MALLOC(oligos, sizeof(char *)*num_instances);
	for(i=0;i<num_instances;i++)
	{
		MALLOC(oligos[i], (strlen(degen_oligo) + 1) * sizeof(char));
		oligos[i][strlen(degen_oligo)] = '\0';
	}

	int pos = 0, index = 1;
	build_oligos(degen_oligo, pos, &oligos, index);


	for(i=0;i<num_instances;i++)
		printf("%s\n", oligos[i]);
	getchar();
}


void test_duplicate()
{
	int i;
	char **oligos;
	MALLOC(oligos, sizeof(char *)*10);
	for(i=0;i<10;i++)
	{
		MALLOC(oligos[i], 10*sizeof(char) + 1);
	}
	strcpy(oligos[0], "AGTC");
	strcpy(oligos[1], "GGGGG");
	strcpy(oligos[2], "CTCTCTC");
	duplicate(&oligos, 3, 2);
	for(i=0;i<9;i++)
	{
		printf("%s\n", oligos[i]);
	}
	printf("done\n");
	getchar();
}
