/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef PSSM_H_
#define PSSM_H_

#include <gsl/gsl_permutation.h>
#include <gsl/gsl_matrix.h>
#include "random_gen.h"

#define ZERO	1.0e-20

gsl_matrix *read_pssm(char *file);
gsl_matrix *read_mat(char *file);

void write_pssm(gsl_matrix *pssm, char *file);
void write_pssm_TRANSFAC_like_format_freq(gsl_matrix *mat, char *file, char *itype);
void read_pssm_TRANSFAC_like_format_counts(gsl_matrix *pssm_out, char *file, double pseudocount);
void read_pssm_TRANSFAC_like_format_freq(gsl_matrix *pssm_out, char *file);
int find_num_cols_TRASNFAC_pssm(char *pssm_TRANSFAC_file);


gsl_matrix *generate_PSSM(int length);
char *draw_from_pssm(gsl_matrix *pssm);

void pssm_to_IUPAC(gsl_matrix *pssm, char *motif);
void IUPAC_to_pssm(char *motif, gsl_matrix *pssm);

gsl_matrix *get_rev_compl_pssm(gsl_matrix *pssm);

double get_pos(gsl_matrix *pssm, int column, char letter);
void set_pos(gsl_matrix *pssm, int column, char letter, double value);

char draw_col_entry(gsl_vector_view vv);
void print_pssm(gsl_matrix *pssm);
double sum(double *arr, int num);

#endif /*PSSM_H_*/
