/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef ANALYSIS_H_
#define ANALYSIS_H_

#include <math.h>
#include "parameters.h"
#include "file.h"
#include "utilities.h"
#include "geneset_scoring.h"
#include "regression.h"

#include <gsl/gsl_permutation.h>
//#include <gsl/gsl_stats_double.h>
//#include <gsl/gsl_stats_int.h>
//#include <gsl/gsl_stats.h>

#define MALLOC(s,t) if(((s) = malloc(t)) == 0) { oops("error: malloc() ");}
#define FREE(p) {free((void *)p);p=0;}
#define oops(s) { perror((s)); exit(EXIT_FAILURE); }


typedef struct{int index;double score;} score_index;

typedef struct
{
	char **raw_input;
	int *index;double *scores;
	char **seq_ids;
	score_index *sort_index;
	int num;
}gene_level_scores;

void load_gene_level_scores(gene_level_scores *s, char *gene_level_scores_file);
void free_gene_level_scores(gene_level_scores *s);

typedef struct
{
	char *id;
	int index_location;
	double index_seq_level_stat;
}seq_details;

typedef struct
{
	char *inputfile;       // contains the name of the inputfile. FORMAT: gene_id<TAB>gene_level_score<newline>...
	char **seqfile_list;   // array of names of sequence files. First entry is the master file. The following entries are ortholgs.
	int num_seqfile_list;  // number of (char *) in array 'seqfile_list'
}dataset;

typedef struct
{
	dataset *d;
	int current_index;
	int num_datasets;
}dataset_file_names;


void run_analysis(params *parameters, IO_params *p_IO);

#endif /*ANALYSIS_H_*/
