/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef EIGSVD_H_
#define EIGSVD_H_

// allows for 'bool' type to be defined, 'true'=1,'fasle'=0
#include <stdbool.h>
#include <math.h>
#include <string.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_rng.h>

#include "random_gen.h"
#include "rmvn.h"

typedef enum {Golub_Reinsch, modified_Golub_Reinsch, jacobi} SVD_type;

typedef struct
{
// SVD: A = U S V^T
	bool estimate_U, estimate_S, estimate_V; int numeig;
}
eigs_SVD_Golub_Kahan_options;


typedef struct
{
// SVD: A = U S V^T
	gsl_matrix *U, *V; // left and right eigenvectors
	gsl_vector *S;     // singular values (in non-increasing order, consisent with U and V)
}
eigs_SVD;



//typedef enum {Golub_Reinsch, modified_Golub_Reinsch, jacobi} SVD_type;
// in           : dim2-by-m
// M            : m-by-m
// out = in * M : dim2-by-m
typedef void(*pt2svd)(gsl_matrix *in, eigs_SVD *out);


// SVD of A (M-by-N), works only if M >= N
//****************************************
/*
This function factorizes the M-by-N matrix A into the singular value 
decomposition A = U S V^T for M >= N. On output the matrix A is replaced 
by U. The diagonal elements of the singular value matrix S are stored in 
the vector S. The singular values are non-negative and form a non-increasing 
sequence from S_1 to S_N. The matrix V contains the elements of V in 
untransposed form. To form the product U S V^T it is necessary to take the 
transpose of V. A work vector of length N as well as additional N-by-N matrix 
is required for 'scratch' space.
*/

void SVD_Golub_Kahan(gsl_matrix *mat, eigs_SVD *eig, eigs_SVD_Golub_Kahan_options options);
void SVD_Golub_Reinsch(gsl_matrix *A, eigs_SVD *out);
void SVD_modified_Golub_Reinsch(gsl_matrix *A, eigs_SVD *out);
void SVD_Jacobi(gsl_matrix *A, eigs_SVD *out);

void QR_decomp(gsl_matrix *A, gsl_matrix *Q, gsl_matrix *R);

//eigvals_eigvecs eigenvec_and_eigenvals_SVD(gsl_matrix *corr_mat);
//void test_SVD();

#endif
