/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "file.h"
//#include "list.h"

FILE *open_file(char *filename, char * r_w)
{
	FILE *fp;
	if((fp = fopen(filename, r_w)) == NULL)
	{
	    	fprintf(stderr, "ERROR: cannot open %s\n", filename);
	    	exit(1);
	}
	else
		return fp;
} 

bool file_exist(char *filename)
{
	FILE *fp;
	if((fp = fopen(filename, "r")) != NULL)
	{
		fclose(fp);
		return true;
	}
	return false;
}

bool exists(char *filename)
{
    FILE *file_p = open_file(filename,"r");
    if(file_p != NULL)
    {
       fclose(file_p);
       return(true);
    }
    return(false);
}

int count_lines(char *file)
{
	FILE *fp = open_file(file, "r");
	int num_read = 0;
	char *lineptr = 0; int n = 0;
	while(getline(&lineptr, &n, fp) != -1)
	{
		num_read++;
		FREE(lineptr);
	}
	FREE(lineptr);
	fclose(fp);
	return num_read;
}

void free_file_lines(char ***lines, int num_lines)
{
	int i;
	for(i=0;i<num_lines;i++)
		FREE((*lines)[i]);
	FREE(*lines);
}

// read lines from a file discarding the newline character and storing the result into array 'char **lines'
void read_file_lines(char *in_file, char ***lines, int *num_lines)
{
	*num_lines = count_lines(in_file);
	MALLOC(*lines, sizeof(char *)*(*num_lines));
	char *lineptr = 0; int n = 0, index = 0;
	
	FILE *fp = open_file(in_file, "r");
	while(getline(&lineptr, &n, fp) != -1)
	{
		if(lineptr[strlen(lineptr) - 1] == '\n')
			lineptr[strlen(lineptr) - 1]= '\0';
		(*lines)[index++] = lineptr;
		lineptr = 0;
	}
	FREE(lineptr);
	fclose(fp);
}
/*
list_node *read_file(char *filename)
{
	int i;
	list *l = create_list();
//	list_node *curr, *start;
//	create_list(&curr);
//	start = curr;

	FILE *fp = open_file(filename, "r");
	char *lineptr = 0; int n = 0;
	while(getline(&lineptr, &n, fp) != -1)
	{
		if(lineptr[strlen(lineptr) - 1] == '\n')
			lineptr[strlen(lineptr) - 1]= '\0';
//		curr->value = lineptr;
		append_list_node(l, create_list_node(lineptr));
//		append_list_node(&curr, create_list_node(0));
		lineptr = 0;
	}
	fclose(fp);
//	return start;
	return l;
}
*/

// ???
void write_to_file(char *filename, char *string)
{
	FILE *f_p = fopen (filename, "w");
	fputs(string,f_p);
	fclose(f_p);
}
/*
char *create_tmp_filename(char *base_name, char *output_directory)
{
	char *tmp_file;
	MALLOC(tmp_file, sizeof(char) * (strlen(base_name) + strlen(output_directory) + 1));
	strcpy(tmp_file, output_directory);
	strcat(tmp_file, base_name);
	return tmp_file;
}

void test_create_tmp_file()
{
	char *base = "base";
	char *dir = "dir/";
	char *tmp = create_tmp_filename(base, dir);
	printf("%s", tmp);
	getchar();
}
*/

// returns name of the current directory
char *get_curr_dir()
{
	return getcwd(NULL,0);
}

//char *name = "/dir";
// creates a subdirectory to the current active directory
void create_dir(char *name)
{
//	char *curr_dir = get_curr_dir();
	char new_dir[200];
//	MALLOC(new_dir, sizeof(char) * (strlen(curr_dir) + strlen(name) + 1));
//	strcpy(new_dir, ".");
	strcpy(new_dir, name);
//	strcat(new_dir, name);
//	printf("new_dir***: %s\n", new_dir);
	int res;
	res = mkdir(new_dir, S_IRWXU);
	if(res == -1)
	{
		switch(errno)
		{
			case  EACCES: printf("Write permission is denied for the parent directory in which the new directory is to be added!\n");break;
			case  EMLINK: printf("The parent directory has too many links (entries)!\n");break;
			case  ENOSPC: printf("The file system doesn't have enough room to create the new directory!\n");break;
			case  EROFS:  printf("The parent directory of the directory being created is on a read-only file system and cannot be modified!\n");break;
	//		case  EEXIST: printf("directory exists: %s\n", new_dir);break;
		}
	}
}

void print_current_dir_contents()
{
	struct dirent **namelist;
    int n;
	n = scandir(".", &namelist, 0, alphasort);
    if(n < 0)
    	perror("scandir");
    else
    {
    	while(n--)
    	{
        	printf("%s\n", namelist[n]->d_name);
            free(namelist[n]);
        }
        free(namelist);
    }
    getchar();
}

// The following sample code will search the current directory for the entry name:
bool search_current_dir(char *name)
{
	struct dirent *dp;
	DIR *dirp = opendir(".");
	while (dirp)
	{
	    errno = 0;
	    if ((dp = readdir(dirp)) != NULL)
	    {
	        if (strcmp(dp->d_name, name) == 0)
	        {
	            closedir(dirp);
	            return true;
	        }
	    }
	    else
	    {
	        if (errno == 0)
	        {
	            closedir(dirp);
	            return false;
	        }
	        closedir(dirp);
	        return false;//READ_ERROR;
	    }
	}
	return false;//OPEN_ERROR;
}
