/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "geneset_scoring.h"

extern int zero_threshold_index;

void estimate_global_MEAN_and_VAR_and_MSS_threshold_exact(double *all_gene_level_stat, int num_all_genes, int geneset_size, int threshold_index, 
													double *global_mean, double *global_var, double *mss)
{	
	int i;
	double sum = 0.0, sum_squares = 0;
	for(i=0;i<num_all_genes;i++)
	{
		if(i > threshold_index)
			break;
		sum += all_gene_level_stat[i];
		sum_squares += all_gene_level_stat[i]*all_gene_level_stat[i];
	}
	*global_mean = sum / num_all_genes;
	*mss = sum_squares / num_all_genes;
	*global_var = ((*mss - (*global_mean)*(*global_mean)) * (num_all_genes - geneset_size))/(geneset_size*(num_all_genes - 1));
}

void estimate_global_VAR_exact(double *all_gene_level_stat, int num_all_genes, int geneset_size, double global_mean, double *global_var)
{	
	int i;
	double sum = 0.0, mean_sum_squared = 0;
	for(i=0;i<num_all_genes;i++)
		mean_sum_squared += all_gene_level_stat[i]*all_gene_level_stat[i];
	mean_sum_squared = mean_sum_squared / num_all_genes;
	double factor = (1.0 / geneset_size) + (1.0 / geneset_size - 1) / (num_all_genes - 1);
	*global_var = (mean_sum_squared - global_mean * global_mean) * factor;
}


void estimate_global_MEAN_and_VAR_exact(double *all_gene_level_stat, int num_all_genes, int geneset_size, double *global_mean, double *global_var)
{	
	int i;
	double sum = 0.0, mean_sum_squared = 0;
	for(i=0;i<num_all_genes;i++)
	{
		sum += all_gene_level_stat[i];
		mean_sum_squared += all_gene_level_stat[i]*all_gene_level_stat[i];
	}
	*global_mean = sum / num_all_genes;
	mean_sum_squared = mean_sum_squared / num_all_genes;
//	printf("((mean_sum_squared-(*global_mean)*(*global_mean)) = %lf\n", ((mean_sum_squared-(*global_mean)*(*global_mean))));
//	printf("num_all_genes = %i\n", num_all_genes);
//	printf("geneset_size = %i\n", geneset_size);
//	printf("num_all_genes - geneset_size = %i\n", num_all_genes - geneset_size);
//	double intermediate = ((double)geneset_size*((double)num_all_genes - 1));
	double factor = (1.0 / geneset_size) + (1.0 / geneset_size - 1) / (num_all_genes - 1);
	*global_var = (mean_sum_squared-(*global_mean)*(*global_mean))*factor;
//	printf("global_var1 = %lf\n", global_var1);

//	printf("(geneset_size*(num_all_genes - 1)) = %i\n", (geneset_size*(num_all_genes - 1)));
//	printf("intermediate = %lf\n", intermediate);
//	*global_var = ((mean_sum_squared-(*global_mean)*(*global_mean))*(num_all_genes - geneset_size)) / intermediate;//(geneset_size*(num_all_genes - 1));
}

void estimate_global_MEAN_and_VAR_exact_intercept_only_reg(double *all_gene_level_stat, int num_all_genes, int geneset_size, double *global_mean, double *global_var)
{	
	estimate_global_MEAN_and_VAR_exact(all_gene_level_stat, num_all_genes, geneset_size, global_mean, global_var);
//	printf("*global_var = %lf\n", *global_var);
	double factor = ((double)num_all_genes - geneset_size) / (num_all_genes - 1);
	*global_var = *global_var / factor;
//	printf("*global_var = %lf\n", *global_var);
//	getchar();
}

void estimate_global_MEAN_and_MEAN_SUM_SQUARES_threshold(double *all_gene_level_stat, int num_all_genes, int threshold_index, double *global_mean, double *mean_sum_squares)
{	
	int i;
	double sum_squares = 0.0, sum = 0.0;
	double val;
	for(i=0;i<num_all_genes;i++)
	{
		if(i > threshold_index)
			break;
		val = all_gene_level_stat[i] - all_gene_level_stat[threshold_index];
		sum += all_gene_level_stat[i];
		sum_squares += all_gene_level_stat[i]*all_gene_level_stat[i];
	}
	*global_mean      = sum / num_all_genes;
	*mean_sum_squares = sum_squares / num_all_genes;
}

void estimate_global_MEAN_and_VAR_threshold(double *all_gene_level_stat, unsigned long *gene_indexes, int num_all_genes, double threshold, double *global_mean, double *global_var)
{	
	int i;
	*global_mean = 0;
	double mean_sum_squared = 0;
	for(i=0;i<num_all_genes;i++)
	{
//		printf("%i: %i\n", i, gene_indexes[i]);getchar();
//		printf("%lf\n", all_gene_level_stat[gene_indexes[i]]);getchar();
		double val = 0;
		if(fabs(all_gene_level_stat[gene_indexes[i]]) <= fabs(threshold))
			break;
		val = all_gene_level_stat[gene_indexes[i]] - threshold;
		*global_mean += val;
		mean_sum_squared += val*val;
	}
	*global_mean /= num_all_genes;
	mean_sum_squared = mean_sum_squared / num_all_genes;
	*global_var = mean_sum_squared - (*global_mean)*(*global_mean);
//	*global_var = ((mean_sum_squared-(*global_mean)*(*global_mean))*(num_all_genes - geneset_size))/(geneset_size*(num_all_genes - 1));
}


void estimate_global_MEAN_and_VAR(double *all_gene_level_stat, unsigned long *gene_indexes, int num_all_genes, double *global_mean, double *global_var)
{	
	int i;
	*global_mean = 0;
	double mean_sum_squared = 0;
	for(i=0;i<num_all_genes;i++)
	{
//		printf("%i: %i\n", i, gene_indexes[i]);getchar();
//		printf("%lf\n", all_gene_level_stat[gene_indexes[i]]);getchar();
		double val = all_gene_level_stat[gene_indexes[i]];
		*global_mean += val;
		mean_sum_squared += val*val;
	}
	*global_mean /= num_all_genes;
	mean_sum_squared = mean_sum_squared / num_all_genes;
	*global_var = mean_sum_squared - (*global_mean)*(*global_mean);	
}

void estimate_global_MEDIAN_and_IQR(double *all_gene_level_stat, unsigned long *gene_indexes, int num_all_genes, double *global_median, double *global_IQR)
{
	int i;
	double scores[num_all_genes];
	for(i=0;i<num_all_genes;i++)
	    	scores[i] = all_gene_level_stat[gene_indexes[i]];
	size_t stride = 1;
	gsl_sort(scores, stride, num_all_genes);
	*global_median = gsl_stats_median_from_sorted_data(scores, stride, num_all_genes);
	double global_IQR_lower = gsl_stats_quantile_from_sorted_data(scores, stride, num_all_genes, 0.25);
	double global_IQR_upper = gsl_stats_quantile_from_sorted_data(scores, stride, num_all_genes, 0.75);
	*global_IQR      	= fabs(global_IQR_upper - global_IQR_lower);
}


/////////////////

void estimate_MEAN_harmonic(unsigned long *gene_indexes, int geneset_size, int num_all_genes, double *harminic_mean)
{	
	int i;
	double harmonic_sum = 0;
	for(i=0;i<geneset_size;i++)
//		*harminic_mean += 1.0 / all_gene_level_stat[gene_indexes[i]];
		harmonic_sum += (1.0 + gene_indexes[i]) / (1.0 * num_all_genes);
	*harminic_mean = num_all_genes / harmonic_sum;
}

double evaluate_random_set_ES_harmonic_mean(int *gene_indexes, int num_genes, double *all_gene_level_stat, int num_all_genes, double global_mean, double global_var)
{
	double harminic_mean;
	estimate_MEAN_harmonic(gene_indexes, num_genes, num_all_genes, &harminic_mean);
	return -harminic_mean;

	int i;
	double sum = 0.0;
	for(i=0;i<num_genes;i++)
	    	sum += all_gene_level_stat[gene_indexes[i]];
	double harmonic_mean = num_genes / sum;
	return harmonic_mean;
//	double Z_score = (mean - global_mean)/sqrt(global_var);
//	return Z_score;
}

/////////////////


double ks_ES(int *genes, int num_genes, int num_all_genes, int *leading_edge_index, int *leading_edge_num_genes)
{
		if(num_genes == 0)
			return 0;
		int i;
		int num_outside_geneset = num_all_genes - num_genes;
		double incr_hit = 1.0 / num_genes;
		double incr_miss = 1.0 / num_outside_geneset;
		double running_ES;
		double P_hit = 0, P_miss = 0, max_ES = 0, min_ES = 0;
		int last_index = 0;
		*leading_edge_num_genes = 0;
		*leading_edge_index = 0;
		for(i=0;i<num_all_genes;i++)
		{
//			printf("genes[%i]: %i\n", last_index, genes[last_index]);fflush(stdout);
			if(last_index < num_genes && genes[last_index] == i)
			{
				P_hit = P_hit + incr_hit;
//				printf("\n%i:P_hit %lf\n", i, P_hit);fflush(stdout);
				last_index++;
			}
			else
			{
				P_miss = P_miss + incr_miss;
//				printf("%i:P_miss %lf\t",i, P_miss);fflush(stdout);
			}
//			printf("running_ES: %lf\n", running_ES);fflush(stdout);
			running_ES = P_hit - P_miss;
			max_ES = (max_ES > running_ES ? max_ES : running_ES);
			min_ES = (min_ES < running_ES ? min_ES : running_ES);
			if(max_ES == running_ES || min_ES == running_ES)
			{
				*leading_edge_index = i;
				*leading_edge_num_genes = last_index;
			}
		}
		double ES = (fabs(max_ES) > fabs(min_ES) ? max_ES : min_ES);
		double normalizing_const = sqrt( (1.0 * num_genes * num_outside_geneset) / num_all_genes);
		return normalizing_const * ES;
}


double weighted_ks_ES(int *genes, int num_genes, double *seq_level_stat, int num_all_genes, double power, int *leading_edge_index, int *leading_edge_num_genes)
{
		if(num_genes == 0 || power == 0.0)
			return ks_ES(genes, num_genes, num_all_genes, leading_edge_index, leading_edge_num_genes);
		int i;
/*
		double N_r = 0;
		for(i=0;i<num_genes;i++)
//			N_r += pow(fabs(seq_level_stat[genes[i]]), power);
			N_r += pow(-seq_level_stat[genes[i]], power);
*/
		double N_geneset = 0, N_outside_geneset = 0;
		int ind = 0;
		for(i=0;i<num_all_genes;i++)
		{	
			if(ind < num_genes && genes[ind] == i)
			{
				N_geneset += pow(-seq_level_stat[i], power);
				ind++;
			}
//			else
//				N_outside_geneset += pow(-seq_level_stat[i], power);
		}
//		N_geneset = -N_geneset;
//		N_outside_geneset = -N_outside_geneset;

		double increment_miss = 1.0/(num_all_genes - num_genes);
//		double increment_miss = 1.0 / N_outside_geneset;

		double ES, running_ES;
		double  max_ES = 0, min_ES = 0, P_hit = 0, P_miss = 0;
		ind = 0;
		for(i=0;i<num_all_genes;i++)
		{
//			printf("genes[%i]: %i\n", last_index, genes[last_index]);fflush(stdout);
			if(ind < num_genes && genes[ind] == i)
			{
//				P_hit = P_hit + pow(fabs(seq_level_stat[i]), power)/N_r;
				P_hit = P_hit + pow(-seq_level_stat[i], power) / N_geneset;
				ind++;
//				printf("%i:P_hit %lf\n", i, P_hit);fflush(stdout);
			}
			else
			{
				P_miss = P_miss + increment_miss;
//				P_miss = P_miss + pow(-seq_level_stat[i], power) / N_outside_geneset;
//				printf("%i:P_miss %lf\t",i, P_miss);fflush(stdout);
			}
			running_ES = P_hit - P_miss;
/*
			if(i % 100 == 0)
			{
				printf("%i: P_hit = %lf\tP_miss = %lf\trunning_ES: %lf\n", i, P_hit, P_miss, running_ES);fflush(stdout);getchar();
			}
*/
			max_ES = MAX(max_ES, running_ES);
			min_ES = MIN(min_ES, running_ES);
			if(running_ES == max_ES || running_ES == min_ES)
				*leading_edge_index = i;
		}

		ES = (max_ES > fabs(min_ES)) ? max_ES : min_ES;
		double normalizing_const = sqrt( (1.0 * num_genes * (num_all_genes - num_genes)) / num_all_genes);
//		double normalizing_const = sqrt( (1.0 * N_geneset * N_outside_geneset) / (N_geneset + N_outside_geneset));
		double normalized_ES = normalizing_const * ES;
//		printf("normalizing_const: %lf\nES = %lf\nnormalized_ES = %lf\n", normalizing_const, ES, normalized_ES);fflush(stdout);
//		getchar();
		return normalized_ES;
}


double evaluate_set_ES_median(int *gene_indexes, int num_genes, double *all_gene_level_stat, int num_all_genes, double median_global_score, double IQR_global)
{
	int i;
	double scores_local[num_genes];
	for(i=0;i<num_genes;i++)
	    	scores_local[i] = all_gene_level_stat[gene_indexes[i]];
	size_t stride = 1;
//	gsl_sort(scores_local, stride, num_genes);
	double median_geneset_score = gsl_stats_median_from_sorted_data(scores_local, stride, num_genes);
//	return (median_geneset_score - median_global_score) / IQR_global;
	return median_geneset_score;

	double geneset_IQR_lower = gsl_stats_quantile_from_sorted_data(scores_local, stride, num_genes, 0.25);
	double geneset_IQR_upper = gsl_stats_quantile_from_sorted_data(scores_local, stride, num_genes, 0.75);
	double geneset_IQR       = fabs(geneset_IQR_upper - geneset_IQR_lower);
	double normalized_median_score = median_geneset_score / geneset_IQR;

//	double median_global_score = gsl_stats_median_from_sorted_data(all_gene_level_stat, stride, num_all_genes);

//	printf("geneset_IQR_lower = %lf\n", geneset_IQR_lower);
//	printf("geneset_IQR_upper = %lf\n", geneset_IQR_upper);
//	printf("median_global_score = %lf\n", median_global_score);getchar();

//	double normalized_median_score = (median_geneset_score - median_global_score) / IPR_global;
	return normalized_median_score;
}

double evaluate_random_set_ES_abs_dev(int *gene_indexes, int num_genes, double *all_gene_level_stat, int num_all_genes, double global_mean, double global_var)
{
	int i;
	double sum = 0.0;
	double data[num_genes];
	for(i=0;i<num_genes;i++)
	{
	    	sum += all_gene_level_stat[gene_indexes[i]];
		data[i] = all_gene_level_stat[gene_indexes[i]];
	}
	double mean = sum / num_genes;

	// global_var <- absdev
	double Z_score = (mean - global_mean)/global_var;
	return Z_score;
/*
	size_t stride = 1;
	double absdev = gsl_stats_absdev_m(data, stride, num_genes, mean);
	return mean / absdev;
*/
}

// here we don't need sorted order for 'all_gene_level_stat'

double evaluate_random_set_ES(int *gene_indexes, int num_genes, double *all_gene_level_stat, int num_all_genes, double global_mean, double global_var)
{
	int i;
	double sum = 0.0;
	for(i=0;i<num_genes;i++)
	{
	    	sum += all_gene_level_stat[gene_indexes[i]];
	}
	double mean = sum / num_genes;
	double Z_score = (mean - global_mean) / sqrt(global_var);
//	printf("mean %8.4f, global mean %8.4f, global var %12.9f, z-score %8.4f\n", mean, global_mean, global_var, Z_score);
	return Z_score;
}


double regression_score(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes);//, geneset_set_scoring_params *scoring_params)

double random_set_ES_bootstrap_Std(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, 
				double observed_score, double bootstrap_Std_threshold, int num_draws, bool is_regression_initialized, double global_location, double global_scale)
{
	int i;
//	int num_draws = 50;
	double draws[num_draws];
	unsigned long num_dest = num_genes;
	unsigned long dest[num_dest];
	double sum = 0.0, sum_squares = 0.0;
	for(i=0;i<num_draws;i++)
	{
		unsigned long *src = genes;
		unsigned long num_src = num_genes;
		sample_with_replacement(dest, num_dest, src, num_src);
		if(is_regression_initialized == true)
			draws[i] = -regression_score(dest, num_dest, all_gene_level_stat, num_all_genes);//, scoring_params);
		else
			draws[i] = -evaluate_random_set_ES(dest, num_dest, all_gene_level_stat, num_all_genes, global_location, global_scale);
	
//		printf("global_location = %5.3lf, global_scale = %5.2lf, draws[%i]:% 5.2lf\n", global_location, global_scale, i, draws[i]);
		sum += draws[i];
		sum_squares += draws[i] * draws[i];
	}
	double bootstrap_ave = sum / num_draws;
	double bootstrap_se = sqrt(sum_squares / num_draws - (bootstrap_ave * bootstrap_ave));
//	double observed_score = evaluate_random_set_ES(genes, num_genes, all_gene_level_stat, num_all_genes, global_location, global_scale);
//	printf("\nbootstrap: %lf [%lf, %lf]", bootstrap_ave, bootstrap_ave - bootstrap_Std_threshold*bootstrap_se, 
//						bootstrap_ave + bootstrap_Std_threshold*bootstrap_se);
	return (bootstrap_ave - bootstrap_Std_threshold*bootstrap_se);
//	return observed_score / bootstrap_se;
}

double evaluate_random_set_ES_threshold(int *gene_indexes, int num_genes, double *all_gene_level_stat, int num_all_genes, double global_mean, double global_var, int threshold_index)
{
	int i;
	double sum = 0.0;
	for(i=0;i<num_genes;i++)
	{
		if(gene_indexes[i] <= threshold_index)
		    	sum += (all_gene_level_stat[gene_indexes[i]] - all_gene_level_stat[threshold_index]);
	}
	double mean = sum / num_genes;
//	printf("geneset mean (%i) = %-20.8lfthreshold_index = %-10i\n", num_genes, mean, threshold_index);fflush(stdout);
	double Z_score = (mean - global_mean)/sqrt(global_var);
	return Z_score;
}

double evaluate_random_set_ES_bootstrap_all_genes
(int *geneset_indexes, int num_geneset_indexes, double *all_gene_level_stat, int *all_genes_bootstrap_indexes, int *bootstrap_indexes_counts, int num_all_genes, double global_mean, double global_var)
{
	int i, j;
	double sum = 0.0;
//	int count = 0;
	for(i=0;i<num_geneset_indexes;i++)
	{
		sum += all_gene_level_stat[geneset_indexes[i]] * bootstrap_indexes_counts[geneset_indexes[i]];
		bootstrap_indexes_counts[geneset_indexes[i]] = 0;
//		count += bootstrap_indexes_counts[all_genes_bootstrap_indexes[i]];
	}
//	double mean = sum / count;
	double mean = sum / num_geneset_indexes;

//	unsigned long m = count;
	unsigned long m = num_geneset_indexes;

	unsigned long G = num_all_genes;
	double Z_score = (mean - global_mean)/sqrt(global_var); 
//	printf("Z_score = %10.3lf\n", Z_score);fflush(stdout);
	return Z_score;
}


// here we don't need sorted order for 'all_gene_level_stat'
double random_set_ES_bootstrap_all_genes(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, int *indexes, double global_mean, double global_var)
{
	int i, j;
	int num_draws = 300;
	double draws[num_draws];
	unsigned long bootstrap_indexes[zero_threshold_index];
	int bootstrap_indexes_counts[num_all_genes];
	for(i=0;i<num_all_genes;i++)
		bootstrap_indexes_counts[i] = 0;
	double sum = 0.0;
	for(i=0;i<num_draws;i++)
	{
//		sample_with_replacement(bootstrap_indexes, num_all_genes, indexes, num_all_genes);
		sample_with_replacement(bootstrap_indexes, zero_threshold_index, indexes, zero_threshold_index);

		for(j=0;j<zero_threshold_index;j++)
			bootstrap_indexes_counts[bootstrap_indexes[j]]++;
//		double global_mean, global_var;
//		estimate_global_MEAN_and_VAR(double *all_gene_level_stat, unsigned long*gene_indexes, int num_all_genes, double *global_mean, double *global_var)
//		estimate_global_MEAN_and_VAR(all_gene_level_stat, bootstrap_indexes, num_all_genes, &global_mean, &global_var);
//		printf("done estimate_global_MEAN_and_VAR\n");getchar();
		draws[i] = evaluate_random_set_ES_bootstrap_all_genes(genes, num_genes, all_gene_level_stat, bootstrap_indexes, bootstrap_indexes_counts, num_all_genes, global_mean, global_var);
//		sum += draws[i];
//		printf("%10.2lf ", draws[i]);fflush(stdout);getchar();
	}
//	double ave = sum / num_draws;
//	return (ave - global_mean) / sqrt(global_var);
	
	size_t stride = 1;	
	gsl_sort(draws, stride, num_draws);
//	double median_Z_score = gsl_stats_median_from_sorted_data(draws, stride, num_draws);
//	return median_Z_score;

	double observed_score = evaluate_random_set_ES(genes, num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);
	double score_25percentile_permuted = gsl_stats_quantile_from_sorted_data(draws, stride, num_draws, 0.25);
	double score_75percentile_permuted = gsl_stats_quantile_from_sorted_data(draws, stride, num_draws, 0.75);
//	double norm_bootstrap = fabs(score_75percentile_permuted - score_25percentile_permuted);
	double norm_bootstrap = fabs((2*observed_score - score_25percentile_permuted) - (2*observed_score - score_75percentile_permuted));
	double normalized_median_median_Z_score = observed_score / norm_bootstrap;
	return normalized_median_median_Z_score;
}

// here we don't need sorted order for 'all_gene_level_stat'
double random_set_ES_stable_threshold(d_gs *dg, double *all_gene_level_stat, int num_all_genes, double global_location, double global_scale, int index_threshold)
{
	double median_global_score = global_location;
	double IPR_global = global_scale;

//	double observed_score = evaluate_random_set_ES_threshold(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, global_location, global_scale, index_threshold);
//	return observed_score;

	int i;
	int num_draws = 300;
	double draws[num_draws];
	unsigned long num_dest = dg->num_genes/2;
	unsigned long dest[num_dest];

	for(i=0;i<num_draws;i++)
	{
		unsigned long *src = dg->genes;
		unsigned long num_src = dg->num_genes;
		sample_without_replacement(dest, num_dest, src, num_src);
//		sample_with_replacement(dest, num_dest, src, num_src);

//		draws[i] = evaluate_set_ES_median(dest, num_dest, all_gene_level_stat, num_all_genes, median_global_score, IPR_global);
		draws[i] = evaluate_random_set_ES_threshold(dest, num_dest, all_gene_level_stat, num_all_genes, global_location, global_scale, index_threshold);
	}

	size_t stride = 1;	
	gsl_sort(draws, stride, num_draws);
	double median_Z_score  = gsl_stats_median_from_sorted_data(draws, stride, num_draws);
	double score_lower_percentile = gsl_stats_quantile_from_sorted_data(draws, stride, num_draws, 0.25);
	double score_upper_percentile = gsl_stats_quantile_from_sorted_data(draws, stride, num_draws, 0.75);

	double norm = fabs(score_upper_percentile - score_lower_percentile);
	double normalized_median_Z_score = median_Z_score / norm;
//	dg->score_5percentile  /= norm;
//	dg->score_95percentile /= norm;

//	double normalized_median_Z_score = median_Z_score / norm;
//	double normalized_median_Z_score = median_Z_score - median_global_score / global_scale;

//	dg->score_5percentile  /= global_scale;
//	dg->score_95percentile /= global_scale;
//	printf("median_Z_score=%10.2lf\n", median_Z_score);//getchar();

//	printf("[%10.2lf%10.2lf]%10.2lf%10.2lf\n", dg->score_5percentile / norm, dg->score_95percentile / norm, median_Z_score, normalized_median_Z_score);fflush(stdout);getchar();
	return normalized_median_Z_score;


//	return median_Z_score;
//	return dg->score_5percentile;
//	return dg->score_95percentile;

/*
	double mad = 0;
	for(i=0;i<num_draws;i++)
		mad += fabs(median_Z_score - draws[i]);
	mad /= num_draws;
//	printf("%10.3lf%10.3lf\n", median_Z_score, mad);fflush(stdout);getchar();
	return median_Z_score / mad;
*/
//	printf("%10.3lf%10.3lf\n", observed_score, median_Z_score);fflush(stdout);
//	printf("%10.3lf%10.3lf%10.3lf%10.3lf", observed_score, median_Z_score, median_Z_score1, median_Z_score2);fflush(stdout);
//	getchar();
}


////////////
/* STABLE */
////////////

// here we don't need sorted order for 'all_gene_level_stat'
double random_set_ES_stable_bootstrap(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, double global_location, double global_scale)
{
//	double median_global_score = global_location;
//	double IPR_global = global_scale;

//	double observed_score = evaluate_random_set_ES(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);
//	double score_threshold = 20;
	int i;
	int num_draws = 200;
//	if(-observed_score > score_threshold)
//		num_draws = 200;
//	else
//		return observed_score;

	double draws[num_draws];
	unsigned long num_dest = num_genes;//MIN(num_genes / 2, 1000);
	unsigned long dest[num_dest];

	double sum = 0.0, sum_squares = 0.0;
	for(i=0;i<num_draws;i++)
	{
		unsigned long *src = genes;
		unsigned long num_src = num_genes;
//		sample_without_replacement(dest, num_dest, src, num_src);
		sample_with_replacement(dest, num_dest, src, num_src);

//		draws[i] = evaluate_set_ES_median(dest, num_dest, all_gene_level_stat, num_all_genes, median_global_score, IPR_global);
		draws[i] = evaluate_random_set_ES(dest, num_dest, all_gene_level_stat, num_all_genes, global_location, global_scale);

//		int leading_edge_index, leading_edge_num_genes;
//		double ks_score = ks_ES(dest, num_dest, num_all_genes, &leading_edge_index, &leading_edge_num_genes);
//		draws[i] = -ks_score;

//		sum += draws[i];
//		sum_squares += draws[i] * draws[i];
	}
//	double ave = sum / num_draws;
//	return ave;

//	return ave / sqrt( (sum_squares/num_draws - ave*ave) / num_draws );
//	return ave / sqrt( (sum_squares/num_draws - ave*ave) );

	size_t stride = 1;	
	gsl_sort(draws, stride, num_draws);
//	return draws[0];

//	double median_Z_score  = gsl_stats_median_from_sorted_data(draws, stride, num_draws);
//	return median_Z_score;

	double score_lower_percentile = gsl_stats_quantile_from_sorted_data(draws, stride, num_draws, 0.25);
	double score_upper_percentile = gsl_stats_quantile_from_sorted_data(draws, stride, num_draws, 0.75);

	double observed_score = evaluate_random_set_ES(genes, num_genes, all_gene_level_stat, num_all_genes, global_location, global_scale);
	double norm_bootstrap = fabs((2*observed_score - score_lower_percentile) - (2*observed_score - score_upper_percentile));
	double normalized_score = observed_score / norm_bootstrap;
	return normalized_score;
}

// here we don't need sorted order for 'all_gene_level_stat'
double random_set_ES_stable(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, double global_location, double global_scale)
{
//	return random_set_ES_stable_bootstrap(genes, num_genes, all_gene_level_stat, num_all_genes, global_location, global_scale);

//	double median_global_score = global_location;
//	double IPR_global = global_scale;

//	double score_threshold = 20;
	int i;
	int num_draws = 300;
//	if(-observed_score > score_threshold)
//		num_draws = 200;
//	else
//		return observed_score;

	double draws[num_draws];
	unsigned long num_dest = num_genes / 2;//MIN(num_genes / 2, 1000);
	unsigned long dest[num_dest];

	double sum = 0.0, sum_squares = 0.0;
	for(i=0;i<num_draws;i++)
	{
		unsigned long *src = genes;
		unsigned long num_src = num_genes;
		sample_without_replacement(dest, num_dest, src, num_src);
//		sample_with_replacement(dest, num_dest, src, num_src);

//		draws[i] = evaluate_set_ES_median(dest, num_dest, all_gene_level_stat, num_all_genes, median_global_score, IPR_global);
		draws[i] = evaluate_random_set_ES(dest, num_dest, all_gene_level_stat, num_all_genes, global_location, global_scale);

//		int leading_edge_index, leading_edge_num_genes;
//		double ks_score = ks_ES(dest, num_dest, num_all_genes, &leading_edge_index, &leading_edge_num_genes);
//		draws[i] = -ks_score;

//		sum += draws[i];
//		sum_squares += draws[i] * draws[i];
	}
//	double ave = sum / num_draws;
//	return ave;

//	return ave / sqrt( (sum_squares/num_draws - ave*ave) / num_draws );
//	return ave / sqrt( (sum_squares/num_draws - ave*ave) );

	size_t stride = 1;	
	gsl_sort(draws, stride, num_draws);

/*
	double median_Z_score  = gsl_stats_median_from_sorted_data(draws, stride, num_draws);
//	printf("[%10.2lf%10.2lf]\n", observed_score, median_Z_score);fflush(stdout);getchar();
	return median_Z_score;
*/

	double observed_score = evaluate_random_set_ES(genes, num_genes, all_gene_level_stat, num_all_genes, global_location, global_scale);
	double score_lower_percentile = gsl_stats_quantile_from_sorted_data(draws, stride, num_draws, 0.25);
	double score_upper_percentile = gsl_stats_quantile_from_sorted_data(draws, stride, num_draws, 0.75);

//	double norm = fabs(score_upper_percentile - score_lower_percentile);
	double norm = fabs((2*observed_score - score_lower_percentile) - (2*observed_score - score_upper_percentile));

/*
	double normalized_median_Z_score = median_Z_score / norm;
	return normalized_median_Z_score;
*/

	double normalized_observed_Z_score = observed_score / norm;
	return normalized_observed_Z_score;

//	double norm = ((dg->score_95percentile - dg->score_5percentile));
//	double normalized_median_Z_score = median_Z_score / norm;
//	dg->score_5percentile  /= norm;
//	dg->score_95percentile /= norm;

//	double normalized_median_Z_score = (median_Z_score - median_global_score) / global_scale;
//	double normalized_median_Z_score = median_Z_score - median_global_score / global_scale;

//	dg->score_5percentile  /= global_scale;
//	dg->score_95percentile /= global_scale;

//	printf("[%10.2lf%10.2lf]%10.2lf%10.2lf\n", dg->score_5percentile / norm, dg->score_95percentile / norm, median_Z_score, normalized_median_Z_score);fflush(stdout);getchar();
//	return normalized_median_Z_score;


//	return median_Z_score;
//	return dg->score_5percentile;
//	return dg->score_95percentile;

/*
	double mad = 0;
	for(i=0;i<num_draws;i++)
		mad += fabs(median_Z_score - draws[i]);
	mad /= num_draws;
//	printf("%10.3lf%10.3lf\n", median_Z_score, mad);fflush(stdout);getchar();
	return median_Z_score / mad;
*/
//	printf("%10.3lf%10.3lf\n", observed_score, median_Z_score);fflush(stdout);
//	printf("%10.3lf%10.3lf%10.3lf%10.3lf", observed_score, median_Z_score, median_Z_score1, median_Z_score2);fflush(stdout);
//	getchar();
}

////////////
/* STABLE */
////////////


void test_gaussian_pdf()
{
	int i;
	double x[10] = {1,2,3,4,5,6,7,8,9,10};
	double expected_mean_geneset_size = 5.5;
	double sigma = 2;
	for(i=0;i<10;i++)
	{

//		double a = 2, b = 5;
		double a = 1, b = 1;
		double weight0 = gsl_ran_beta_pdf(x[i] / 10, a, b);
		double weight  = gsl_ran_gaussian_pdf(x[i]-expected_mean_geneset_size, sigma);
		double weight1 = gsl_expm1(-(x[i]-expected_mean_geneset_size)*(x[i]-expected_mean_geneset_size) / (sigma*sigma)) + 1;
		printf("\nweight0 = %-10.5lfweight = %-10.5lf\tweight1 = %-10.5lf", weight0, weight, weight1);
//		printf("\nx = %-10.2lfweight = %-10.5lf", x, weight);
	}
	printf("\n");
	exit(0);
}

///////////////////////
/* Adaptive to Noise */
///////////////////////


double random_set_ES_adaptive_to_noise_level(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params)
{
	double *global_mean_sum_squares = scoring_params->mean_sum_squares_estimates;
//	test_gaussian_pdf();
	int i, j;
	int num_partitions = 500;
	double draws[num_partitions];
	double partition_threshold_fraction[num_partitions];
//	double a = 15, b = 30;
//	double a = 1.0, b = 1.0;
//	double a = .5, b = .5;
	double a = 15, b = 1;


//	double mean = 0.3, var = 0.1;
//	mean = 0.5
//	var = 12
//	a = mean * ( (mean - mean*mean) / var - 1 )
//	b = (1 - mean)*(a - 1)
//	x = rbeta(2000,a,b)
//	hist(x, nclass=n.bins(x))

	ran_beta(partition_threshold_fraction, num_partitions, a, b);
	double copy_all_gene_level_stat[num_all_genes];

	int gene_indexes[num_all_genes];
	for(i=0;i<num_all_genes;i++)
		gene_indexes[i] = i;

//	printf("%s\n", dg->degen_oligo);fflush(stdout);

	double global_mean, global_var;
/*
	double expected_mean_geneset_size = (0.4*num_all_genes);
	double expected_sd_geneset_size = (0.2*num_all_genes);

//	double ks_ES_arr[num_all_genes];
	int cumsum[num_all_genes];
	cumsum[0] = 0;
	int count = 0, ind = 0;
	for(i=0;i<num_all_genes;i++)
	{
		if(i == genes[ind])
		{
			count++;
			ind++;
		}
		cumsum[i] = count;
//		ks_ES_arr[i] = ks_ES(dg->genes, cumsum[i], num_all_genes);
	}
*/

	double weight = 1.0;//gsl_ran_beta_pdf((1.0 * dg->num_genes) / num_all_genes, 2, 5);

	double sum = 0.0, sum_squares = 0.0;
	int partition_threshold;
	int part_threshold = (int)(0.01*num_all_genes);
	for(i=0;i<num_partitions;i++)
	{
//		partition_threshold = (int)(partition_threshold_fraction[i] * num_all_genes);
		partition_threshold = (int)(partition_threshold_fraction[i] * zero_threshold_index);

		if(partition_threshold < part_threshold)
			partition_threshold = part_threshold;
		if(partition_threshold > num_all_genes - part_threshold)
			partition_threshold = num_all_genes - part_threshold;

//		printf("\n%i: partition_threshold = %i\n", i+1, partition_threshold);fflush(stdout);
		global_mean = scoring_params->location_estimates[partition_threshold];
		global_var = ( (scoring_params->mean_sum_squares_estimates[partition_threshold] - global_mean * global_mean) * (num_all_genes - num_genes) ) / ( num_genes * (num_all_genes - 1) );
/*
		double x = dg->num_genes - expected_mean_geneset_size;
		double bandwidth = expected_sd_geneset_size;
//		double weight = gsl_ran_gaussian_pdf(x, sigma);
		double weight = gsl_expm1(-x*x / bandwidth) + 1;
*/
//		int leading_edge_index, leading_edge_num_genes;
//		double ks_score = ks_ES(dg->genes, cumsum[partition_threshold], num_all_genes, &leading_edge_index, &leading_edge_num_genes);
//		double weight = ks_score;
		draws[i] =  evaluate_random_set_ES_threshold(genes, num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var, partition_threshold);
/*
		int leading_edge_index, leading_edge_num_genes;
		double ks_score = ks_ES(dg->genes, cumsum[partition_threshold], num_all_genes, &leading_edge_index, &leading_edge_num_genes);
		draws[i] = -ks_score;
*/
//		double random_set = evaluate_random_set_ES_threshold(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var, partition_threshold);
//		draws[i] = 0.5*leading_edge_score + 0.5*random_set;

		sum += draws[i];
//		sum_squares += draws[i] * draws[i];
	}
	double ave_score = sum / num_partitions;
//	printf("ave_score = %10.2lf\n", ave_score);
	return ave_score;

//	if(ave_score < 0)
//		return -log( MAX(-ave_score, 10e-10) );
//	else
//		return 0;

	size_t stride = 1;	
	gsl_sort(draws, stride, num_partitions);
/*
	double median_Z_score = gsl_stats_median_from_sorted_data(draws, stride, num_partitions);
	return median_Z_score;
*/
/*
	if(median_Z_score < 0)
		return -log( MAX(-median_Z_score, 10e-10) );
	else
		return 0;
*/
/*
	double score_lower_percentile = gsl_stats_quantile_from_sorted_data(draws, stride, num_partitions, 0.25);
	double score_upper_percentile = gsl_stats_quantile_from_sorted_data(draws, stride, num_partitions, 0.75);
	double norm = fabs(score_upper_percentile - score_lower_percentile) + 10e-10;
	if(median_Z_score < 0)
		return -log( MAX(-median_Z_score / norm, 10e-10) );
	else
		return 0;
*/

//	return draws[0] / norm;
	return -log( MAX(-draws[0], 10e-10) );


/*
	if(-ave_score >= 30.0)// || (ave_score > 10.0 && ave_score < 12.0))
	{
		double weights[num_all_genes];
		int last_occ_index = - 1;
		for(i=0;i<num_all_genes;i++)
		{
			double a = 1 + (double)cumsum[i];
			double b = 1 + (double)(dg->num_genes - cumsum[i]);
//			weights[i] = gsl_ran_beta_pdf((1.0 * i) / num_all_genes, a, b);
			weights[i] = gsl_ran_beta_pdf((double)cumsum[i] / dg->num_genes, a, b);
			if(cumsum[i] == cumsum[num_all_genes - 1] && last_occ_index == -1)
				last_occ_index = i;		
		}
	
		double p_mean = ((double)dg->num_genes) / num_all_genes;
		double p_mean1 = ((double)dg->num_genes) / last_occ_index;
		double binom_pval2;

		for(i=0;i<num_all_genes;i++)
		{
			if(i % 100 == 0)// && i <= last_occ_index )
			{
				double shrinkage = 10e-10;
//				double p_hat = (double)cumsum[i] / i + shrinkage;
				double p_hat = (double)cumsum[i] / dg->num_genes + shrinkage;
				double ci = (p_hat - p_mean) / ( sqrt( (p_hat*(1-p_hat)) / i) );
				double ci1 = (p_hat - p_mean1) / ( sqrt( (p_hat*(1-p_hat)) / i) );

				double binom_pval1 = gsl_cdf_binomial_P(cumsum[i], p_mean, dg->num_genes);
				if(i <= last_occ_index )
					binom_pval2 = gsl_cdf_binomial_P(cumsum[i], p_mean1, dg->num_genes);
//				double ci = p_hat / ( sqrt( (p_hat*(1-p_hat)) / dg->num_genes) );
				printf("%i: cumsum[i] = %i, ks_ES_arr = %-10.4lf, ci1 = %-10.4lf, binom_pval2 = %-20.18lf, weigth = %-10.8lf\n", i, cumsum[i], ks_ES_arr, ci1, binom_pval2, weights[i]);
			}
		}
		printf("p_mean = %20.15lf, num_all_genes = %i, geneset_size = %i, score = %-15.8lf\n", p_mean, num_all_genes, dg->num_genes, ave_score);
		getchar();
	}
*/
//	double rank_weight = ks_ES(dg->genes, cumsum[num_all_genes - 1], last_occ_index);//gsl_ran_beta_pdf(x, a, b);
//	double rank_weight = ks_ES(dg->genes, cumsum[num_all_genes - 1], num_all_genes);//gsl_ran_beta_pdf(x, a, b);
//	if(rank_weight < 0)
//		rank_weight = 0;
//	return ave_score * rank_weight;
}


///////////////////////
/* Adaptive to Noise */
///////////////////////



void get_smoothed_evidence(double *all_gene_level_stat, int num_all_genes, gsl_matrix *kNN, double *smoothed_values_out)
{
	int i, j, ii;

//	for(i=0;i<num_all_genes;i++)
//		smoothed_values_out[i] = all_gene_level_stat[i];
//	return;

	int num_draws = 100;
	double lambdas[num_draws];
	double a = 10, b = 1;
	ran_beta(lambdas, num_draws, a, b);

/*
	lambdas[0] = 0;
	for(j=1;j<num_draws;j++)
		lambdas[j] = lambdas[j-1] + 1.0 / (num_draws - 1);
*/
	for(i=0;i<num_all_genes;i++)
		smoothed_values_out[i] = 0;

	for(j=0;j<num_draws;j++)
	{
		double lambda = 1;//lambdas[j];
		for(i=0;i<num_all_genes;i++)
		{
			double shrinkage_target = 0;
			int num_neighbours = 0;
	//		printf("%i-by-%i\n", kNN->size1, kNN->size2);

			double sum = 0.0;
			for(ii=1;ii<kNN->size2;ii++)
			{
				int ind = (int)gsl_matrix_get(kNN, i, ii);
//				printf("ind = %i\n", ind);
//				if(ind > 0)
//				{
					sum += all_gene_level_stat[ind];// / (ii + 1);
					num_neighbours++;
//					printf("%i: %lf\n", ind, all_gene_level_stat[ind]);
//				}
			}
			shrinkage_target = sum / num_neighbours;
			smoothed_values_out[i] += (lambda*all_gene_level_stat[i] + (1-lambda)*shrinkage_target) / num_draws;
//			printf("all_gene_level_stat = %lf\n", all_gene_level_stat[i]); 
//			printf("shrinkage_target = %lf\n", shrinkage_target);
		}
		printf("smoothed_values_out[i] = %lf\n", smoothed_values_out[i]); getchar();
	}
}


void preprocess_thresholds_using_kNN(double *all_gene_level_stat, int num_all_genes, gsl_matrix *kNN, geneset_set_scoring_params *scoring_params)
{
	int i, ii, j;
	int gene_indexes[num_all_genes];
	double copy_all_gene_level_stat[num_all_genes];
	for(i=0;i<num_all_genes;i++)
		gene_indexes[i] = i;

//	printf("before\n");
//	for(i=0;i<20;i++)
//		printf("%lf\n", all_gene_level_stat[i]);

	double smoothed_values[num_all_genes];
	get_smoothed_evidence(all_gene_level_stat, num_all_genes, kNN, smoothed_values);

/*
	double lambda = 0.0;
	double smoothed_values[num_all_genes];
	for(i=0;i<num_all_genes;i++)
	{
		double shrinkage_target = 0;
//		printf("%i-by-%i\n", kNN->size1, kNN->size2);
		for(ii=0;ii<kNN->size2;ii++)
		{
			int ind = (int)gsl_matrix_get(kNN, i, ii);
			shrinkage_target += all_gene_level_stat[ind] / (ii + 1);
//			printf("%i: %lf\n", ind, all_gene_level_stat[ind]);
		}
//		printf("shrinkage_target = %lf\n", shrinkage_target);
		smoothed_values[i] = lambda*all_gene_level_stat[i] + (1-lambda)*shrinkage_target;
//		getchar();
	}
*/

	for(i=0;i<num_all_genes;i++)
		all_gene_level_stat[i] = smoothed_values[i];

//	printf("after\n");
//	for(i=0;i<20;i++)
//		printf("%lf\n", all_gene_level_stat[i]);
//	getchar();

/*
	double global_mean, global_var;
	global_means[0] = 0;
	global_vars[0]  = 1;
	global_means[1] = 0;
	global_vars[1]  = 1;
	for(i=2;i<num_all_genes;i++)
	{
		for(j=0;j<num_all_genes;j++)
			copy_all_gene_level_stat[j] = all_gene_level_stat[j];
		for(j=0;j<num_all_genes;j++)
		{
			if(j <= i)
				copy_all_gene_level_stat[j] = all_gene_level_stat[j] - all_gene_level_stat[i];
			else	
				copy_all_gene_level_stat[j] = 0;
		}
		estimate_global_MEAN_and_VAR(copy_all_gene_level_stat, gene_indexes, num_all_genes, &global_mean, &global_var);
		global_means[i] = global_mean;
		global_vars[i]  = global_var;
	}
*/
}





void evaluate_geneset_median_and_IQR(int *gene_indexes, int num_genes, double *all_gene_level_stat, int num_all_genes, double *median, double *IQR)
{
	int i;
	double scores_local[num_genes];
	for(i=0;i<num_genes;i++)
	    	scores_local[i] = all_gene_level_stat[gene_indexes[i]];
	size_t stride = 1;
//	gsl_sort(scores_local, stride, num_genes);
	*median = gsl_stats_median_from_sorted_data(scores_local, stride, num_genes);
	double geneset_IQR_lower = gsl_stats_quantile_from_sorted_data(scores_local, stride, num_genes, 0.25);
	double geneset_IQR_upper = gsl_stats_quantile_from_sorted_data(scores_local, stride, num_genes, 0.75);
	*IQR = fabs(geneset_IQR_upper - geneset_IQR_lower);
}

void estimate_global_MEDIAN_and_IQR_random_set(double *all_gene_level_stat, unsigned long *all_gene_indexes, int num_all_genes, 
					int *random_gene_set, int gene_set_size, int num_draws, double *median, double *IQR)
{	
	int i, j;
	double draws_IQR[num_draws], draws_median[num_draws];
	for(i=0;i<num_draws;i++)
	{
		sample_without_replacement(random_gene_set, gene_set_size, all_gene_indexes, num_all_genes);
		evaluate_geneset_median_and_IQR(random_gene_set, gene_set_size, all_gene_level_stat, num_all_genes, &draws_median[i], &draws_IQR[i]);
//		printf("median = %lf\tIQR = %lf\n", median, IQR);getchar();
	}
	size_t stride = 1;
	*median = gsl_stats_mean(draws_median, stride, num_draws);
	*IQR = gsl_stats_mean(draws_IQR, stride, num_draws);
//	printf("median = %lf\tIQR = %lf\n", *global_median, *global_IQR);getchar();
/*
	gsl_sort(draws_median, stride, num_draws);
	*global_median = gsl_stats_median_from_sorted_data(draws_median, stride, num_draws);
	gsl_sort(draws_IQR, stride, num_draws);
	*global_IQR = gsl_stats_median_from_sorted_data(draws_IQR, stride, num_draws);
*/
}


//double glob_var_tmp, glob_mean_tmp;
void estimate_global_MEAN_and_VAR_random_set(double *all_gene_level_stat, unsigned long *all_gene_indexes, int num_all_genes, 
					int *random_gene_set, int gene_set_size, int num_draws, double *mean, double *var)
{	
	int i, j;
//	double draws_var[num_draws], draws_mean[num_draws];

	*mean = 0.0; *var = 0.0;
	double sum_squares, sum, tmp, tmp1, tmp2;

	double gene_set_vals[gene_set_size];
	for(i=0;i<num_draws;i++)
	{
//		printf("before\n");fflush(stdout);
		sample_without_replacement(random_gene_set, gene_set_size, all_gene_indexes, num_all_genes);
//		printf("after\n");fflush(stdout);
		sum_squares = 0.0, sum = 0.0;
		for(j=0;j<gene_set_size;j++)
		{
			sum += all_gene_level_stat[random_gene_set[j]];
			sum_squares += all_gene_level_stat[random_gene_set[j]] * all_gene_level_stat[random_gene_set[j]];
			gene_set_vals[j] = all_gene_level_stat[random_gene_set[j]];
		}
		*mean += sum / gene_set_size;
		*var += sum_squares / gene_set_size - (sum / gene_set_size)*(sum / gene_set_size);
//		draws_mean[i] = sum / gene_set_size;
//		draws_var[i]  = (sum_squares / gene_set_size - (sum / gene_set_size)*(sum / gene_set_size)) / gene_set_size;
	}
	*mean = (*mean) / num_draws;
	*var  = (*var)  / (num_draws * gene_set_size);

/*
	size_t stride = 1;
	gsl_sort(draws_var, stride, num_draws);
	*var = gsl_stats_median_from_sorted_data(draws_var, stride, num_draws);
	gsl_sort(draws_mean, stride, num_draws);
	*mean = gsl_stats_median_from_sorted_data(draws_mean, stride, num_draws);
*/
/*
	size_t stride = 1;
	double absdev = gsl_stats_absdev_m(gene_set_vals, stride, gene_set_size, *mean);
	*var = absdev;
*/
}


// double gsl_stats_absdev (const double data[], size_t stride, size_t n)
// double gsl_stats_absdev_m (const double data[], size_t stride, size_t n, double mean)
void estimate_global_MEAN_and_VAR_KS_enrichment_score(double *all_gene_level_stat, unsigned long *all_gene_indexes, int num_all_genes, 
					int *random_gene_set, int gene_set_size, int num_draws, double *mean, double *var)
{
	int i;
//	int random_gene_set[geneset_size];
	double sum_positive_random_scores = 0, sum_squares_positive_random_scores = 0;
	double score;
	int leading_edge_index_dummy, leading_edge_num_genes_dummy;
	for(i=0;i<num_draws;i++)
	{	
		do
		{
//			printf("score =%lf\n", score);
			sample_without_replacement(random_gene_set, gene_set_size, all_gene_indexes, num_all_genes);
		}
		while( (score = ks_ES(random_gene_set, gene_set_size, num_all_genes, &leading_edge_index_dummy, &leading_edge_num_genes_dummy)) < 0);
		sum_positive_random_scores += score;
		sum_squares_positive_random_scores += score*score;
	}
	*mean = sum_positive_random_scores / num_draws;
//	*vars_KS = ( sum_squares_positive_random_scores / num_random - (*mean_KS)*(*mean_KS) ) / num_random;
	*var = ( sum_squares_positive_random_scores / num_draws - (*mean)*(*mean) ) / num_draws;
}

/*
// double gsl_stats_absdev (const double data[], size_t stride, size_t n)
// double gsl_stats_absdev_m (const double data[], size_t stride, size_t n, double mean)
void estimate_global_MEAN_and_VAR_KS_enrichment_score_old(double *all_gene_level_stat, int num_all_genes, int geneset_size, 
				double power, int num_random, double *mean_KS, double *vars_KS, geneset_set_scoring_params *scoring_params)
{
	int i;
//	int random_gene_set[geneset_size];
	double sum_positive_random_scores = 0, sum_squares_positive_random_scores = 0;
	double score = 0;
	int leading_edge_index_dummy, leading_edge_num_genes_dummy;
	for(i=0;i<num_random;i++)
	{	
		do
		{
//			printf("score =%lf\n", score);
			sample_without_replacement(scoring_params->ks_params.random_gene_set, geneset_size, scoring_params->indexes, num_all_genes);
		}
		while( (score = weighted_ks_ES(scoring_params->ks_params.random_gene_set, geneset_size, all_gene_level_stat, num_all_genes, power, &leading_edge_index_dummy, &leading_edge_num_genes_dummy)) < 0);
//		while( (score = ks_ES(scoring_params->ks_params.random_gene_set, geneset_size, num_all_genes, &leading_edge_index_dummy, &leading_edge_num_genes_dummy)) < 0);
		sum_positive_random_scores += score;
		sum_squares_positive_random_scores += score*score;
	}
	*mean_KS = sum_positive_random_scores / num_random;
//	*vars_KS = ( sum_squares_positive_random_scores / num_random - (*mean_KS)*(*mean_KS) ) / num_random;
	*vars_KS = ( sum_squares_positive_random_scores / num_random - (*mean_KS)*(*mean_KS) );
}
*/

void preprocess_thresholds_no_kNN_empirical(double *all_gene_level_stat, int num_all_genes, int min_geneset_size, geneset_set_scoring_params *scoring_params)
{
	int i, j;

	int initilaized_so_far = 0;
	int gene_set_size = min_geneset_size;

	double prev_location, prev_scale, location_estimate, scale_estimate;
	scoring_params->pre_calculate_location_and_scale(all_gene_level_stat, scoring_params->indexes, num_all_genes, scoring_params->random_gene_set, gene_set_size, scoring_params->num_draws, 
								&prev_location, &prev_scale);
	for(i=0;i<gene_set_size;i++)
	{
		scoring_params->location_estimates[i] = prev_location;
		scoring_params->scale_estimates[i] = prev_scale;
	}
	int last_processed_gene_set_size;

	// step_size control parameters
/*	int step_size = 5, increment = 5, decrement_ratio = 2, max_step_size = 300;
	double ratio_lower_threshold = 1.05;
	double ratio_upper_threshold = 1.15;
*/
	double ratio_scale1, ratio_scale2, max_scale, ratio_location_1, ratio_location_2, max_location, max; 
	for(i=gene_set_size+1;i<num_all_genes;i++)
	{
		gene_set_size = i;
		if(gene_set_size % scoring_params->step_size_control_params.step_size == 0)
		{
			scoring_params->pre_calculate_location_and_scale(all_gene_level_stat, scoring_params->indexes, num_all_genes, scoring_params->random_gene_set, gene_set_size, scoring_params->num_draws, 
								&scoring_params->location_estimates[gene_set_size], &scoring_params->scale_estimates[gene_set_size]);
			for(j=initilaized_so_far;j<gene_set_size;j++)
			{
				scoring_params->location_estimates[j] = (scoring_params->location_estimates[gene_set_size] + prev_location) / 2;
				scoring_params->scale_estimates[j]  = (scoring_params->scale_estimates[gene_set_size] + prev_scale) / 2;
			}
			initilaized_so_far = gene_set_size + 1;

			ratio_scale1 = scoring_params->scale_estimates[gene_set_size] / prev_scale;
			ratio_scale2 = prev_scale / scoring_params->scale_estimates[gene_set_size];
			max_scale = (ratio_scale1 > ratio_scale2 ? ratio_scale1 : ratio_scale2);

			ratio_location_1 = scoring_params->location_estimates[gene_set_size] / prev_location;
			ratio_location_2 = prev_location / scoring_params->location_estimates[gene_set_size];
			max_location = (ratio_location_1 > ratio_location_2 ? ratio_location_1 : ratio_location_2);

			if(scoring_params->step_size_control_params.ratio_type == location)
				max = max_location;
			else
				max = max_scale;

			if(max < scoring_params->step_size_control_params.ratio_lower_threshold && scoring_params->step_size_control_params.step_size < scoring_params->step_size_control_params.max_step_size)
				scoring_params->step_size_control_params.step_size += scoring_params->step_size_control_params.increment_step_size;
			else
			{
				if(max > scoring_params->step_size_control_params.ratio_upper_threshold && (int)(scoring_params->step_size_control_params.step_size / scoring_params->step_size_control_params.decrement_ratio_step_size) > 1)
					scoring_params->step_size_control_params.step_size = (int)(scoring_params->step_size_control_params.step_size / scoring_params->step_size_control_params.decrement_ratio_step_size);
			}

			prev_location = scoring_params->location_estimates[gene_set_size];
			prev_scale = scoring_params->scale_estimates[gene_set_size];
			last_processed_gene_set_size = gene_set_size;
		//	printf("%i: max_ratio = %-20.15lfstep_size = %i\n", gene_set_size, max, scoring_params->step_size_control_params.step_size);
		}
	}
	for(i=last_processed_gene_set_size+1;i<num_all_genes;i++)
	{
		scoring_params->location_estimates[i] = scoring_params->location_estimates[last_processed_gene_set_size];
		scoring_params->scale_estimates[i] = scoring_params->scale_estimates[last_processed_gene_set_size];
	}

// James-Stein Shrinakge (Strimmer paper)
/*	
	double copy_global_vars[num_all_genes];
	for(i=0;i<num_all_genes;i++)
		copy_global_vars[i] = scoring_params->location_estimates[i];

	size_t stride = 1;	
	gsl_sort(copy_global_vars, stride, num_all_genes);
	double median_global_vars = gsl_stats_median_from_sorted_data(copy_global_vars, stride, num_all_genes);
//	printf("median_global_vars =%-15.10lf\n", median_global_vars);

	double sum_var = 0, mse = 0;
	for(i=0;i<num_all_genes;i++)
	{
		sum_var += var_of_global_var[i];
		mse += (scoring_params->location_estimates[i] - median_global_vars) * (scoring_params->location_estimates[i] - median_global_vars);
	}
	double ratio = sum_var / mse;
	double lambda = ratio < 1? ratio : 1;

	for(i=0;i<num_all_genes;i++)
		scoring_params->location_estimates[i] = lambda*median_global_vars + (1 - lambda)*scoring_params->location_estimates[i];
	printf("sum_var = %-15.10lfmse = %-15.10lfratio = %-15.10lflambda = %-15.10lf\n", sum_var, mse, ratio, lambda);
*/
}

void preprocess_thresholds_no_kNN(double *all_gene_level_stat, int num_all_genes, int min_geneset_size, geneset_set_scoring_params *scoring_params)
{
//	time_t t  = start_timer();
//	printf("pre-processing global location and scale estimates...\n");
	if(scoring_params->pre_calculate_global_datasets_location_and_scale != 0)
	{
		scoring_params->pre_calculate_global_datasets_location_and_scale(all_gene_level_stat, scoring_params->indexes, num_all_genes, 
				&scoring_params->global_dataset_location, &scoring_params->global_dataset_scale);
	}
	else
	{
		printf("ERROR: scoring_params->pre_calculate_global_datasets_location_and_scale == 0!\n"); 
		exit(1);
	}

	int threshold_index, gene_set_size;
	if(scoring_params->pre_calculate_location_and_scale != 0)
	{
		preprocess_thresholds_no_kNN_empirical(all_gene_level_stat, num_all_genes, min_geneset_size, scoring_params);
	}
	else
	{
//		printf("pre_calculate_location_and_scale == 0\n");
		int i;
	/*
		double copy_all_gene_level_stat[num_all_genes];
		for(i=0;i<num_all_genes;i++)
			copy_all_gene_level_stat[i] = all_gene_level_stat[i];
	*/
		for(i=min_geneset_size;i<num_all_genes;i++)
		{
	/*
	//		double threshold = copy_all_gene_level_stat[num_all_genes - i];
			int threshold_index = num_all_genes - i;
	//		copy_all_gene_level_stat[num_all_genes - i] = 0;
	
			estimate_global_MEAN_and_MEAN_SUM_SQUARES_threshold(all_gene_level_stat, num_all_genes, threshold_index, &scoring_params->global_mean, &scoring_params->global_mss);
	//		estimate_global_MEAN_and_VAR_threshold(copy_all_gene_level_stat, scoring_params->indexes, num_all_genes, threshold, &scoring_params->global_mean, &scoring_params->global_var);
			scoring_params->location_estimates[num_all_genes - i] = global_mean;
			scoring_params->global_mean_sum_squares[num_all_genes - i] = global_mss;
	//		scoring_params->scale_estimates[num_all_genes  - i] = global_var;
	*/
			gene_set_size = i;
	//		if(scoring_params->geneset_ES == geneset_adaptive_to_noise_level_ES)
			threshold_index = gene_set_size;
	//		else
	//			threshold_index = num_all_genes - 1;

			estimate_global_MEAN_and_VAR_and_MSS_threshold_exact(all_gene_level_stat, num_all_genes, gene_set_size, threshold_index, 
				&scoring_params->location_estimates[i], &scoring_params->scale_estimates[i], &scoring_params->mean_sum_squares_estimates[i]);
//			estimate_global_MEAN_and_VAR_and_MSS_threshold_exact(all_gene_level_stat, num_all_genes, gene_set_size, 
//							threshold_index, &scoring_params->location_estimates[i], &scoring_params->scale_estimates[i], &scoring_params->mean_sum_squares_estimates[i]);


//			estimate_global_MEAN_and_VAR_exact(all_gene_level_stat, num_all_genes, i, &scoring_params->location_estimates[i], &scoring_params->scale_estimates[i]);
	//		estimate_global_MEDIAN_and_IQR_random_set(all_gene_level_stat, scoring_params->indexes, num_all_genes, 
	//								gene_set_size, &scoring_params->global_medians[i], &scoring_params->global_IQRs[i]);
//			if(i<50 || i>num_all_genes - 50 || i % 200 == 0)
//			{
//				printf("%i: global_medians = %-20.15lfglobal_IQRs = %-20.15lf\n", i, scoring_params->global_medians[i], scoring_params->global_IQRs[i]);
//				fflush(stdout);
//			}
		}
		for(i=0;i<min_geneset_size;i++)
		{
			scoring_params->location_estimates[i] = scoring_params->location_estimates[min_geneset_size];
			scoring_params->scale_estimates[i] = scoring_params->scale_estimates[min_geneset_size];
			scoring_params->mean_sum_squares_estimates[i] = scoring_params->mean_sum_squares_estimates[min_geneset_size];
		}
	}
//	printf("done\n");
//	print_time(t, stdout);
}


extern gsl_matrix *covariates;
//extern gsl_vector *response;
extern gsl_vector *beta;
extern gsl_matrix *cov;
extern gsl_multifit_linear_workspace *work;
extern gsl_matrix *U_global;

double regression_score(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes)//, geneset_set_scoring_params *scoring_params)
{
//	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
//		return 0;

	if(covariates->size1 != num_all_genes)
	{
		printf("ERROR: 'regression_score', covariates->size1(%i) != num_all_genes(%i)\n", covariates->size1, num_all_genes);
		exit(1);
	}
	size_t n = covariates->size1, p = covariates->size2;
//	gsl_multifit_linear_workspace *work = gsl_multifit_linear_alloc(n, p);
//	gsl_matrix *X = gsl_matrix_alloc(n, p);
//	gsl_matrix_set_all(X, 1.0);

	gsl_vector_view vview = gsl_matrix_column(covariates, p - 1);
//	gsl_vector_set_all(&vview.vector, 1.0);
//	vview = gsl_matrix_column(X, 1);
	gsl_vector_set_zero(&vview.vector);

	int i;
	int col_index = p - 1;
	for(i=0;i<num_genes;i++)
		gsl_matrix_set(covariates, genes[i], col_index, 1.0);// / sqrt(num_genes));

	if(false)//U_global != 0)
	{
//		printf("U_global != 0\n");
//		getchar();
		gsl_vector_view col_view = gsl_matrix_column(covariates, col_index);
		double *y = col_view.vector.data;
		double y_corrected[col_view.vector.size];
		subtract_off_covariate_effects(y, col_view.vector.size, U_global, y_corrected);
		for(i=0;i<covariates->size1;i++)
			gsl_matrix_set(covariates, i, col_index, y_corrected[i]);// / sqrt(num_genes));
	}

	vview = gsl_vector_view_array(all_gene_level_stat, num_all_genes);

	double chisq;
	// y = X*beta
	// cov   -> variance-covariance matrix of the model parameters, estimated from the scatter of the observations about the best-fit
	// chisq -> sum of squares of the residuals from the best-fit
//	gsl_multifit_linear(covariates, &vview.vector, beta, cov, &chisq, work);

	double tol = 0.01; size_t rank;
	gsl_multifit_linear_svd(covariates, &vview.vector, tol, &rank, beta, cov, &chisq, work);

/*
//	printf("vview.vector.size: %i\n", vview.vector.size);
	gsl_vector *weights = gsl_vector_alloc(vview.vector.size);

//	printf("before gsl_multifit_linear_residuals\n");
//	This function computes the vector of residuals r = y - X c for the observations y, coefficients c and matrix of predictor variables X. 
	gsl_multifit_linear_residuals(covariates, &vview.vector, beta, weights);

//	printf("weights\n\n");
	for(i=0;i<weights->size;i++)
	{	
//		printf("%lf\n", weights->data[i]);
		weights->data[i] = 1.0 / ((weights->data[i]) * (weights->data[i]));
	}
//	getchar();

	gsl_multifit_wlinear(covariates, weights, &vview.vector, beta, cov, &chisq, work);

	gsl_vector_free(weights);
*/

/*
	// do SVD
	eigs_SVD out;
	out.U = gsl_matrix_alloc(covariates->size1, covariates->size2);
	out.S = gsl_vector_alloc(covariates->size2);
	out.V = gsl_matrix_alloc(covariates->size2, covariates->size2);
	gsl_matrix *covariates_copy = gsl_matrix_alloc(covariates->size1, covariates->size2);
	
	gsl_matrix_memcpy(covariates_copy, covariates);
	//	printf("svd: modified_Golub_Reinsch\n");
	SVD_modified_Golub_Reinsch(covariates_copy, &out);
	// do SVD

	gsl_matrix *beta_hat = gsl_matrix_alloc(covariates, p);
	gsl_matrix *X_tr_time_X_inv = gsl_matrix_alloc(covariates->size2, covariates->size2);
	mat_mat_multiply(&mview_U.matrix, S_diag, U_times_S, CblasNoTrans, CblasNoTrans);

//////////////////////////
// Cholesky-based solution
//////////////////////////
//	This function factorizes the positive-definite square matrix A into the Cholesky decomposition A = L L^T. 
//	On output the diagonal and lower triangular part of the input matrix A contain the matrix L. The upper triangular 
//	part of the input matrix contains L^T, the diagonal terms being identical for both L and L^T. If the matrix is 
//	not positive-definite then the decomposition will fail, returning the error code GSL_EDOM. 
	gsl_linalg_cholesky_decomp(gsl_matrix * A)

//	This function solves the system A x = b in-place using the Cholesky decomposition of A into the matrix cholesky given 
//	by gsl_linalg_cholesky_decomp. On input x should contain the right-hand side b, which is replaced by the solution on output. 
	gsl_linalg_cholesky_svx (const gsl_matrix * cholesky, gsl_vector * x)

/////////////////////
// SVD-based solution
/////////////////////
//    	This function solves the system A x = b using the singular value decomposition (U, S, V) of A given by gsl_linalg_SV_decomp.
//	Only non-zero singular values are used in computing the solution. The parts of the solution corresponding to singular values 
//	of zero are ignored. Other singular values can be edited out by setting them to zero before calling this function.
//	In the over-determined case where A has more rows than columns the system is solved in the least squares sense, returning the 
//	solution x which minimizes ||A x - b||_2. 
	gsl_linalg_SV_solve(gsl_matrix *U, gsl_matrix *V, gsl_vector *S, const gsl_vector *b, gsl_vector *x)


/////////////////////////////
// Householder-based solution
/////////////////////////////
//	This function solves the system A x = b in-place using Householder transformations. On input x should contain the right-hand 
//	side b, which is replaced by the solution on output. The matrix A is destroyed by the Householder transformations. 
	gsl_linalg_HH_svx(gsl_matrix * A, gsl_vector * x)


////////////////////
// QR-based solution
////////////////////
//	This function factorizes the M-by-N matrix A into the QRP^T decomposition A = Q R P^T. On output the diagonal and 
//	upper triangular part of the input matrix contain the matrix R. The permutation matrix P is stored in the permutation p. 
//	The sign of the permutation is given by signum. It has the value (-1)^n, where n is the number of interchanges in the 
//	permutation. The vector tau and the columns of the lower triangular part of the matrix A contain the Householder 	
//	coefficients and vectors which encode the orthogonal matrix Q. The vector tau must be of length k=\min(M,N). The matrix Q 
//	is related to these components by, Q = Q_k ... Q_2 Q_1 where Q_i = I - \tau_i v_i v_i^T and v_i is the Householder vector 
//	v_i = (0,...,1,A(i+1,i),A(i+2,i),...,A(m,i)). This is the same storage scheme as used by LAPACK. On output the norms of 
//	each column of R are stored in the vector norm. 
	gsl_linalg_QRPT_decomp (gsl_matrix * A, gsl_vector * tau, gsl_permutation * p, int *signum, gsl_vector * norm)

//	This function solves the system A x = b in-place using the QRP^T decomposition of A into (QR,tau,p). On input x should 
//	contain the right-hand side b, which is replaced by the solution on output. 
	gsl_linalg_QRPT_svx (const gsl_matrix * QR, const gsl_vector * tau, const gsl_permutation * p, gsl_vector * x)
*/


//	printf("beta = %lf\n", gsl_vector_get(beta, 1));
//	printf("sd(beta) = %lf\n", sqrt(gsl_matrix_get(cov, 1, 1)));
//	double tss = gsl_stats_tss(y->data, 1, n);

//	printf("chisq = %lf\n", chisq);
//	printf("tss = %lf\n", tss);
	double se = sqrt(gsl_matrix_get(cov, p - 1, p - 1));
//	printf("se = %lf\n", se);
//	getchar();
	double shrinkage = 0.0;//
	double score = gsl_vector_get(beta, p - 1) / (se + shrinkage);
//	double score = 1.0 - chisq / tss;

//	gsl_matrix_free(cov);
//	gsl_vector_free(beta);

//	printf("score = %lf\n", score);
//	printf("score1 = %lf\n", score1);
//	getchar();
	return score;
}

extern bool is_regression_initialized;
extern double global_sample_mean, global_sample_var;

double **permuted_gene_level_stat;

bool first_perm_response = true;
void get_permutation_stats(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, 
				double global_mean, double global_var, double *mean, double *sd)
{
	int j;
	int num_perm_response = 200;
	if(first_perm_response == true)
	{
		permuted_gene_level_stat = malloc(sizeof(double *) * num_perm_response);
		for(j=0;j<num_perm_response;j++)
		{
			permuted_gene_level_stat[j] = malloc(sizeof(double) * num_all_genes);
			memcpy(permuted_gene_level_stat[j], all_gene_level_stat, sizeof(double) * num_all_genes);
			permute(permuted_gene_level_stat[j], num_all_genes);
		}
		first_perm_response = false;
	}	
	double sum = 0, sum_sq = 0;
	double perm_scores[num_perm_response];
//	FILE *fp = fopen("permuted_scores", "w");
	for(j=0;j<num_perm_response;j++)
	{
//		perm_scores[j] = -random_set_ES_bootstrap_Std(genes, num_genes, permuted_gene_level_stat[j], num_all_genes, global_mean, global_var);
		perm_scores[j] = -evaluate_random_set_ES(genes, num_genes, permuted_gene_level_stat[j], num_all_genes, global_mean, global_var);
//		fprintf(fp, "%lf\n", perm_scores[j]);
//		printf("%10.2lf", perm_scores[j]);
		sum += perm_scores[j]; sum_sq += perm_scores[j] * perm_scores[j];
	}
//	fclose(fp);
	*mean = sum / num_perm_response;
	*sd = sqrt(sum_sq / num_perm_response - (*mean) * (*mean));
//	printf("mean =%10.2lf, sd =%10.2lf", *mean, *sd);
//	getchar();
}



bool overlap_regions(int region1_count, int region2_count)
{	
	return region1_count > 0 && region2_count > 0;
}

bool union_regions(int region1_count, int region2_count)
{	
	return region1_count > 0 || region2_count > 0;
}

bool index_included(int index, int *a)
{
	return a[index] > 0;
}

// assume arrays are of the same length
int count_num_overlap(int *arr1, int *arr2, int num)
{
	int i, j;
	int num_overlap = 0;
	for(i=0;i<num;i++)
	{
		if(overlap_regions(arr1[i], arr2[i]) == true)
			num_overlap++;
	}
	return num_overlap;
}

int *get_ovelap(int *arr1, int *arr2, int num, int *num_ovelap)
{
	int i;
	*num_ovelap = count_num_overlap(arr1, arr2, num);
	int *overlap = malloc(sizeof(int) * (*num_ovelap));
	int ind = 0;
	for(i=0;i<num;i++)
	{
		if(overlap_regions(arr1[i], arr2[i]) == true)
		{
			overlap[ind] = i;
			ind++;
		}
	}
	return overlap;
}

int count_num_union(int *arr1, int *arr2, int num)
{
	int i;
	int num_union = 0;
	for(i=0;i<num;i++)
	{
		if(union_regions(arr1[i], arr2[i]) == true)
			num_union++;
	}
	return num_union;
}


int *get_union(int *arr1, int *arr2, int num, int *num_union)
{
	int i;
	*num_union = count_num_union(arr1, arr2, num);
	int *set_union = malloc(sizeof(int) * (*num_union));
	int ind = 0;
	for(i=0;i<num;i++)
	{
		if(union_regions(arr1[i], arr2[i]) == true)
		{
			set_union[ind] = i;
			ind++;
		}
	}
	return set_union;
}

extern double mean_before_centering;

double euclidean_norm(int *a, double *evidence_scores, int num)
{
	int i;
	double norm = 0;
	for(i=0;i<num;i++)
	{
		if(index_included(i, a) == true)
//			norm += (-evidence_scores[i] + mean_before_centering);
			norm += (evidence_scores[i] * evidence_scores[i]);
	}
	return sqrt(norm);
}

double eval_weighted_similarity(int *arr1, int *arr2, double *evidence_scores, int num_evidence_scores)
{
	int i, j;
	int num_overlap = 0;
	double weighted_sim = 0;
	for(i=0;i<num_evidence_scores;i++)
	{
		if(overlap_regions(arr1[i], arr2[i]) == true)
		{
//			weighted_sim += (-evidence_scores[i] + mean_before_centering);
			weighted_sim += (evidence_scores[i] * evidence_scores[i]);
			num_overlap++;
		}
	}
	double norm1 = euclidean_norm(arr1, evidence_scores, num_evidence_scores);
	double norm2 = euclidean_norm(arr2, evidence_scores, num_evidence_scores);
	weighted_sim = weighted_sim / (norm1 * norm2);
//	if(num_overlap > 0)
//		weighted_sim = weighted_sim / num_overlap;
	return weighted_sim;
}

void get_permutation_distance_targets_sets(int *genes1, int *genes2, double *all_gene_level_stat, int num_all_genes, double *mean, double *sd)
{
	int j;
	int num_perm_response = 200;
	if(first_perm_response == true)
	{
		permuted_gene_level_stat = malloc(sizeof(double *) * num_perm_response);
		for(j=0;j<num_perm_response;j++)
		{
			permuted_gene_level_stat[j] = malloc(sizeof(double) * num_all_genes);
			memcpy(permuted_gene_level_stat[j], all_gene_level_stat, sizeof(double) * num_all_genes);
			permute(permuted_gene_level_stat[j], num_all_genes);
		}
		first_perm_response = false;
	}	
	double sum = 0, sum_sq = 0;
	double perm_scores[num_perm_response];
//	FILE *fp = fopen("permuted_scores", "w");
	for(j=0;j<num_perm_response;j++)
	{
//		perm_scores[j] = -random_set_ES_bootstrap_Std(genes, num_genes, permuted_gene_level_stat[j], num_all_genes, global_mean, global_var);
		perm_scores[j] = eval_weighted_similarity(genes1, genes2, permuted_gene_level_stat[j], num_all_genes);
//		perm_scores[j] = -evaluate_random_set_ES(genes, num_genes, permuted_gene_level_stat[j], num_all_genes, global_mean, global_var);
//		fprintf(fp, "%lf\n", perm_scores[j]);
//		printf("%10.2lf", perm_scores[j]);
		sum += perm_scores[j]; sum_sq += perm_scores[j] * perm_scores[j];
	}
//	fclose(fp);
	*mean = sum / num_perm_response;
	*sd = sqrt(sum_sq / num_perm_response - (*mean) * (*mean));
//	printf("mean =%10.2lf, sd =%10.2lf", *mean, *sd);
//	getchar();
}


void get_permutation_overlap_targets_sets(int *genes1, int *genes2, double *all_gene_level_stat, int num_all_genes, 
				double global_mean, double global_var, double *mean, double *sd)
{
	int num_ovelap;
	int *overlap_genes = get_ovelap(genes1, genes2, num_all_genes, &num_ovelap);
	int j;
	int num_perm_response = 200;
	if(first_perm_response == true)
	{
		permuted_gene_level_stat = malloc(sizeof(double *) * num_perm_response);
		for(j=0;j<num_perm_response;j++)
		{
			permuted_gene_level_stat[j] = malloc(sizeof(double) * num_all_genes);
			memcpy(permuted_gene_level_stat[j], all_gene_level_stat, sizeof(double) * num_all_genes);
			permute(permuted_gene_level_stat[j], num_all_genes);
		}
		first_perm_response = false;
	}	
	double sum = 0, sum_sq = 0;
	double perm_scores[num_perm_response];
//	FILE *fp = fopen("permuted_scores", "w");
	for(j=0;j<num_perm_response;j++)
	{
		perm_scores[j] = -evaluate_random_set_ES(overlap_genes, num_ovelap, permuted_gene_level_stat[j], num_all_genes, global_mean, global_var);
//		fprintf(fp, "%lf\n", perm_scores[j]);
//		printf("%10.2lf", perm_scores[j]);
		sum += perm_scores[j]; sum_sq += perm_scores[j] * perm_scores[j];
	}
//	fclose(fp);
	*mean = sum / num_perm_response;
	*sd = sqrt(sum_sq / num_perm_response - (*mean) * (*mean));
	free(overlap_genes);
//	printf("mean =%10.2lf, sd =%10.2lf", *mean, *sd);
//	getchar();
}


//double significance_threshold_filter_global = 3.0;
bool pass_significance_threshold(double obs_score, double mean_random, double significance_threshold_filter, double sd_random)
{
	return obs_score > mean_random + significance_threshold_filter*sd_random;
}

bool pass_significance_threshold_bootstrap(double CI_lower_bound)
{
	return CI_lower_bound > 0;
}


int num_agreement_scores = 0;
int num_agreement_pass = 0;
int num_agreement_fail = 0;
int num_agreement_pass_fail = 0;
int num_agreement_fail_pass = 0;
int total_num_scores = 0;

bool first_time_global_mean = true;
double global_mean_exact_null_ES;

double geneset_random_set_exact_null_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, 
								geneset_set_scoring_params *scoring_params)
{
	// if input option 'use_regression == yes' => 'is_regression_initialized == true'
	// do regression scoring. Otherwise do standard cERMIT scoring (as described in the paper)
	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
		return 0.0;

	total_num_scores +=1;

	int i;
	if(first_time_global_mean == false)
	{
		global_mean_exact_null_ES = 0;
		for(i=0;i<num_all_genes;i++);
			global_mean_exact_null_ES += all_gene_level_stat[i];
		global_mean_exact_null_ES =  global_mean_exact_null_ES / num_all_genes;
		first_time_global_mean = false;
		printf("first: global_mean_exact_null_ES = %lf\n", global_mean_exact_null_ES);
	}


	double global_var, global_mean;
	estimate_global_MEAN_and_VAR_exact(all_gene_level_stat, num_all_genes, num_genes, &global_mean, &global_var);
	
	// high evidence -> large positive score
	double observed_score;
	if(is_regression_initialized == true)
		observed_score = -regression_score(genes, num_genes, all_gene_level_stat, num_all_genes);//, scoring_params);
	else
		observed_score = -evaluate_random_set_ES(genes, num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);

	if(scoring_params->bootstrap_motif_score == true)
	{
		double bootstrap_Std_threshold = scoring_params->bootstrap_Std_threshold;
		int num_draws = scoring_params->num_bootstrap_draws;
		double CI_lower_bound = random_set_ES_bootstrap_Std(genes, num_genes, all_gene_level_stat, num_all_genes, observed_score, 
					bootstrap_Std_threshold, num_draws, is_regression_initialized, global_mean, global_var);
		if(pass_significance_threshold_bootstrap(CI_lower_bound) == false)
			observed_score = 0;
	}
	return -observed_score;
}


double geneset_random_set_empirical_null_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params)
{
	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
		return 0;
	double global_mean = scoring_params->location_estimates[num_genes];
	double global_var  = scoring_params->scale_estimates[num_genes];
	return evaluate_random_set_ES(genes, num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);
}

double geneset_ks_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params)
{
	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
//		return HUGE;
		return 0;
	int leading_edge_index, leading_edge_num_genes;
	double ks_score = ks_ES(genes, num_genes, num_all_genes, &leading_edge_index, &leading_edge_num_genes);
	if(scoring_params->pre_calculate_location_and_scale == 0)
		return -ks_score;
	else
	{
		double global_mean = scoring_params->location_estimates[num_genes];
		return -ks_score / global_mean;
	}
}

double geneset_stable_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params)
{
	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
		return 0;
	double global_mean = scoring_params->location_estimates[num_genes];
	double global_var  = scoring_params->scale_estimates[num_genes];
	return random_set_ES_stable(genes, num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);
}

double geneset_stable_bootstrap_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params)
{
	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
		return 0;
	double global_mean = scoring_params->location_estimates[num_genes];
	double global_var  = scoring_params->scale_estimates[num_genes];
//	return random_set_ES_stable(genes, num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);
	return random_set_ES_stable_bootstrap(genes, num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);
}

double geneset_adaptive_to_noise_level_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params)
{
	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
		return 0;
	return random_set_ES_adaptive_to_noise_level(genes, num_genes, all_gene_level_stat, num_all_genes, scoring_params);
}

double geneset_bootstrap_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params)
{
	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
//		return HUGE;
		return 0;

	double global_mean = scoring_params->location_estimates[num_genes];
	double global_var  = scoring_params->scale_estimates[num_genes];
	return random_set_ES_bootstrap_all_genes(genes, num_genes, all_gene_level_stat, num_all_genes, scoring_params->indexes, global_mean, global_var);
}

double geneset_median_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params)
{
	if(num_genes < scoring_params->min_gene_set_size_absolute_threshold || num_genes > scoring_params->max_gene_set_size_absolute_threshold)
//		return HUGE;
		return 0;
	double global_median = scoring_params->location_estimates[num_genes];
	double global_IQR    = scoring_params->scale_estimates[num_genes];
	return evaluate_set_ES_median(genes, num_genes, all_gene_level_stat, num_all_genes, global_median, global_IQR);
}

/*
// here we don't need sorted order for 'all_gene_level_stat'
double random_set_ES(d_gs *dg, double *all_gene_level_stat, int num_all_genes, double global_location, double global_scale, geneset_set_scoring_params *scoring_params)
{
	double global_mean = scoring_params->location_estimates[dg->num_genes];
	double global_var  = scoring_params->scale_estimates[dg->num_genes];

	// 'preprocess_thresholds_no_kNN_new' (ORIGINAL random set statistic + empirical scaling)
//	return evaluate_random_set_ES(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);
/*
	// 'preprocess_thresholds_no_kNN_new' (unweighted KS statistic + empirical scaling)
	int leading_edge_index, leading_edge_num_genes;
	double ks_score = ks_ES(dg->genes, dg->num_genes, num_all_genes, &leading_edge_index, &leading_edge_num_genes);
	return -ks_score / scoring_params->ks_params.means_KS[dg->num_genes];
*/

/*
	// no pre-processing required (unweighted KS statistic + analytic scaling)
	int leading_edge_index, leading_edge_num_genes;
	double ks_score = ks_ES(dg->genes, dg->num_genes, num_all_genes, &leading_edge_index, &leading_edge_num_genes);
	return -ks_score;
*/
//	double power = 0;
//	double weighted_ks_score = weighted_ks_ES(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, power);
//	return -(weighted_ks_score / fabs(means_KS[dg->num_genes]));
//	return -weighted_ks_score;
//	return -( (weighted_ks_score - means_KS[dg->num_genes]) / sqrt(vars_KS[dg->num_genes]) ) ;

	// 'estimate_global_MEAN_and_VAR_exact' (ORIGINAL random set statistic + analytic scaling)
//	return evaluate_random_set_ES(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);

//	return random_set_ES_stable(dg, all_gene_level_stat, num_all_genes, global_mean, global_var);
//	return random_set_ES_adaptive_to_noise_level(dg, all_gene_level_stat, num_all_genes, scoring_params->global_mean_sum_squares, scoring_params);

//	double bootstrap_score =  random_set_ES_bootstrap_all_genes(dg, all_gene_level_stat, num_all_genes);

/*
	// nned to make the necessary changes in 'estimate_global_MEAN_and_VAR_random_set'
	double global_median = scoring_params->global_medians[dg->num_genes];
	double global_IQR  = scoring_params->global_IQRs[dg->num_genes];
	return evaluate_set_ES_median(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, global_median, global_IQR);
*/
//	return evaluate_random_set_ES_harmonic_mean(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);

/*
//	num_scored++;
	double observed_score = -ks_score;//evaluate_random_set_ES(dg->genes, dg->num_genes, all_gene_level_stat, num_all_genes, global_mean, global_var);
//	printf("observed_score = %lf\n", observed_score);getchar();
//	tol = 5;//fabs(running_sum_scores / num_scored - current_best_score) / 2;
	if(true)//fabs(observed_score - current_best_score) >= tol)
	{
		double new_score = random_set_ES_stable(dg, all_gene_level_stat, num_all_genes, global_mean, global_var);
//		double new_score = -ks_ES(dg->genes, dg->num_genes, num_all_genes, &leading_edge_index, &leading_edge_num_genes);
		current_best_score = MIN(observed_score, new_score);
//		running_sum_scores += new_score;
		return new_score;
	}
	else
	{
//		running_sum_scores += observed_score;
		return observed_score;
	}

*/
/*
	int leading_edge_index, leading_edge_num_genes;
	double ks_score = ks_ES(dg->genes, dg->num_genes, num_all_genes, &leading_edge_index, &leading_edge_num_genes);
	return -ks_score;

	int i;
	double sum = 0.0;
	for(i=0;i<dg->num_genes;i++)
	    	sum += all_gene_level_stat[dg->genes[i]];
	double mean_d_gs = sum / dg->num_genes;
	double Z_score = (mean_d_gs - global_mean)/sqrt(global_var);

	int ind = 0;
	while(ind < strlen(dg->degen_oligo) && (dg->degen_oligo[ind] == 'A' || dg->degen_oligo[ind] == 'G' ||dg->degen_oligo[ind] == 'C' ||dg->degen_oligo[ind] == 'T'))
		ind++;
	
	if(ind == strlen(dg->degen_oligo) && ind == 6)
	{
		FILE *fp = fopen("scores_6_mers", "a");
		fprintf(fp, "%i\t%i\t%i\t%lf\t%lf\n", leading_edge_index, leading_edge_num_genes, dg->num_genes, -Z_score, ks_score);
		fclose(fp);
	}
	else
	{
		FILE *fp = fopen("scores_evolved", "a");
		fprintf(fp, "%i\t%i\t%i\t%lf\t%lf\t", leading_edge_index, leading_edge_num_genes, dg->num_genes, -Z_score, ks_score);
		if(leading_edge_num_genes > 10)
		{
			double sum1 = 0.0;
			for(i=0;i<leading_edge_num_genes;i++)
				sum1 += all_gene_level_stat[dg->genes[i]];
			double mean_d_gs1 = sum1 / leading_edge_num_genes;
			double Z_score1 = (mean_d_gs1 - global_means[leading_edge_num_genes])/sqrt(global_vars[leading_edge_num_genes]);
			fprintf(fp, "%lf\n", -Z_score1);
		}
		else
			fprintf(fp, "%lf\n", 0);
		fclose(fp);
	}
	return -ks_score;

//	printf("Z_score = %5.2lf\n", Z_score);fflush(stdout);

	double Z_score1 = Z_score;
	if(leading_edge_num_genes > 10)
	{
		double sum1 = 0.0;
		for(i=0;i<leading_edge_num_genes;i++)
			sum1 += all_gene_level_stat[dg->genes[i]];
		double mean_d_gs1 = sum1 / leading_edge_num_genes;
		Z_score1 = (mean_d_gs1 - global_means[leading_edge_num_genes])/sqrt(global_vars[leading_edge_num_genes]);
	
	//	printf("Z_score1 = %5.2lf\n", Z_score1);fflush(stdout);
	}
*/

// original random set (with no pre-processing of means/vars)
/*
 	size_t stride = 1;
 	double x_bar_mean = gsl_stats_mean(all_gene_level_stat, stride, num_all_genes);

	int m = dg->num_genes;
	int G = num_all_genes;
	double mean_sum_squared = 0;
	for(i=0;i<G;i++)
		mean_sum_squared += (all_gene_level_stat[i])*(all_gene_level_stat[i]);
	mean_sum_squared = mean_sum_squared / G;
	double x_bar_var = ((mean_sum_squared-x_bar_mean*x_bar_mean)*(G-m))/(m*(G-1));

	// mean=0,var=1 r.v. under H_null: category is not enriched in differentially expressed genes
	double Z_score = (mean_d_gs-x_bar_mean)/sqrt(x_bar_var);
	return Z_score;

}
*/

/*
void preprocess_thresholds_no_kNN_empirical(double *all_gene_level_stat, int num_all_genes, int min_geneset_size, geneset_set_scoring_params *scoring_params)
{
	preprocess_thresholds_no_kNN_empirical_new(all_gene_level_stat, num_all_genes, min_geneset_size, scoring_params);
	return;

	bool fast = true;
	time_t t  = start_timer();

	int i, j;

	printf("pre-processing global means and variances...\n");
	double global_mean, global_var;
	estimate_global_MEAN_and_VAR(all_gene_level_stat, scoring_params->indexes, num_all_genes, &global_mean, &global_var);

	int initilaized_so_far = 0;
	int gene_set_size = min_geneset_size;

	double prev_global_mean, prev_global_var;
//	estimate_global_MEAN_and_VAR_random_set(all_gene_level_stat, scoring_params->indexes, num_all_genes, gene_set_size, &prev_global_mean, &prev_global_var);
	estimate_global_MEDIAN_and_IQR_random_set(all_gene_level_stat, scoring_params->indexes, num_all_genes, gene_set_size, &prev_global_mean, &prev_global_var);

	double power = 0;
	int num_random_ks = 400;
	double mean_KS, var_KS, prev_mean_KS, prev_var_KS;
	estimate_global_MEAN_and_VAR_KS_enrichment_score(all_gene_level_stat, num_all_genes, gene_set_size, power, num_random_ks, &prev_mean_KS, &prev_var_KS, scoring_params);

	for(i=0;i<gene_set_size;i++)
	{
		scoring_params->global_means[i] = prev_global_mean;
		scoring_params->global_vars[i] = prev_global_var;

		scoring_params->ks_params.means_KS[i] = prev_mean_KS;
		scoring_params->ks_params.vars_KS[i] = prev_var_KS;
	}

	int last_processed_gene_set_size;

	// step_size control parameters
	int step_size = 5, increment = 5, decrement_ratio = 2, max_step_size = 300;
	double ratio_lower_threshold = 1.05;
	double ratio_upper_threshold = 1.15;

	double ratio_var_ks_1, ratio_var_ks_2, max_var_ks, ratio_var1, ratio_var2, max_var, max; 
	double ratio_mean_ks_1, ratio_mean_ks_2, max_mean_ks; 
	for(i=gene_set_size+1;i<num_all_genes;i++)
	{
		gene_set_size = i;
		if(gene_set_size % step_size == 0)
		{
			estimate_global_MEAN_and_VAR_KS_enrichment_score(all_gene_level_stat, num_all_genes, gene_set_size, power, num_random_ks, &mean_KS, &var_KS, scoring_params);
			scoring_params->ks_params.means_KS[gene_set_size] = mean_KS;
			scoring_params->ks_params.vars_KS[gene_set_size] = var_KS;

		//	estimate_global_MEAN_and_VAR_random_set(all_gene_level_stat, scoring_params->indexes, num_all_genes, gene_set_size, &global_mean, &global_var);
			estimate_global_MEDIAN_and_IQR_random_set(all_gene_level_stat, scoring_params->indexes, num_all_genes, gene_set_size, &global_mean, &global_var);
	
			scoring_params->global_means[gene_set_size] = global_mean;
			scoring_params->global_vars[gene_set_size]  = global_var;
//			var_of_global_var[gene_set_size]  = var_global_var;

			for(j=initilaized_so_far;j<gene_set_size;j++)
			{
				scoring_params->global_means[j] = (global_mean + prev_global_mean) / 2;
				scoring_params->global_vars[j]  = (global_var + prev_global_var) / 2;

				scoring_params->ks_params.means_KS[j] = (mean_KS + prev_mean_KS) / 2;
				scoring_params->ks_params.vars_KS[j] = (var_KS + prev_var_KS) / 2;
		
			}
			initilaized_so_far = gene_set_size + 1;

			// KS entichment score
			ratio_var_ks_1 = var_KS / prev_var_KS;
			ratio_var_ks_2 = prev_var_KS / var_KS;
			max_var_ks = (ratio_var_ks_1 > ratio_var_ks_2 ? ratio_var_ks_1 : ratio_var_ks_2);

			ratio_mean_ks_1 = mean_KS / prev_mean_KS;
			ratio_mean_ks_2 = prev_mean_KS / mean_KS;
			max_mean_ks = (ratio_mean_ks_1 > ratio_mean_ks_2 ? ratio_mean_ks_1 : ratio_mean_ks_2);

			// random set enrichment score
			ratio_var1 = global_var / prev_global_var;
			ratio_var2 = prev_global_var / global_var;
			max_var = (ratio_var1 > ratio_var2 ? ratio_var1 : ratio_var2);
	
//			max = max_var;
//			max = max_var_ks;
			max = max_mean_ks;

			prev_global_mean = global_mean;
			prev_global_var = global_var;

			prev_mean_KS = mean_KS;
			prev_var_KS = var_KS;

			if(max < ratio_lower_threshold && step_size < max_step_size)
				step_size += increment;
			else
			{
				if(max > ratio_upper_threshold && (int)(step_size / decrement_ratio) > 1)
					step_size = (int)(step_size / decrement_ratio);
			}
			last_processed_gene_set_size = gene_set_size;

			printf("%i: max_ratio = %-20.15lfstep_size = %i\n", gene_set_size, max, step_size);
//			printf("(gene set size = %i): MAX(mean_KS / prev_mean_KS, prev_mean_KS / mean_KS) = %-15.10lf\n", gene_set_size, max_mean);fflush(stdout);
//			printf("(gene set size = %i): MAX(var_KS / prev_var_KS, prev_var_KS / var_KS) = %-15.10lf\n", gene_set_size, max_var);fflush(stdout);
//			printf("%i: step_size = %i\n", i, step_size);

		//	printf("%i: global_mean = %-10.2lfglobal_var = %10.2lf\n", gene_set_size, global_mean, global_var);
		}
	}

	for(i=last_processed_gene_set_size+1;i<num_all_genes;i++)
	{
		scoring_params->global_means[i] = scoring_params->global_means[i];
		scoring_params->global_IQRs[i] = scoring_params->global_vars[i];
	}

	for(i=0;i<num_all_genes;i++)
	{
		scoring_params->global_medians[i] = scoring_params->global_means[last_processed_gene_set_size];
		scoring_params->global_vars[i] = scoring_params->global_vars[last_processed_gene_set_size];

		scoring_params->ks_params.means_KS[i] = scoring_params->ks_params.means_KS[last_processed_gene_set_size];
		scoring_params->ks_params.vars_KS[j] = scoring_params->ks_params.vars_KS[last_processed_gene_set_size];
	}


*/

// James-Stein Shrinakge (Strimmer paper)
/*	
	double copy_global_vars[num_all_genes];
	for(i=0;i<num_all_genes;i++)
		copy_global_vars[i] = global_vars[i];

	size_t stride = 1;	
	gsl_sort(copy_global_vars, stride, num_all_genes);
	double median_global_vars = gsl_stats_median_from_sorted_data(copy_global_vars, stride, num_all_genes);
//	printf("median_global_vars =%-15.10lf\n", median_global_vars);

	double sum_var = 0, mse = 0;
	for(i=0;i<num_all_genes;i++)
	{
		sum_var += var_of_global_var[i];
		mse += (global_vars[i] - median_global_vars) * (global_vars[i] - median_global_vars);
	}
	double ratio = sum_var / mse;
	double lambda = ratio < 1? ratio : 1;

	for(i=0;i<num_all_genes;i++)
		global_vars[i] = lambda*median_global_vars + (1 - lambda)*global_vars[i];
	printf("sum_var = %-15.10lfmse = %-15.10lfratio = %-15.10lflambda = %-15.10lf\n", sum_var, mse, ratio, lambda);

//	printf("done\n");
//	print_time(t, stdout);
}
*/
