/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef SET_SCORING_H_
#define SET_SCORING_H_

#include "geneset_scoring_interface.h"
#include "degen_gene_set.h"
#include "parameters.h"
#include "random_gen.h"
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_statistics_int.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_fit.h>


/*
// geneset scoring implementations
//double geneset_random_set_ES                (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_random_set_exact_null_ES    (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_random_set_empirical_null_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_ks_ES                       (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_median_ES                   (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_stable_ES                   (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_adaptive_to_noise_level_ES  (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_bootstrap_ES                (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_stable_bootstrap_ES	   (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
*/

void estimate_global_MEDIAN_and_IQR_random_set       (double *all_gene_level_stat, unsigned long *all_gene_indexes, int num_all_genes, 
							int *random_gene_set, int gene_set_size, int num_draws, double *median, double *IQR);
void estimate_global_MEAN_and_VAR_random_set         (double *all_gene_level_stat, unsigned long *all_gene_indexes, int num_all_genes, 
							int *random_gene_set, int gene_set_size, int num_draws, double *mean, double *var);
void estimate_global_MEAN_and_VAR_KS_enrichment_score(double *all_gene_level_stat, unsigned long *all_gene_indexes, int num_all_genes, 
							int *random_gene_set, int gene_set_size, int num_draws, double *mean, double *var);

void estimate_global_MEAN_and_VAR(double *all_gene_level_stat, unsigned long *gene_indexes, int num_all_genes, double *global_mean, double *global_var);
void estimate_global_MEDIAN_and_IQR(double *all_gene_level_stat, unsigned long *gene_indexes, int num_all_genes, double *global_median, double *global_IQR);

// Kolmogorov-Smirnoff test (unweighted)
//double ks_ES(int *genes, int num_genes, int num_all_genes, int *leading_edge_index, int *leading_edge_num_genes);

void preprocess_thresholds_no_kNN(double *all_gene_level_stat, int num_all_genes, int min_geneset_size, geneset_set_scoring_params *scoring_params);
void preprocess_thresholds_using_kNN(double *all_gene_level_stat, int num_all_genes, gsl_matrix *kNN, geneset_set_scoring_params *scoring_params);

double evaluate_random_set_ES(genes, num_genes, all_gene_level_stat, num_all_genes, global_sample_mean, global_var);
void estimate_global_VAR_exact(double *all_gene_level_stat, int num_all_genes, int geneset_size, double global_mean, double *global_var);
void get_perumtation_stats(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, double global_mean, double global_var, double *mean, double *sd);
bool pass_significance_threshold(double obs_score, double mean_random, double significance_threshold_filter, double sd_random);
void get_permutation_distance_targets_sets(int *genes1, int *genes2, double *all_gene_level_stat, int num_all_genes, double *mean, double *sd);

double eval_weighted_similarity(int *arr1, int *arr2, double *evidence_scores, int num_evidence_scores);

int *get_union(int *arr1, int *arr2, int num, int *num_union);
int count_num_union(int *arr1, int *arr2, int num);
int *get_ovelap(int *arr1, int *arr2, int num, int *num_ovelap);
int count_num_overlap(int *arr1, int *arr2, int num);

void get_permutation_overlap_targets_sets(int *genes1, int *genes2, double *all_gene_level_stat, int num_all_genes, 
					double global_mean, double global_var, double *mean, double *sd);

double geneset_random_set_exact_null_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);



//double random_set_ES_bootstrap_all_genes(d_gs *dg, double *all_gene_level_stat, int num_all_genes);
//void bootstrap_random_set_ES_stable(d_gs *dg, double *all_gene_level_stat, int num_all_genes, double global_location, double global_scale);
#endif
