/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */


#ifndef GENESET_SCORING_INTERFACE_H_
#define GENESET_SCORING_INTERFACE_H_

#include <stdbool.h>

typedef double(*pt2estimate_global_location_and_scale_random_set)(double *all_gene_level_stat, unsigned long *all_gene_indexes, int num_all_genes, 
					int *random_gene_set, int gene_set_size, int num_draws, double *global_location, double *global_scale);
typedef double(*pt2estimate_global_location_and_scale)(double *all_gene_level_stat, unsigned long *gene_indexes, int num_all_genes, double *global_location, double *global_scale);

typedef enum {location, scale} adaptive_size_ratio_type;

typedef struct
{
	int step_size, increment_step_size, decrement_ratio_step_size, max_step_size;
	double ratio_lower_threshold, ratio_upper_threshold;
	adaptive_size_ratio_type ratio_type;
}
adaptive_step_size_control;

typedef struct
{
	// geneset statistics for different size genesets
	double *location_estimates, *scale_estimates, *mean_sum_squares_estimates;

	// gene indexes from 0 to (num_seq - 1)
	int *indexes;

	// subset of region indexes used to store contents of genesets draws at random
	int *random_gene_set;

	// toal number of sequence regions
	int num_seq;

	// number of random draws to generate the location/scale pair for ach geneset size
	int num_draws;

	// global statistics for all sequence regions
	double global_dataset_location, global_dataset_scale;

	double max_gene_set_size_percentage_threshold; // input by the user
	double min_gene_set_size_percentage_threshold; // input by the user
	int max_gene_set_size_absolute_threshold;      // estimated once all genes have been read
	int min_gene_set_size_absolute_threshold;      // input by the user

	// use a bootstrap filter when scoring a motif based on resampling with replacement of the set of target sequence regions
	// a motif to be considered significant if (bootstrap_mean - bootstrap_Std_threshold*bootstrap_sd) > 0
	bool bootstrap_motif_score;
	double bootstrap_Std_threshold;
	int num_bootstrap_draws;

	// gene index where the noise starts
//	int threshold_index;

//	bool estimate_empirical_means_and_vars;
//	bool estimate_empirical_medians_and_IQRs;
	pt2estimate_global_location_and_scale_random_set pre_calculate_location_and_scale;
	pt2estimate_global_location_and_scale            pre_calculate_global_datasets_location_and_scale;

	adaptive_step_size_control step_size_control_params;
}
geneset_set_scoring_params;

// geneset scoring implementations
//double geneset_random_set_ES                (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_random_set_exact_null_ES    (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_random_set_empirical_null_ES(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_ks_ES                       (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_median_ES                   (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_stable_ES                   (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_adaptive_to_noise_level_ES  (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_bootstrap_ES                (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);
double geneset_stable_bootstrap_ES	   (int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);


/*
typedef struct
{
	// geneset statistics for different size genesets
	double *means_KS, *vars_KS;
	
	// subset of region indexes used to store contents of genesets draws at random
	int *random_gene_set;
}
ks_scoring_params;

typedef struct
{
	// geneset statistics for different size genesets
	double *global_means, *global_vars, *global_mean_sum_squares, *global_medians, *global_IQRs;

	// gene indexes from 0 to (num_seq - 1)
	int *indexes;

	// toal number of sequence regions
	int num_seq;
	
	// parameters used in Kolmogorov-Smirnoff geneset scoring
	ks_scoring_params ks_params;

	// global statistics for all sequence regions
	double global_mean, global_median, global_var, global_IQR;

	double max_gene_set_size_percentage_threshold; // input by the user
	double min_gene_set_size_percentage_threshold; // input by the user
	int max_gene_set_size_absolute_threshold;      // estimated once all genes have been read
	int min_gene_set_size_absolute_threshold;      // input by the user

//	bool estimate_empirical_means_and_vars;
//	bool estimate_empirical_medians_and_IQRs;
	pt2estimate_global_location_and_scale_random_set pre_calculate_location_and_scale;
}
geneset_set_scoring_params;
*/
typedef double(*pt2geneset_ES)(int *genes, int num_genes, double *all_gene_level_stat, int num_all_genes, geneset_set_scoring_params *scoring_params);

#endif
