/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */


#include "hypergeometric.h"

/*
	k    # overlap
	n1   # motif 1
	t    # motif 2
	n2   # total number genes - n1
*/

double hypergeometric_p_value(unsigned int k, unsigned int n1, unsigned int n2, unsigned int t)
{
	return gsl_cdf_hypergeometric_Q(k, n1, n2, t);
	//MIN(gsl_cdf_hypergeometric_Q(k, n1, n2, t), gsl_cdf_hypergeometric_P(k, n1, n2, t));
}

void test_hypergeometric(char *num_overlap)
{
	unsigned int num_k = atoi(num_overlap);
	unsigned int n1 = 4000;
	unsigned int n2 = 20000-n1, t = 3000;
	printf("num_k: %i\nn1: %i\nn2: %i\nt: %i\n", num_k, n1, n2, t);
	double res = hypergeometric_p_value(num_k, n1, n2, t);
//	gsl_cdf_hypergeometric_Q(num_k, n1, n2, t);
	printf("res     : %1.2E\n", res);
	printf("log(res): %lf\n", log(res));
	getchar();
}
