/*------------------------------------------------------------------------------------*/
/*                                                                                    */
/*    Copyright (C) 2008 Georges QUENOT  LIG-CNRS                                     */
/*    Version 1.00 Last revision: December 22, 2008                                   */
/*                                                                                    */
/*    This file is part of KNNLSB (K Nearest Neighbor Linear Scan Baseline).          */
/*                                                                                    */
/*    KNNLSB is free software: you can redistribute it and/or modify                  */
/*    it under the terms of the GNU General Public License as published by            */
/*    the Free Software Foundation, either version 3 of the License, or               */
/*    (at your option) any later version.                                             */
/*                                                                                    */
/*    KNNLSB is distributed in the hope that it will be useful,                       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of                  */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   */
/*    GNU General Public License for more details.                                    */
/*                                                                                    */
/*    You should have received a copy of the GNU General Public License               */
/*    along with KNNLSB.  If not, see <http://www.gnu.org/licenses/>.                 */
/*                                                                                    */
/*------------------------------------------------------------------------------------*/

#include "my_knn.h"

// especially efficicent if na >> nk, nb >> 1 and nc >> 1.
// asymptotic execution time is in O(na*nb*nc/np)

// na :  number of "document" vectors;
// nb :  number of "query" vectors;
// nc :  number of vector components;
// nk :  number of searched nearest neighbors;

#define NC 10
#define NA 11
#define NK 5
#define NB 11
//#define NP 1
/*
#define NC 100
#define NA 100000
#define NK 10
#define NB 1000
*/
/*------------------------------------------------------------------------------------*/

#define FAILURE 0
#define SUCCESS 1

int argindex(int argc, char *argv[], char *name, int n)
{
  int i;
  for (i = 1; i < argc-n; i++) if (!strcmp(name,argv[i])) return(i);
  return(0);
}

void argremove(int *pargc, char *argv[], int i, int n)
{
  int j;
  *pargc -= n+1;
  for (j = i; j < *pargc; j++) argv[j] = argv[j+n+1];
}

int getbarg(int *pargc, char *argv[], char *name)
{
  int i;
  if ((i = argindex(*pargc,argv,name,0))) argremove(pargc,argv,i,0);
  return(i != 0);
}

int getiarg(int *pargc, char *argv[], char *name, int idef, int *piarg)
{
  int i;
  *piarg = idef;
  if ((i = argindex(*pargc,argv,name,1))) {
    if (sscanf(argv[i+1],"%d",piarg) != 1) {
      fprintf(stderr,"ERROR: %s option expects a valid integer\n",name);
      return(FAILURE);
    }
    argremove(pargc,argv,i,1);
  }
  return(SUCCESS);
}

size_t getzuarg(int *pargc, char *argv[], char *name, size_t udef, size_t *puarg)
{
  int i;
  *puarg = udef;
  if ((i = argindex(*pargc,argv,name,1))) {
    if (sscanf(argv[i+1],"%zu",puarg) != 1) {
      fprintf(stderr,"ERROR: %s option expects a valid integer\n",name);
      return(FAILURE);
    }
    argremove(pargc,argv,i,1);
  }
  return(SUCCESS);
}

/*------------------------------------------------------------------------------------*/

fdata **genrand(int n, int nc)
{
  size_t i,ic;
  fdata **m;
  if ((m = malloc2a(n,nc,sizeof(fdata))) == NULL) return(NULL);
  for (i = 0; i < n; i++) for (ic = 0; ic < nc; ic++)
    m[i][ic] = (rand()&0xFFFFFF)/((fdata) 0xFFFFFF);
  return(m);
}

/*------------------------------------------------------------------------------------*/

void test_kNN(int argc, char *argv[])
{
	int ib, ik;
	size_t na, nb, nc, nk;
	int dc;

	if (getzuarg(&argc,argv,"-NA",NA,&na) == FAILURE) exit(1);
	if (getzuarg(&argc,argv,"-NB",NB,&nb) == FAILURE) exit(1);
	if (getzuarg(&argc,argv,"-NC",NC,&nc) == FAILURE) exit(1);
	if (getzuarg(&argc,argv,"-NK",NK,&nk) == FAILURE) exit(1);
	dc = getbarg(&argc,argv,"-CHI2") ? 2 : getbarg(&argc,argv,"-COS");

	fdata **ma,**mb;
	if ((ma = genrand(na,nc)) == NULL) exit(1);
	if ((mb = genrand(nb,nc)) == NULL) exit(1);

//	printf("%zd-core, cache size: %zd KB\n\n",get_nb_cores(),get_cache_size()/1024);
 
	output_matrix(ma, na, nc, 10, 10);
//	output_matrix(mb, nb, nc, 10, 10);

	KEY **key0 = find_kNN(ma, ma, na, nb, nc, nk, dc);
	
	printf("features = %i\nsamples = %i\nk = %i\n", nc, na, nk);
	output_kNN_result(key0, nb, nk, 10, 10);

	free2a(ma,na,nc,sizeof(fdata));
	free2a(mb,nb,nc,sizeof(fdata));
	free2a(key0,nb,nk,sizeof(KEY));

//	printf("Maximum allocated memory: %12zu bytes\n",MaxUsedMemory());
//	printf("Still allocated memory:   %12zu bytes\n",UsedMemory());
//	printf("\nDone\n");

}

/*------------------------------------------------------------------------------------*/
