/*------------------------------------------------------------------------------------*/
/*                                                                                    */
/*    Copyright (C) 2008 Georges QUENOT  LIG-CNRS                                     */
/*    Version 1.00 Last revision: December 22, 2008                                   */
/*                                                                                    */
/*    This file is part of KNNLSB (K Nearest Neighbor Linear Scan Baseline).          */
/*                                                                                    */
/*    KNNLSB is free software: you can redistribute it and/or modify                  */
/*    it under the terms of the GNU General Public License as published by            */
/*    the Free Software Foundation, either version 3 of the License, or               */
/*    (at your option) any later version.                                             */
/*                                                                                    */
/*    KNNLSB is distributed in the hope that it will be useful,                       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of                  */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   */
/*    GNU General Public License for more details.                                    */
/*                                                                                    */
/*    You should have received a copy of the GNU General Public License               */
/*    along with KNNLSB.  If not, see <http://www.gnu.org/licenses/>.                 */
/*                                                                                    */
/*------------------------------------------------------------------------------------*/

//#include "knnp.h"
#include "my_knn.h"

// especially efficient if na >> nk, nb >> 1 and nc >> 1.
// asymptotic execution time is in O(na*nb*nc/np)

/*
ma  document vectors, two-dimensional (na*nk) array of fdata;
mb  query vectors, two-dimensional (nb*nk) array of fdata;
na  number of "document" vectors;
nb  number of "query" vectors;
nc  number of vector components;
nk  number of searched nearest neighbors;
dc  2 : use chi square distance; 
    1 : use angle between vectors;
    0 : use Euclidian distance.

return: nb-by-nk array of type KEY
        typedef struct {
          int k;
          fdata d;
        } KEY;
*/
KEY **find_kNN(fdata **ma, fdata**mb, size_t na, size_t nb, size_t nc, size_t nk, int dc)
{
	// print some statistcal information : knnp execution time, number of floating point 
	// operations per second, main memory bandwidth, cache bandwidth (per core).
	int sta = 0;

	// number of threads to be run in parallel or 0 to let the program set it to the number of available cores if get_nb_cores() works.
	size_t np = 0;

	// specify the cache size on the system or 0 to let the program get it from the system if get_cache_size() works.
	size_t cs = 0; 
	
	fdata *sa = getnorm(ma,na,nc,dc);
	KEY **result;
  	if((result = knnp(ma,mb,sa,na,nb,nc,nk,np,cs,dc,sta)) == NULL) 
	{
		printf("ERROR: 'find_kNN'!\n");
		exit(1);
	}
	free(sa);
	return result;
}



void output_matrix(fdata **ma, size_t na, size_t nc, int max_num_rows, int max_num_cols)
{
	int ib, ik;
	printf("%i-by-%i\n", na, nc);
	for (ib = 0; ib < na && (ib < max_num_rows); ib++) 
	{
		for (ik = 0; ik < nc && (ik < max_num_cols); ik++)
			printf("%-7.2lf",ma[ib][ik]);
		printf("\n");
	}
}

// na  number of "document" vectors;
// nb  number of "query" vectors;
// nc  number of vector components;
void output_kNN_result(KEY **key0, size_t nb, size_t nc, int max_num_rows, int max_num_cols)
{
	int ib, ik;
	printf("%i-by-%i\n", nb, nc);
	for (ib = 0; ib < nb && (ib < max_num_rows); ib++) 
	{
		for (ik = 0; ik < nc && (ik < max_num_cols); ik++)
			printf("%-5d%5.2f\n",key0[ib][ik].k, key0[ib][ik].d);
		printf("\n");
	}
}

/*------------------------------------------------------------------------------------*/
