/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

//driver module, initialize parameter values and run the motif finding

#include "params_reader.h"
#include "parameters.h"
#include "analysis.h"
#include "IUPACalphabet.h"

#include "preprocess.h"
#include "regression.h"

//extern bool consider_rev_compl_flag;
//extern double significance_threshold_filter_global;

int main(int argc, char*argv[])
{
	params p; IO_params p_IO;
	p.fast_mode = false;
	if(strcmp(argv[30], "yes") == 0)
		p.fast_mode = true;

	p.significance_threshold_filter = 3.0;
//	significance_threshold_filter_global = p.significance_threshold_filter;

	// regression params
	char *use_regression = argv[28];
	if(strcmp(use_regression, "yes") == 0)
		p_IO.cov_file = argv[29];
	else if(strcmp(use_regression, "no") == 0)
		p_IO.cov_file = 0;
	else
	{
		printf("INPUT ERROR: 'use_regression' must be set to 'yes' or 'no' value!\n");
		exit(1);
	}

	// bootstrap params
	char *bootstrap_motif_score = argv[26];
	if(strcmp(bootstrap_motif_score, "yes") == 0)
	{
		p.scoring_params.bootstrap_motif_score = true;
		p.scoring_params.bootstrap_Std_threshold = atof(argv[25]);
		p.scoring_params.num_bootstrap_draws = atoi(argv[27]);
		if(p.scoring_params.num_bootstrap_draws <= 0)
			p.scoring_params.num_bootstrap_draws = 20; // default
	}
	else
	{
		p.scoring_params.bootstrap_motif_score = false;
		p.scoring_params.bootstrap_Std_threshold = atof(argv[25]);
	}

	// some file names
	char out_dir[500];
	char p_val_file[500]; 
	char random_scores_file[500];
	if(argv[3][strlen(argv[3]) - 1] == '/')
		sprintf(out_dir, "%s", argv[3]);
	else
		sprintf(out_dir, "%s/", argv[3]);
	sprintf(p_val_file, "%s%s", out_dir, argv[23]);

	sprintf(random_scores_file, "%s%s", out_dir, argv[21]);
//	MALLOC(p_IO.null_scores_filename, sizeof(char)*(strlen(random_scores_file))+1);
	MALLOC(p_IO.null_scores_filename, sizeof(char)*1000);
	sprintf(p_IO.null_scores_filename, "%s", random_scores_file);

//	bool parameters_log_created = false;

	// significance params
	p.significance.num_random_permutations = atoi(argv[22]);
	p.significance.num_random_permutations_remaining = p.significance.num_random_permutations;
	if(p.significance.num_random_permutations > 0)
	{
		printf("evaluating significance: %i random draws\n", p.significance.num_random_permutations);
		p.significance.evaluate_significance = true;
	}
	else
		p.significance.evaluate_significance = false;


	//read and initialize parameter values, implementation inside 'params_reader.c'
	read_input(argc,argv,&p,&p_IO);
	//run the motif finding algorithm, implementation inside 'analysis.c'
	run_analysis(&p,&p_IO);
	//free memory
	free(p_IO.null_scores_filename);
	free_oligos(&p.oligo_seeds);

	return 0;
}
