/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "matrix_multiply.h"

// compute the matrix-matrix product and sum C = \alpha op(A) op(B) + \beta C	
// where op(A) = A, A^T, A^H for TransA = CblasNoTrans, CblasTrans
void mat_mat_multiply(gsl_matrix *A, gsl_matrix *B, gsl_matrix *C, CBLAS_TRANSPOSE_t TransA, CBLAS_TRANSPOSE_t TransB)
{
	double alpha = 1, beta = 0;
	int res = gsl_blas_dgemm(TransA, TransB, alpha, A, B, beta, C);
	if(res != 0)
		printf("ERROR: 'mat_mat_multiply', 'gsl_blas_dgemm', %i\n", res);
}


// int gsl_blas_dsymm(CBLAS_SIDE_t Side, CBLAS_UPLO_t Uplo, double alpha, const gsl_matrix * A, const gsl_matrix * B, double beta, gsl_matrix * C)
void mat_mat_multiply_sym(gsl_matrix *A, gsl_matrix *B, gsl_matrix *C, CBLAS_SIDE_t Side)
{
	double alpha = 1, beta = 0;
	// CblasUpper -> the upper triangle and diagonal of A are used
	// CblasLower -> the lower triangle and diagonal of A are used
	// C = A * B
	int res = gsl_blas_dsymm(Side, CblasUpper, alpha, A, B, beta, C);
	if(res != 0)
		printf("ERROR: 'gsl_blas_dgemm1', %i\n", res);
}

