/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

// main motif discovery module

#include "motif_finder.h"

typedef enum {positive, negative} motif_score_sign;

//extern double top_scores[200];
//extern int num_top_scores;
extern double *top_random_scores;
extern int random_index;
extern bool consider_rev_compl_flag;
extern const int no_primary_IUPAC_nucleotides;

int get_seq_length(char *text_filename, char delim)
{
//	printf("text_filename: %s\n", text_filename);
	FILE *fp;
	if( (fp = fopen(text_filename, "r")) == NULL )
	{
   		printf("'get_seq_length': could not open file: %s\n", text_filename);
    		exit(2);
	}
	fclose(fp);

  	fp = fopen(text_filename, "r");
  	char *text = 0;
  	size_t text_length = 0;
 // 	getdelim(&text, &text_length, '$', fp);
  	// takes an 'int' as a delim value
 	getdelim(&text, &text_length, delim, fp);

  	fclose(fp);
   	int length = 2 * strlen(text);
  	FREE(text);
  	return length;
}

// extracts 1) the sequences length (including the reverse complement) and stores it in 'seq_length'
//          2) number of sequences and stores it in 'num_seq'
void get_seq_info(char *text_filename, char *delim, int *num_seq, int *seq_length, bool ignore_strand)
{
	FILE *fp;
	if( (fp = fopen(text_filename, "r")) == NULL )
	{
   		printf("'get_seq_info': could not open file: %s\n", text_filename);
    		exit(2);
	}
	fclose(fp);
	fp = fopen(text_filename, "r");
	char *line = NULL;
	size_t len = 0; ssize_t read;
	int total_read = 0;
	int count = 0; int length_one_strand_only;
	while ((read = getdelim(&line, &len, (char)delim, fp)) != -1)
	{
		length_one_strand_only = read;
		count++;
		FREE(line);
	}
	FREE(line);
  	fclose(fp);
	*num_seq = count; //no double stranded /2;
  	*seq_length = length_one_strand_only; //no double stranded * 2; // double in order to account for the reverse complement
//	printf("delim = %i\n*num_seq = %i\n*seq_length = %i\n", delim, *num_seq , *seq_length);exit(0);
}

void read_sequence_input(char *text_filename, seq_data *seq, char *delim)
{
	FILE *fp;
	if( (fp = fopen(text_filename, "r")) == NULL )
	{
   		printf("'get_seq_info': could not open file: %s\n", text_filename);
    		exit(2);
	}
	fclose(fp);

	seq->delim = delim;
	get_seq_info(text_filename, seq->delim, &seq->num_seq, &seq->seq_length, true);//seq->ignore_strand);

	fp = fopen(text_filename, "r");
 	char *text = NULL;
  	size_t text_length = 0;
   	getdelim(&text, &text_length, '\n' , fp);
	fclose(fp);

	char *build_options = "copy_text;free_text";
	seq->index = malloc(sizeof(TSA_Un_copy *));
	int e;
  	// 'text_length' DOESN'T contain the correct value (?why?), so use 'strlen(text)' instead
	if((e = build_index(text, strlen(text), build_options, seq->index)) != 0)
	{
	  	printf("could not build index: %s\n", error_index(e));
	   	exit(1);
	}
	seq->is_initialized = true;
}

#include "degen_gene_set.h"

//#include <stdbool.h>

bool only_plot = false;

//extern int *reference_positions;
//extern int offset_top_score_global;

// require co-occurrence upstream of a gene (irrespective of offset from TSS)
int count_ovelapping_genes(d_gs *motif1, d_gs *motif2, void **index, int num_seq)
{
	int i;
	int *genes1 = motif1->genes;//extract_genes(motif1_instances_tree->locations_tree, index, seq_length_global, motif1->num_genes);
	int *genes2 = motif2->genes;//extract_genes(motif2_instances_tree->locations_tree, index, seq_length_global, motif2->num_genes);

	int overlap_count = 0, ind1 = 0;
//	printf("motif2->num_genes: %i\tmotif1->num_genes: %i\n", motif2->num_genes, motif1->num_genes);
	for(i=0;i<motif2->num_genes;i++)
	{
	//	printf("%i: %i\t%i\t%i\t%i\n", i, genes1[i], motif1->genes[i], genes2[i], motif2->genes[i]);

		while(ind1 < motif1->num_genes && genes1[ind1] < genes2[i])
			ind1++;
	//	printf("%i: %i\t%i\t%i\t%i\n", i, genes1[ind1], motif1->genes[ind1], genes2[i], motif2->genes[i]);
		if(genes1[ind1] == genes2[i])
		{
		//	printf("ind: %i\ti:%i\toverlap_count:%i\n", ind1, i, overlap_count);
			overlap_count++;
		}
	}
//	printf("%s{%i}\t%s{%i}\t%i\n", motif1->degen_oligo, motif1->num_genes,
//								   motif2->degen_oligo, motif2->num_genes, overlap_count);
//	getchar();
//	destroy_motif_instances(&motif1_instances_tree);
//	destroy_motif_instances(&motif2_instances_tree);
	return overlap_count;
}

int left_offset = -1;
int right_offset = 1;
bool within_offset_range(int ind1, int ind2)
{
	return ind1 >= ind2 - left_offset && ind1 <= ind2 - right_offset;
}



// require co-occurrence upstream of a gene at the SAME offset from TSS
void count_ovelapping_gene_offsets_new(d_gs *motif1, d_gs *motif2, void **index, int num_seq, int seq_length,
									int *num_offset_specific_gene_overlaps, int *total_num_gene_overlaps)
{
//	printf("%s(%i)\n%s(%i)\n", motif1->degen_oligo, motif1->num_genes, motif2->degen_oligo, motif2->num_genes);
	int i;
	int offset_threshold = (motif1->oligo_length < motif2->oligo_length ? motif1->oligo_length / 2 : motif2->oligo_length / 2);

	unsigned char **oligos_motif1 = motif1->instantiations;
	int num_oligos_motif1 = motif1->instantiations_length;
	degen_oligo_occurrences dg_motif1_occ;
	degen_oligo_occurrences_init(&dg_motif1_occ, num_oligos_motif1, num_seq);
	find_degen_oligo_set_occurrences_ALL(index, oligos_motif1, num_oligos_motif1, seq_length, num_seq, &dg_motif1_occ);

	unsigned char **oligos_motif2 = motif2->instantiations;
	int num_oligos_motif2 = motif2->instantiations_length;
	degen_oligo_occurrences dg_motif2_occ;
	degen_oligo_occurrences_init(&dg_motif2_occ, num_oligos_motif2, num_seq);
	find_degen_oligo_set_occurrences_ALL(index, oligos_motif2, num_oligos_motif2, seq_length, num_seq, &dg_motif2_occ);

	(*total_num_gene_overlaps) = 0;
	(*num_offset_specific_gene_overlaps) = 0;
	for(i=0;i<num_seq;i++)
	{
		int offset_incr = 0, gene_index_incr = 0;
		if(dg_motif1_occ.match_info[i].num_occ_in_region > 0 && dg_motif2_occ.match_info[i].num_occ_in_region > 0)
		{
			gene_index_incr = 1;
			if(fabs(dg_motif1_occ.match_info[i].index_start_offset - dg_motif2_occ.match_info[i].index_start_offset) <= offset_threshold)
				offset_incr = 1;
		}
		(*total_num_gene_overlaps) += gene_index_incr;
		(*num_offset_specific_gene_overlaps) += offset_incr;
	}
	(*num_offset_specific_gene_overlaps) = (*total_num_gene_overlaps);
	degen_oligo_occurrences_free(&dg_motif2_occ);
	degen_oligo_occurrences_free(&dg_motif1_occ);
}


// require co-occurrence upstream of a gene at the SAME offset from TSS
void count_ovelapping_gene_offsets_batch(d_gs *reference_motif, d_gs **motifs, int *motif_indexes, int num_indexes, void **index, int num_seq, int seq_length,
									int *num_offset_specific_gene_overlaps, int *total_num_gene_overlaps)
{
	int i;
	for(i=0;i<num_indexes;i++)
	{
		count_ovelapping_gene_offsets_new(reference_motif, motifs[motif_indexes[i]], index, num_seq, seq_length,
									&num_offset_specific_gene_overlaps[i], &total_num_gene_overlaps[i]);
	}
}

/*
// count co-occurences between any oligo from 2 sets of motifs
void count_ovelapping_gene_clusters(d_gs **motifs1, d_gs **motifs2, int *motif_indexes1, int *motif_indexes2, int num_indexes1, int num_indexes2, void **index,
									int num_seq, int seq_length, int total_num_gene_overlaps)
{
	unsigned char **oligos1;
	int num_oligos1;
	int i;
	for(i=0;i<num_indexes1;i++)
	{
		oligos1 = motifs1->instantiations;
		num_oligos1 += motifs1->instantiations_length;
	}
	degen_oligo_occurrences dg_motifs1_occ;
	degen_oligo_occurrences_init(&dg_motifs1_occ, num_oligos1, num_seq);
	find_degen_oligo_set_occurrences_ALL(index, oligos1, num_oligos1, seq_length, num_seq, &dg_motifs1_occ);

}*/

double hypergeometric_test_criteria(d_gs *motif1, d_gs *motif2, void **index, int num_seq, int seq_length, int num_offset_specific_gene_overlaps)
{
	int k        = num_offset_specific_gene_overlaps;
	int n1       = motif1->num_genes;
	int n2       = num_seq - n1;
	int t        = motif2->num_genes;
	double p_val = hypergeometric_p_value(k,n1,n2,t);
	return p_val;
}

double get_similarity(d_gs *motif1, d_gs *motif2, bool *rev_compl, int required_core_length, int *offset_top_score)//, cluster_sim_criteria sim_criterion)
{
	double offset1, offset2;
	double sim1 = eval_PSSM_similarity(motif1->pssm, motif2->pssm, required_core_length, offset_top_score);
	//double sim1 = distance_metric_DRIM(motif1->degen_oligo, motif2->degen_oligo);

	offset1 = *offset_top_score;

	if(consider_rev_compl_flag == false)
	{
		*offset_top_score = offset1;
		*rev_compl = false;
		return sim1;
	}

	gsl_matrix *rev_compl_pssm = get_rev_compl_pssm(motif1->pssm);
	double sim2 = eval_PSSM_similarity(rev_compl_pssm, motif2->pssm, required_core_length, offset_top_score);
	//double sim2 = distance_metric_DRIM(get_rev_compl(motif1->degen_oligo), motif2->degen_oligo);
	offset2 = *offset_top_score;
	gsl_matrix_free(rev_compl_pssm);

	double sim_score = MAX(sim1, sim2);
	if(sim_score == sim1)
	{
		*offset_top_score = offset1;
		*rev_compl = false;
	}
	else
	{
		*offset_top_score = offset2;
		*rev_compl = true;
	}
	return sim_score;
}

bool not_yet_assigned_to_a_cluster(int *c_index, int genset_index){ return c_index[genset_index] == -1;}

void init_cluster_index(int *c_index, int num)
{
	int i;
	for(i=0;i<num;i++)
		c_index[i] = -1;
}

void init_cluster_order(int *c_order_index, int num)
{
	int i;
	for(i=0;i<num;i++)
		c_order_index[i] = i;
}

// if rev_compl_flags == true -> optimal similarity is achieved with reverse complement of one of the motifs
bool *rev_compl_flags;
// offset of shorter w.r.t longer motif, producing maximum similarity score
int *optimal_offsets;
// significance estimate for co-localization of two motifs, based on overlap of genes where present
double *p_val = 0;



// c_index -> cluster index, upon function completion initialized with the cluster labels for individual motifs in 'd_gs **arr'
// new procedure of clustering evolved motifs:
// Start with the sorted list of evolved motifs
// For the top motif, select all motifs with similarity > threshold (0.75) and compute score of the intersection -> retain all for which this score is > score of top motif
// Repeat until no more motifs left
void cluster_output_by_joint_score(d_gs **arr, int num, int *c_index, void **index, int num_seq, int seq_length, params *p)
{
	int i, j, ii, k;
	// index denotes geneset/motif number in the score-sorted list, the value indicates the cluster membership
	double similarity_cluster_entries[num];
	for(i=0;i<num;i++)
		similarity_cluster_entries[i] = -1;
	int *explained_peaks_current = malloc(sizeof(int) * p->d.num_seq_level_stats);
	for(i=0;i<num;i++)
	{
		p_val[i] = -1;
		rev_compl_flags[i] = false;
		optimal_offsets[i] = 0;
	}

	int motif_indexes[num];
	int num_offset_specific_gene_overlaps[num], total_num_gene_overlaps[num];

	int *intersect_genes, *union_genes;
	int num_intersect_genes, num_union_genes;
	double score_intersect, score_union;
	int sign_multiplier = -1; // TODO
	double score_fraction = 0.95; // amount of allowed variability for best score

	//printf("M1 idx\tM1 oligo\tM1 score\tM1 targets\tM2 idx\tM2 oligo\tM2 score\tM2 targets\tM1-M2 sim\tM1&M2 score\tM1&M2 targets\tM1|M2 score\tM1|M2 targets\n");
	int next = 0;
	for(i=0;i<num;i++)
	{
		if(not_yet_assigned_to_a_cluster(c_index, i) == true)
		{
			c_index[i] = next;

			while(i+1 < num && strcmp(arr[i]->degen_oligo, arr[i+1]->degen_oligo) == 0)
			{
				c_index[i+1] = next;
				i++;
			}
			next++;

			// initialize explained peaks count for current cluster
			for(ii=0;ii<p->d.num_seq_level_stats;ii++)
				explained_peaks_current[ii] = 0;
			for(ii=0;ii<arr[i]->num_genes;ii++)
				explained_peaks_current[arr[i]->genes[ii]] += 1;
			int num_peaks_curr = arr[i]->num_genes;
			double prev_fraction_cum_curr = (1.0 * arr[i]->num_genes) / p->d.num_seq_level_stats;

			int top_scoring_cluster_num_targets = arr[i]->num_genes;

			// 'good' scores are LARGE and POSITIVE
			double top_cluster_score = -arr[i]->score;

//			printf("=cluster_output= cluster %i, motif %s, score %f\n",i, arr[i]->degen_oligo,top_cluster_score);

			for(j=i+1;j<num;j++)
			{
				bool rev_compl;
				int offset_top_score;
				double sim = get_similarity(arr[j], arr[i], &rev_compl, p->required_core_length, &offset_top_score);
				bool similar_flag = (sim >= p->cluster_sim_threshold);
				if(
				   (not_yet_assigned_to_a_cluster(c_index, j) == true)
				   && (similar_flag == true)
				  )
				{ // motif j is similar with motif i
					int num_new_peaks = 0;
					for(ii=0;ii<arr[j]->num_genes;ii++)
					{
						if(explained_peaks_current[arr[j]->genes[ii]] == 0)
							num_new_peaks += 1;
						explained_peaks_current[arr[j]->genes[ii]] += 1;
					}
					double fraction_cum_curr = (1.0 * (num_new_peaks + num_peaks_curr)) / p->d.num_seq_level_stats;
					num_peaks_curr += num_new_peaks;
					prev_fraction_cum_curr = fraction_cum_curr;
					// compute score for motifs i and j (intersection)
					intersect_genes = get_intersect_genes(arr[i]->genes, arr[i]->num_genes, arr[j]->genes, arr[j]->num_genes, p->d.num_seq_level_stats, &num_intersect_genes);
					score_intersect = sign_multiplier * p->geneset_ES(intersect_genes, num_intersect_genes, p->d.seq_level_stats, p->d.num_seq_level_stats, &p->scoring_params);
					// compute score for motifs i or j (union)
					union_genes = get_union_genes(arr[i]->genes, arr[i]->num_genes, arr[j]->genes, arr[j]->num_genes, p->d.num_seq_level_stats, &num_union_genes);
					score_union = sign_multiplier * p->geneset_ES(union_genes, num_union_genes, p->d.seq_level_stats, p->d.num_seq_level_stats, &p->scoring_params);
					/*// print some info
					if(strcmp(arr[j]->degen_oligo, arr[j-1]->degen_oligo) != 0) // test motif j not to be the same as the one before (printing unique j)
						printf("%i\t%s\t%4.2f\t%i\t%i\t%s\t%4.2f\t%i\t%4.2f\t%4.2f\t%i\t%4.2f\t%i\n", i, arr[i]->degen_oligo, sign_multiplier *arr[i]->score, arr[i]->num_genes, j, arr[j]->degen_oligo, sign_multiplier *arr[j]->score, arr[j]->num_genes, sim, score_intersect, num_intersect_genes, score_union, num_union_genes);*/
					// test if scores are good for clustering
					if((score_intersect>score_fraction*sign_multiplier *arr[i]->score)||(score_union>score_fraction*sign_multiplier *arr[i]->score))
					{
						c_index[j] = c_index[i]; // motif j in the same cluster as motif i
						similarity_cluster_entries[j] = sim; // similarity
						optimal_offsets[j] = offset_top_score; // motif offset
					}
				}
			}
		}
	}
	free(explained_peaks_current);
}

// c_index -> cluster index, upon function completion initialized with the cluster labels for individual motifs in 'd_gs **arr'
void cluster_output(d_gs **arr, int num, int *c_index, void **index, int num_seq, int seq_length, params *p)
{
	int i, j, ii;
	// index denotes geneset/motif number in the score-sorted list, the value indicates the cluster membership
	double similarity_cluster_entries[num];
	for(i=0;i<num;i++)
		similarity_cluster_entries[i] = -1;
	int *explained_peaks_current = malloc(sizeof(int) * p->d.num_seq_level_stats);
	for(i=0;i<num;i++)
	{
		p_val[i] = -1;
		rev_compl_flags[i] = false;
		optimal_offsets[i] = 0;
	}

	int motif_indexes[num];
	int num_offset_specific_gene_overlaps[num], total_num_gene_overlaps[num];

	int next = 0;
	for(i=0;i<num;i++)
	{
		if(not_yet_assigned_to_a_cluster(c_index, i) == true)
		{
			c_index[i] = next;

			while(i+1 < num && strcmp(arr[i]->degen_oligo, arr[i+1]->degen_oligo) == 0) 
			{
				c_index[i+1] = next;
				i++;
			}
			next++;
			
			// intialize explained peaks count for current cluster
			for(ii=0;ii<p->d.num_seq_level_stats;ii++)
				explained_peaks_current[ii] = 0;
			for(ii=0;ii<arr[i]->num_genes;ii++)
				explained_peaks_current[arr[i]->genes[ii]] += 1;
			int num_peaks_curr = arr[i]->num_genes;
			double prev_fraction_cum_curr = (1.0 * arr[i]->num_genes) / p->d.num_seq_level_stats;

			int top_scoring_cluster_num_targets = arr[i]->num_genes;

			// 'good' scores are LARGE and POSITIVE
			double top_cluster_score = -arr[i]->score;

//			printf("=cluster_output= cluster %i, motif %s, score %f\n",i, arr[i]->degen_oligo,top_cluster_score);

			int offset_top_score;
			int num_indexes = 0;
			for(j=i+1;j<num;j++)
			{
				bool rev_compl;
				double sim = get_similarity(arr[j], arr[i], &rev_compl, p->required_core_length, &offset_top_score);
				bool similar_flag = (sim >= p->cluster_sim_threshold);
				if(
				   (not_yet_assigned_to_a_cluster(c_index, j) == true)
				   && (similar_flag == true)
				  )
				{
					int num_new_peaks = 0;
					for(ii=0;ii<arr[j]->num_genes;ii++)
					{
						if(explained_peaks_current[arr[j]->genes[ii]] == 0)
							num_new_peaks += 1;
						explained_peaks_current[arr[j]->genes[ii]] += 1;
					}
					double fraction_cum_curr = (1.0 * (num_new_peaks + num_peaks_curr)) / p->d.num_seq_level_stats;
					num_peaks_curr += num_new_peaks;
					motif_indexes[num_indexes] = j;
					num_indexes++;
					prev_fraction_cum_curr = fraction_cum_curr;
				}
			}
			count_ovelapping_gene_offsets_batch(arr[i], arr, motif_indexes, num_indexes, index, num_seq, seq_length,
									num_offset_specific_gene_overlaps, total_num_gene_overlaps);
			for(j=0;j<num_indexes;j++)
			{
				int motif_ind = motif_indexes[j];
				int offset_top_score;
				double p_val_tmp = hypergeometric_test_criteria(arr[i], arr[motif_ind], index, num_seq, seq_length, num_offset_specific_gene_overlaps[j]);
				if(p_val_tmp < p->hypegeom_p_val_cutoff)
				{
					p_val[motif_ind] = p_val_tmp;
					c_index[motif_ind] = c_index[i];

					similarity_cluster_entries[motif_ind] = get_similarity(arr[motif_ind], arr[i], &rev_compl_flags[motif_ind], p->required_core_length, &offset_top_score);
					optimal_offsets[motif_ind] = offset_top_score;//offset_top_score_global;
					//printf("=cluster_output= -----------, similar motif %s, score %f\n",arr[motif_ind]->degen_oligo,-arr[motif_ind]->score);
				}
			}
		}
	}
	free(explained_peaks_current);
}

cluster *get_cluster_member_indexes(int target_cluster_index, clustered_motifs *c_motifs)
{
	int i;
	cluster *c;
	MALLOC(c, sizeof(cluster));
	c->num = 0;
	for(i=0;i<c_motifs->num_unique_motifs;i++)
		if(c_motifs->c_index[c_motifs->indexes_of_unique_motifs[i]] == target_cluster_index)
			c->num++;
	MALLOC(c->cluster_indexes, sizeof(int) * c->num);
	int index = 0;
	for(i=0;i<c_motifs->num_unique_motifs;i++)
	{
		if(c_motifs->c_index[c_motifs->indexes_of_unique_motifs[i]] == target_cluster_index)
		{
			c->cluster_indexes[index] = c_motifs->indexes_of_unique_motifs[i];
			index++;
		}
	}
	return c;
}

void extract_cluster_targets_array(int target_cluster_index, clustered_motifs *c_motifs, int *explained_peaks_current, int num_explained_peaks_current)
{
	int i, j;
	for(i=0;i<num_explained_peaks_current;i++)
		explained_peaks_current[i] = 0;
	
	cluster *c = get_cluster_member_indexes(target_cluster_index, c_motifs);

	int num_new_peaks = 0;
	for(i=0;i<c->num;i++)
	{
		d_gs *motif = c_motifs->motifs[c->cluster_indexes[i]];
		for(j=0;j<motif->num_genes;j++)
		{
			explained_peaks_current[motif->genes[j]] += 1;
		}
	}
	free(c->cluster_indexes);
	free(c);
}


#include "geneset_scoring.h"

// after this cm->cluster_scores[i] will contain the scores for union of all motifs in a cluster
void score_clusters(clustered_motifs *cm, motif_score_sign sign, params *p)
{
	int num_explained_peaks = p->d.num_seq_level_stats;
	int *explained_peaks = malloc(sizeof(int) * num_explained_peaks);
	int sign_multiplier = (sign == positive ? 1:-1);
	int i, num_overlap_genes;
	int *overlap_genes;
	double score;

/*	for(i=0;i<p->d.num_seq_level_stats;i++)
		printf("p->d.seq_level_stats %i - %8.4f\n", i, p->d.seq_level_stats[i]);*/

	for(i=0;i<cm->num_clusters;i++)
	{
		extract_cluster_targets_array(i, cm, explained_peaks, num_explained_peaks);
		overlap_genes = get_ovelap(explained_peaks, explained_peaks, num_explained_peaks, &num_overlap_genes);
		score = sign_multiplier * p->geneset_ES(overlap_genes, num_overlap_genes, p->d.seq_level_stats, p->d.num_seq_level_stats, &p->scoring_params);
		if(score<0.01)
		{
			printf("-----> Score %e, cluster %i, overlapping genes %i\n", score, i+1, num_overlap_genes);
		}
		cm->cluster_scores[i] = score;
		free(overlap_genes);
	}
}

// this sorts all clusters by the scores in cm->cluster_scores[i] - renumbers them
void sort_clusters_by_score(clustered_motifs *cm)
{
	int i;
	sort_order so = decreasing;
	int *idx;
	// sort the scores and retrieve the new indices
	idx = sort_double_idx(cm->cluster_scores, cm->num_clusters, so);
	// renumber the cluster index of all motifs
	for(i=0;i<cm->total_num_motifs;i++)
	{
		cm->c_index[i] = idx[cm->c_index[i]];
	}
	// reorder the scores
	double *tmp_scores;
	MALLOC(tmp_scores, sizeof(double) * cm->num_clusters);
	for(i=0;i<cm->num_clusters; i++)
		tmp_scores[i] = cm->cluster_scores[idx[i]];
	cm->cluster_scores = tmp_scores;
}

// verifies if clusters have empty_cluster_score, and if so, delete the empty cluster
// (modifies cm->cluster_scores & cm->num_clusters, renumbers cm->total_num_motifs)
void eliminate_empty_clusters(clustered_motifs *cm, double empty_cluster_score)
{
	int i, j;
	for(i=0;i<cm->num_clusters;i++)
	{
		if(cm->cluster_scores[i]==empty_cluster_score) //cluster i is empty
		{
			// renumber all clusters that follow (-1)
			for(j=0;j<cm->total_num_motifs;j++)
			{
				if(cm->c_index[j]>i)
				{
					cm->c_index[j] -= 1;
				}
			}
			// move scores left in vector
			for(j=i;j<cm->num_clusters-1;j++)
			{
				cm->cluster_scores[j] = cm->cluster_scores[j+1];
			}
			cm->cluster_scores[j] = empty_cluster_score;
			// move merged indices left
			for(j=i;j<cm->num_clusters-1;j++)
			{
				strcpy(cm->merged_clusters[j], cm->merged_clusters[j+1]);
			}
			// eliminate the last empty cluster
			cm->num_clusters -= 1;
			// repeat and break the loops
			eliminate_empty_clusters(cm, empty_cluster_score);
			return;
		}
	}
}

// print num_clusters, cluster_scores and c_index for clustered_motifs
void print_clusters_info(clustered_motifs *cm)
{
	int i;
	printf("Clusters - num_clusters %i\n", cm->num_clusters);
	printf("Clusters - cluster_scores:\n");
	for(i=0;i<cm->num_clusters;i++)
	{
		printf("%4.2lf\t", cm->cluster_scores[i]);
	}
	printf("\n");
	// c_index
	printf("Clusters - c_index:\n");
	for(i=0;i<cm->total_num_motifs;i++)
	{
		printf("%i\t", cm->c_index[i]);
	}
	printf("\n");
/*	// indexes_of_unique_motifs
	printf("Clusters - indexes_of_unique_motifs:\n");
	for(i=0;i<cm->num_unique_motifs;i++)
	{
		printf("%i\t", cm->indexes_of_unique_motifs[i]);
	}
	printf("\n");*/
}



// compute scores for all pairs of clusters, and merge the first pair for which the score improves
void merge_cluster_co_binding(clustered_motifs *cm, motif_score_sign sign, params *p, double empty_cluster_score)
{
	int num_explained_peaks = p->d.num_seq_level_stats;
	int *explained_peaks1 = malloc(sizeof(int) * num_explained_peaks);
	int *explained_peaks2 = malloc(sizeof(int) * num_explained_peaks);

	int sign_multiplier = (sign == positive ? 1:-1);
	int target_cluster_index1, target_cluster_index2;
	cluster *c1, *c2;
	int i, j, k;

//	print_clusters_info(cm);

	int *overlap_genes, *union_genes;
	int num_overlap_genes, num_union_genes;
	double score1, score2, score_overlap, score_union, max_score_union = 0;
	int max_ind1, max_ind2;
	char *tmp, *sep = ",";
	MALLOC(tmp, 3*sizeof(char) + 1);

	for(i=0;i<cm->num_clusters;i++)
	{
		extract_cluster_targets_array(i, cm, explained_peaks1, num_explained_peaks);
		score1 = cm->cluster_scores[i];
		if(score1!=empty_cluster_score)
		{
			for(j=i+1;j<cm->num_clusters;j++)
			{
				extract_cluster_targets_array(j, cm, explained_peaks2, num_explained_peaks);
				score2 = cm->cluster_scores[j];
				if(score2!=empty_cluster_score)
				{
					// compute scores for clusters and their overlap
					// overlap of cluster 1 and cluster 2
					overlap_genes = get_ovelap(explained_peaks1, explained_peaks2, num_explained_peaks, &num_overlap_genes);
					score_overlap = sign_multiplier * p->geneset_ES(overlap_genes, num_overlap_genes, p->d.seq_level_stats, p->d.num_seq_level_stats, &p->scoring_params);
					free(overlap_genes);
					if(score_overlap>0)
					{// consider the pair only if their overlap has some positive score
						// union of cluster 1 and cluster 2
						union_genes = get_union(explained_peaks1, explained_peaks2, num_explained_peaks, &num_union_genes);
						score_union = sign_multiplier * p->geneset_ES(union_genes, num_union_genes, p->d.seq_level_stats, p->d.num_seq_level_stats, &p->scoring_params);
						free(union_genes);
						if((score_union > score1 && score_union > score2 && score_union > max_score_union) && (score_overlap > score1 || score_overlap > score2)) //
						{
							//printf("C%i, C%i, %5.2lf, %5.2lf, %5.2lf, %5.2lf\n", i + 1, j + 1, score1, score2, score_union, score_overlap);
							max_ind1 = i;
							max_ind2 = j;
							max_score_union = score_union;
						}

					}
				}
			}

		}
	}
	if(max_score_union>0)
	{
		// merge clusters max_ind1 & max_ind2 (new index = max_ind1), and create new cluster set
//		printf("%i + %i - %4.2lf\n", max_ind1 + 1, max_ind2 + 1, max_score_union);
		for(k=0;k<cm->total_num_motifs;k++)
		{
			if(cm->c_index[k]==max_ind2)
			{
				cm->c_index[k] = max_ind1;
			}
		}
		// store the indices
		sprintf(tmp, "%i", max_ind2+1);
		strcat(cm->merged_clusters[max_ind1], sep);
		strcat(cm->merged_clusters[max_ind1], tmp);
		// assign new scores
		cm->cluster_scores[max_ind1] = max_score_union;
		cm->cluster_scores[max_ind2] = empty_cluster_score; // this will ensure that the cluster is skipped
		// repeat
		merge_cluster_co_binding(cm, sign, p, empty_cluster_score);
	}
}

// extract the non-identical motif indexes into '*indexes_of_unique_motifs'
void extract_unique_motifs(d_gs **arr, int num, int **indexes_of_unique_motifs, int *num_unique)
{
	int j,k;
	*num_unique = 0;
	int motif_ind_local[num]; bool already_present;
	for(j=0;j<num;j++)
	{
		already_present = false;

		for(k=0;k<*num_unique;k++)
		{
			char *rev_compl_oligo = rev_compl(arr[motif_ind_local[k]]->degen_oligo);
			if(consider_rev_compl_flag == false)
			{
				FREE(rev_compl_oligo);
				MALLOC(rev_compl_oligo, sizeof(char) * strlen(arr[motif_ind_local[k]]->degen_oligo) + 1);
				strcpy(rev_compl_oligo, arr[motif_ind_local[k]]->degen_oligo);
			}
			// change to 'false' to keep identical occurrences
			if(eq(arr[j]->degen_oligo, arr[motif_ind_local[k]]->degen_oligo) == true || eq(arr[j]->degen_oligo, rev_compl_oligo) == true)
				already_present = true;//false;//
			FREE(rev_compl_oligo);
		}
		if(already_present == false)
		{
			motif_ind_local[*num_unique] = j;
			(*num_unique)++;
		}
	}
	MALLOC(*indexes_of_unique_motifs, sizeof(int) * (*num_unique));

	for(j=0;j<(*num_unique);j++)
	{
		(*indexes_of_unique_motifs)[j] = motif_ind_local[j];
//		printf("indexes_of_unique_motifs %i: %s %i\n",(*indexes_of_unique_motifs)[j], arr[(*indexes_of_unique_motifs)[j]]->degen_oligo, optimal_offsets[(*indexes_of_unique_motifs)[j]]);
	}
//	printf("Number of unique motifs: %i\n", *num_unique);
//	getchar();
}

// new procedure of clustering evolved motifs
void cluster_motifs_by_joint_score(d_gs **dg_evolved, int num_evolved, clustered_motifs *cm, params *p)
{
	cm->total_num_motifs = num_evolved;
	cm->motifs = dg_evolved;

	int i;
	init_cluster_index(cm->c_index, cm->total_num_motifs);

	// upn exit of 'cluster_output_by_joint_score', 'cm->c_index' is initialized with the cluster indexes of the motifs in the array 'cm->motifs'
	cluster_output_by_joint_score(cm->motifs, cm->total_num_motifs, cm->c_index, p->d.seq.index, p->d.num_seq_level_stats, p->d.seq.seq_length, p);

	// defined in 'utilites.c'
	// upon exit 'cm->indexes_of_unique_motifs' contains the indexes of the unique motifs in array 'dg_evolved'
	// and 'cm->num_unique_motifs' contains the number for elements in 'cm->indexes_of_unique_motifs'
	extract_unique_motifs(cm->motifs, cm->total_num_motifs, &cm->indexes_of_unique_motifs, &cm->num_unique_motifs);

	cm->num_clusters = get_num_clusters(cm->c_index, cm->total_num_motifs);
}


void cluster_motifs(d_gs **dg_evolved, int num_evolved, clustered_motifs *cm, params *p)
{
	cm->total_num_motifs = num_evolved;
	cm->motifs = dg_evolved;

	int i;
	init_cluster_index(cm->c_index, cm->total_num_motifs);

	// upn exit of 'cluster_output', 'cm->c_index' is initialized with the cluster indexes of the motifs in the array 'cm->motifs'
	cluster_output(cm->motifs, cm->total_num_motifs, cm->c_index, p->d.seq.index, p->d.num_seq_level_stats, p->d.seq.seq_length, p);
//	printf("before 'extract_unique_motifs'\n");fflush(stdout);

	// defined in 'utilites.c'
	// upon exit 'cm->indexes_of_unique_motifs' contains the indexes of the unique motifs in array 'dg_evolved'
	// and 'cm->num_unique_motifs' contains the number for elements in 'cm->indexes_of_unique_motifs'
	extract_unique_motifs(cm->motifs, cm->total_num_motifs, &cm->indexes_of_unique_motifs, &cm->num_unique_motifs);

	cm->num_clusters = get_num_clusters(cm->c_index, cm->total_num_motifs);
}

void destroy_motif_clusters(clustered_motifs *cm)
{
	int i;
	FREE(cm->indexes_of_unique_motifs);
//	FREE(cm->sim_scores_ensemble_pssm_to_known_pssm);
}

d_gs *construct_union_geneset(char *union_oligo, char *genes1, int num_genes1, int *genes2, int num_genes2, params *p)
{
	int union_genes[p->d.num_seq_level_stats]; int num_union_genes;
	get_union_genes(genes1, num_genes1, genes2, num_genes2, p->d.num_seq_level_stats, union_genes, &num_union_genes);
	if(num_union_genes == 0)
		return 0;
	d_gs *union_genset = create_sub_geneset(union_oligo, p->d.seq.index, union_genes, num_union_genes, p->d.num_seq_level_stats, p->d.seq.seq_length, p->scoring, p);
	return union_genset;
}

int count_draws;

bool promising_motif_seed(bool fast_mode, d_gs *motif_seed)
{
	return fast_mode == false || (motif_seed != 0 && motif_seed->score < 0);
}

bool run_motif_finder(params *p, IO_params *p_IO)
{
	time_t t;
	t = start_timer();
	int i,j,k;
	double *seq_level_stat = p->d.seq_level_stats;
	int num_seq            = p->d.num_seq_level_stats;
	score_type st          = p->scoring;
	motif_score_sign sign  = negative;

	// create gene sets corresponding to loaded oligos
	if(p->oligo_seeds.is_empty == true)
	{
		printf("'run_motif_finder': p->oligo_seeds.is_empty == true!");
		exit(1);
	}
	d_gs *dg1[p->oligo_seeds.num_oligos];

	if(p->use_conservation == true)
	{
		for(i=0;i<p->oligo_seeds.num_oligos;i++)
	  		dg1[i] = create_geneset_applying_orth_filter(p->oligo_seeds.oligos[i], p, 0);
	}
	else
	{
	 	for(i=0;i<p->oligo_seeds.num_oligos;i++)
	 	{
	 		//////////////////////////********************************
	 		//printf("\nOLIGO [%i]: %s\n",i,p->oligo_seeds.oligos[i]);
	 		//////////////////////////********************************
	 		dg1[i] = create_d_gs(p->oligo_seeds.oligos[i], p, 0);

	 	}

	}

	d_gs *dg[p->oligo_seeds.num_oligos];
	int ind = 0;

	int num_to_evolve = p->oligo_seeds.num_oligos, swap_ind = p->oligo_seeds.num_oligos - 1;
	for(i=0;i<p->oligo_seeds.num_oligos;i++)
	{
		if(promising_motif_seed(p->fast_mode, dg1[i]) == true)
		{
			dg[ind] = dg1[i];
			ind++;
		}
		else
			free_d_gs(dg1[i]);
	}

	if(p->significance.evaluate_significance == false)
		printf("\n# motif seeds = %i\n", ind);

	num_to_evolve = ind;
	sort_order so = increasing;
	sort_d_gs(dg, num_to_evolve, so);
	FILE *fp = stdout;

	// evolve motifs using a greedy strategy
	d_gs *dg_evolved[num_to_evolve];
	int curr_index = 0, num_evolved = 0;
	d_gs *top, *top_dg_orth;
	d_gs *top_so_far = 0;

	int improvement = 1;
	double top_running_score = 0.0, eps=1.0;
	char out_str[1000];

	for(i=0;i<num_to_evolve;i++)
	{
		t = start_timer();
		sprintf(out_str, "%-5i%-15s (%i)%5.2lf  =>  ", i+1, dg[curr_index]->degen_oligo, dg[curr_index]->num_genes, -dg[curr_index]->score);

		top = evolve_motif(dg[curr_index], p);


		if(p->significance.evaluate_significance == false && -top->score > top_running_score*eps)
		{
			top_running_score = -top->score;

			fprintf(fp, "%s%-15s (%i)%5.2lf% 10i\n", 
					out_str, top->degen_oligo, top->num_genes, -top->score, top->num_evolution_steps);fflush(stdout);

		}

		if(top != 0 && -top->score > 0)
		{
			if(p->significance.evaluate_significance == false)
			{
				if(top_running_score < -top->score)
				{
					top_running_score = -top->score;
					top_so_far = top;
					count_draws = 0;
				}
			}
			dg_evolved[num_evolved] = top;
			num_evolved++;
		}
		else
		{
			free_d_gs(top);
		}
		curr_index++;
	}

	// post-process independently evolved motifs
	sort_d_gs(dg_evolved, num_evolved, so);

	int num_significant_predictions = num_evolved;
	if(p->significance.evaluate_significance == true)
	{
		if(num_significant_predictions > 0)
		{
			FILE *fp = open_file(p_IO->null_scores_filename, "a");
			fprintf(fp, "%lf\n", -dg_evolved[0]->score);
			fclose(fp);
			top_random_scores[random_index] = -dg_evolved[0]->score;
			random_index++;
			p->significance.num_random_permutations_remaining--;
		}
		else
		{
//			printf("failed to evolve randomized score!\n");
		}
	}
	else
	{
		if(num_significant_predictions > 0)
		{		
			int c_index[num_evolved];
			int num_evolved_at_least_one_step = 0;
			for(i=0;i<num_evolved;i++)
			{
				if(is_valid_score(-dg_evolved[i]->score) == true)
					num_evolved_at_least_one_step++;
			}
			if(num_evolved_at_least_one_step == 0)
				printf("no evolved motifs!\n");


			// this determines how many of the evolved seeds to participate in the final clustering
			int num_to_consider = num_evolved_at_least_one_step;
	
			MALLOC(rev_compl_flags, sizeof(bool) * num_to_consider);
			MALLOC(optimal_offsets, sizeof(int) * num_to_consider);
			MALLOC(p_val, sizeof(double) * num_to_consider);

			clustered_motifs motif_clusters;
			MALLOC(motif_clusters.c_index, sizeof(int) * num_to_consider);
			
//			printf("cluster_motifs\n");getchar();
			//cluster_motifs(dg_evolved, num_to_consider, &motif_clusters, p); // from Stoyan
			cluster_motifs_by_joint_score(dg_evolved, num_to_consider, &motif_clusters, p);

// ***************** EXTRA OUTPUT ********************
			if(1) // output independently evolved motifs and their similarity matrix
			{
				output_evolved_motifs(dg_evolved, num_to_consider, num_seq, motif_clusters.c_index, sign, p, p_IO);
				output_similarity_matrix(dg_evolved, num_to_consider, p, p_IO);
			}
// ***************************************************

			// printf("generate_summary_output\n");getchar();
			generate_summary_output(dg_evolved, num_to_consider, num_seq, 0, motif_clusters.c_index, motif_clusters.sim_score_top_ensemble_pssm_to_known_pssm, sign, p, p_IO);
			// generate pssm and plot logos - clusters
			output_top_cluster_ensemble_pssm(&motif_clusters, p_IO, p);
			//print_clusters_info(&motif_clusters);



			destroy_motif_clusters(&motif_clusters);
			FREE(p_val);
			FREE(optimal_offsets);
			FREE(rev_compl_flags);
		}
	}
	free_d_gs_batch(dg_evolved, num_evolved);
	return num_significant_predictions > 0;
}


void swap_oligo_with_rev_compl(d_gs *dg, int index_in_array)
{
	char *tmp_oligo;
	gsl_matrix *tmp_pssm; 

	tmp_oligo = rev_compl(dg->degen_oligo);
	strcpy(dg->degen_oligo, tmp_oligo);

	tmp_pssm = get_rev_compl_pssm(dg->pssm);
	gsl_matrix_memcpy(dg->pssm, tmp_pssm);

	gsl_matrix_free(tmp_pssm);
	FREE(tmp_oligo);

	if(rev_compl_flags[index_in_array] == true)
		rev_compl_flags[index_in_array] = false;
	else
		rev_compl_flags[index_in_array] = true;
}

int find_top_cluster_motif_index(int *c_index, int *unique, int num_unique, int cluster_number)
{	
	int i;
	for(i=0;i<num_unique;i++)
	{
		if(c_index[unique[i]] == cluster_number)
			return unique[i];
	}
	return -1;
}

gsl_matrix *generate_cropped_pssm(gsl_matrix *ensemble_pssm, int star_col_index, int num_cols_out)
{
	gsl_matrix *ensemble_pssm_out = gsl_matrix_alloc(no_primary_IUPAC_nucleotides, num_cols_out);
	gsl_matrix_view mview_out = gsl_matrix_submatrix(ensemble_pssm, 0, star_col_index, no_primary_IUPAC_nucleotides, num_cols_out);
	gsl_matrix_memcpy(ensemble_pssm_out, &mview_out.matrix);
	return ensemble_pssm_out;
}


// threshold -> remove all flanking positions that have less than 'threshold' frequency of the most frequent nucleotide
gsl_matrix *crop_pssm_ends(gsl_matrix *ensemble_pssm, double threshold)
{
	if(threshold < 0.125)
		threshold = 0.125; // 0.25 for 4 column PSSM, 0.125 for 8 column PSSM
	int i,j;
	bool fl1 = false, fl2 = false;
//	gsl_vector *v = gsl_vector_alloc(4);
	int col_out1 = 0, col_out2 = ensemble_pssm->size2 - 1;

	for(i=0;i<ensemble_pssm->size2 / 2;i++)
	{
		double max = 0;
		for(j=0;j<ensemble_pssm->size1;j++)
			max = MAX(max, gsl_matrix_get(ensemble_pssm, j, i));
		if(fl1 == false && max < threshold)
			col_out1++;
		else
			fl1 = true;
		max = 0;
		for(j=0;j<ensemble_pssm->size1;j++)
			max = MAX(max, gsl_matrix_get(ensemble_pssm, j, ensemble_pssm->size2  - i - 1));
		if(fl2 == false && max < threshold)
			col_out2--;
		else
			fl2 = true;
	}
	int num_cols_out = col_out2 - col_out1 + 1;
	gsl_matrix *ensemble_pssm_out =	generate_cropped_pssm(ensemble_pssm, col_out1, num_cols_out);
	return ensemble_pssm_out;
}


void output_top_cluster_ensemble_pssm(clustered_motifs *motif_clusters, IO_params *p_IO, params *p)
{
	int i,j,curr_cluster_index;
	d_gs **dg_evolved   = motif_clusters->motifs;
	int num_to_consider = motif_clusters->total_num_motifs;
	int *unique         = motif_clusters->indexes_of_unique_motifs;
	int num_unique      = motif_clusters->num_unique_motifs;

	int indexes_of_top_cluster_members[num_unique];
	int offsets_of_top_cluster_members[num_unique];
	double scores_of_top_cluster_members[num_unique];
	bool rev_compl_flags_of_top_cluster_members[num_unique];

	double fraction_of_top_sequences_to_consider = p->fraction_of_top_sequences_to_consider;

	int how_many_top_sequences = (int)(fraction_of_top_sequences_to_consider * p->d.seq.num_seq);
	
	int num_to_output;
	num_to_output = p_IO->num_motif_clusters_to_output;
	if(num_unique < num_to_output)
	{
		num_to_output = num_unique;
		p_IO->num_motif_clusters_to_output = num_unique;
	}

	int top_cluster_members[num_unique];
	int ind_top = 0;
	for(i=0;i<num_unique;i++)
	{
		if(motif_clusters->c_index[unique[i]] == ind_top)
		{
			top_cluster_members[ind_top] = unique[i];
			ind_top++;
		}
	}

	for(curr_cluster_index = 0; curr_cluster_index < num_to_output;curr_cluster_index++)
	{
		// helper arrays containing information on the individual cluster members
		int num_in_top_cluster = 0, min_offset = 0, max_offset = 0;
		int right_overhang = 0, left_overhang = 0;

		int top_cluster_motif_index = find_top_cluster_motif_index(motif_clusters->c_index, unique, num_unique, curr_cluster_index);
		if(top_cluster_motif_index == -1)
			break;

		double top_cluster_score = dg_evolved[top_cluster_motif_index]->score;
		int top_cluster_num_genes = dg_evolved[top_cluster_motif_index]->num_genes;
		double score_threshold = -top_cluster_score * p->cluster_pssm_threshold_fraction;

		for(i=0;i<num_unique;i++)
		{
			if(motif_clusters->c_index[unique[i]] == curr_cluster_index)
			{
				bool targets_enough_seq_regions = true;//dg_evolved[unique[i]]->num_genes >= 0.5 * top_cluster_num_genes;
				if
				(
					(targets_enough_seq_regions == true)
					&& (p_val[unique[i]] <= p->hypegeom_p_val_cutoff)
					&& (-dg_evolved[unique[i]]->score >= score_threshold)
//					(curr_cluster_index + 1 >= ind_top || -dg_evolved[unique[i]]->score >= -dg_evolved[top_cluster_members[curr_cluster_index+1]]->score)
				)//p_val[unique[i]] <= p->hypegeom_p_val_cutoff)
				{
					d_gs *cluster_member = dg_evolved[unique[i]];
					// assign index of the cluster member
					indexes_of_top_cluster_members[num_in_top_cluster] = unique[i];
	
					if(dg_evolved[top_cluster_motif_index]->oligo_length > dg_evolved[unique[i]]->oligo_length)
						offsets_of_top_cluster_members[num_in_top_cluster] = optimal_offsets[unique[i]];
					else
						offsets_of_top_cluster_members[num_in_top_cluster] = -optimal_offsets[unique[i]];
					scores_of_top_cluster_members[num_in_top_cluster] = fabs(dg_evolved[unique[i]]->score);
		
					// set min and max offset for the ensemble motif
					if(min_offset > optimal_offsets[unique[i]])
						min_offset = optimal_offsets[unique[i]];
		
					if(max_offset < optimal_offsets[unique[i]])
						max_offset = optimal_offsets[unique[i]];
		
					right_overhang = MAX(right_overhang, optimal_offsets[unique[i]] + dg_evolved[unique[i]]->oligo_length);
					build_pssm_from_top_sequences_only(p->d.seq.index, p->d.seq.seq_length, p->d.seq.num_seq, cluster_member, how_many_top_sequences);

					// set up revelse compl flag to false as we swap if necessary
					if(rev_compl_flags[unique[i]] == true)
						swap_oligo_with_rev_compl(dg_evolved[unique[i]], unique[i]);
					rev_compl_flags_of_top_cluster_members[num_in_top_cluster] = rev_compl_flags[unique[i]];
					num_in_top_cluster++;
				}
			}
		}

		int ensemble_matrix_length = right_overhang - min_offset;
		gsl_matrix *ensemble_pssm  = gsl_matrix_alloc(no_primary_IUPAC_nucleotides, ensemble_matrix_length);
		gsl_matrix_set_zero(ensemble_pssm);
		int pos;

		int max_count_occurrences = 0;
		int num_occurrences_for_position[ensemble_pssm->size2];
		for(pos = min_offset; pos < right_overhang;pos++)
		{
			int ind = pos - min_offset;
			num_occurrences_for_position[ind] = 0;
			int num_for_curr_position = 0;
			for(i=0;i<num_in_top_cluster;i++)
			{
				d_gs *top_cluster_member = dg_evolved[indexes_of_top_cluster_members[i]];
				if(offsets_of_top_cluster_members[i] <= ind && (top_cluster_member->oligo_length > ind - offsets_of_top_cluster_members[i]) )
					num_occurrences_for_position[ind] += dg_evolved[indexes_of_top_cluster_members[i]]->num_genes;
			}
			max_count_occurrences = max_count_occurrences < num_occurrences_for_position[pos] ? num_occurrences_for_position[ind] : max_count_occurrences;
		}
		for(pos = min_offset; pos < right_overhang; pos++)
		{
			int ind = pos;// - min_offset;
			double sum_of_scores_for_current_position = 0;
			d_gs *top_cluster_member;
			int num_for_curr_position = 0;
			for(i=0;i<num_in_top_cluster;i++)
			{
				top_cluster_member = dg_evolved[indexes_of_top_cluster_members[i]];
				if(offsets_of_top_cluster_members[i] <= ind && (top_cluster_member->oligo_length > ind - offsets_of_top_cluster_members[i]) )
					sum_of_scores_for_current_position += scores_of_top_cluster_members[i];
			}

			double total_freq = 0;
			for(i=0;i<num_in_top_cluster;i++)
			{
				top_cluster_member = dg_evolved[indexes_of_top_cluster_members[i]];
				if( offsets_of_top_cluster_members[i] <= pos && (top_cluster_member->oligo_length > pos - offsets_of_top_cluster_members[i]) )
				{
					for(j=0;j<no_primary_IUPAC_nucleotides;j++)
					{
						gsl_matrix *tmp_pssm = top_cluster_member->pssm;
						double freq_cluster_member = gsl_matrix_get(tmp_pssm, j, pos - offsets_of_top_cluster_members[i]);
						double weighted_freq_cluster_member = (scores_of_top_cluster_members[i] / sum_of_scores_for_current_position) * freq_cluster_member;
						double curr_freq = gsl_matrix_get(ensemble_pssm, j, pos - min_offset);
						double new_freq = curr_freq + weighted_freq_cluster_member;
						gsl_matrix_set(ensemble_pssm, j, pos - min_offset, new_freq);
					}
				}
			}
		}

		// remove all flanking positions that have less than 'threshold' frequencey of the most frequent nucleotide
		double threshold = p->pssm_crop_threshold;
		gsl_matrix *ensemble_pssm_cropped = crop_pssm_ends(ensemble_pssm, p->pssm_crop_threshold);
		gsl_matrix_free(ensemble_pssm);	
		ensemble_pssm = ensemble_pssm_cropped;
		p->ensemble_pssm[curr_cluster_index] = gsl_matrix_alloc(ensemble_pssm->size1, ensemble_pssm->size2);
		gsl_matrix_memcpy(p->ensemble_pssm[curr_cluster_index], ensemble_pssm);
		gsl_matrix *tmp_pssm = dg_evolved[top_cluster_motif_index]->pssm;
		dg_evolved[top_cluster_motif_index]->pssm = ensemble_pssm;

		char suffix[25];
		sprintf(suffix, "_motif%i", curr_cluster_index + 1);
		create_logo_using_pssm_freq(dg_evolved[top_cluster_motif_index], p_IO, p, suffix, &motif_clusters->sim_score_top_ensemble_pssm_to_known_pssm);
		dg_evolved[top_cluster_motif_index]->pssm = tmp_pssm;
	}
}

void create_logo_using_pssm_freq(d_gs *top, IO_params *p_IO, params *p, char *suffix, double *sim_score_top_ensemble_pssm_to_known_pssm)
{
//	printf("Create logo for the cluster with top evolved motif: %s\n", top->degen_oligo);fflush(stdout);
	char web_logo_num_ev[strlen(p_IO->out_weblogo) + strlen("_prediction") + strlen(suffix) + 2];
	strcpy(web_logo_num_ev, p_IO->out_weblogo);
	strcat(web_logo_num_ev, suffix);


	FILE *fp = open_file(web_logo_num_ev, "w");
	bool rev_flag = false;
	gsl_matrix *rev_compl_top_pssm = get_rev_compl_pssm(top->pssm);
	double similarity = -1; //	similarity is computed by similarity_score_Harbison_with_overhang_no_background in similarity_metrics.c
	if(p->known_pssm != 0)
	{
		int required_core_length = p->required_core_length; int offset_top_score;
		if(consider_rev_compl_flag == false)
		{
			similarity = eval_PSSM_similarity(top->pssm, p->known_pssm, required_core_length, &offset_top_score);
			//similarity = distance_metric_DRIM(top->degen_oligo, p-> ???)
			rev_flag = false;
		}
		else
		{
			double similarity1 = eval_PSSM_similarity(top->pssm         , p->known_pssm,required_core_length, &offset_top_score);
			double similarity2 = eval_PSSM_similarity(rev_compl_top_pssm, p->known_pssm,required_core_length, &offset_top_score);

			similarity = MAX(similarity1, similarity2);
			rev_flag = similarity1 < similarity2;
			gsl_matrix_free(rev_compl_top_pssm);
		}
	}

	char file_pssm[500];
	if(false)//similarity != -1)
	{
		sprintf(file_pssm, "%s_similarity_%3.2lf", web_logo_num_ev, similarity);
		*sim_score_top_ensemble_pssm_to_known_pssm = similarity;
	}
	else
	{
		sprintf(file_pssm, "%s", web_logo_num_ev);
		*sim_score_top_ensemble_pssm_to_known_pssm = 0;
	}

	gsl_matrix *pssm_out_from_TRANSFAC = gsl_matrix_alloc(top->pssm->size1, top->pssm->size2);
	if(rev_flag == false)
	{
//		printf("rev_flag == false\n");fflush(stdout);
//		generate_instantiations_from_pssm_freq(top->pssm, web_logo_num_ev);
		write_pssm_TRANSFAC_like_format_freq(top->pssm, web_logo_num_ev, p_IO->img_type);
//		read_pssm_TRANSFAC_like_format_freq(pssm_out_from_TRANSFAC, file_pssm);
	}
	else
	{		
//		printf("rev_flag == true\n");fflush(stdout);
//		generate_instantiations_from_pssm_freq(rev_compl_top_pssm, web_logo_num_ev);
		write_pssm_TRANSFAC_like_format_freq(rev_compl_top_pssm, web_logo_num_ev, p_IO->img_type);
//		read_pssm_TRANSFAC_like_format_freq(pssm_out_from_TRANSFAC, file_pssm);
	}
	gsl_matrix_free(pssm_out_from_TRANSFAC);

	// pssm is built based on literature consensus from the occurrences in the given sequence set
	if(false)//p->known_pssm != 0)
	{
		char web_logo_num_ev1[strlen(p_IO->out_weblogo) + strlen("_literature")+1];
		strcpy(web_logo_num_ev1, p_IO->out_weblogo);
		strcat(web_logo_num_ev1, "_literature");
		write_pssm_TRANSFAC_like_format_freq(p->known_pssm, web_logo_num_ev1, p_IO->img_type);
		unlink(web_logo_num_ev1);
	}

	gsl_matrix_free(rev_compl_top_pssm);
	char command[500];
//	sprintf(command, "mv %s %s_TRANSFAC_like.txt", web_logo_num_ev, web_logo_num_ev);
//	system(command);

//	sprintf("mv %s .pdf %s.pdf", web_logo_num_ev, web_logo_num_ev);
//	system(command);
//	printf("deleting %s\n", web_logo_num_ev);
//	unlink(web_logo_num_ev);
}


int get_num_clusters(int *c_index, int num)
{
	int num_clusters = 1;
	int i;
	for(i=0;i<num;i++)
	{
//		printf("c_index[%i] = %i\n", i, c_index[i]);
		if(c_index[i] + 1 > num_clusters)
			num_clusters = c_index[i] + 1;
	}
	return num_clusters;
}
//extern char **seq_ids_global;

void output_geneset_ids(d_gs *geneset, params *p, IO_params *p_IO)
{
	int i;
	char PCC_out_path[200];
	strcpy(PCC_out_path, p_IO->output_dir);//output_directory_global);
	strcat(PCC_out_path, "gene_set_ids");
	FILE *fp1 = fopen(PCC_out_path, "a");

	fprintf(fp1, "%4i %-20s", geneset->num_genes, geneset->degen_oligo);
  	for(i=0;i<geneset->num_genes;i++)
	{
//		printf("%i: %s\n", arr[0]->genes[i], seq_ids_global[arr[0]->genes[i]]);
		fprintf(fp1, "%s,", p->d.seq.seq_ids[geneset->genes[i]]);
	}
	fprintf(fp1, "\n");
	fflush(fp1);
	fclose(fp1);
}



inline void print_spaces(FILE *fp, int num)
{
	int k;
	for(k=0;k<num;k++)
		fprintf(fp, " ");
}

inline void print_spaces2string(char *s, int num)
{
	int k;
	for(k=0;k<num;k++)
		strcat(s, " ");
}

double estimate_Gaussian_fit_pval(double *top_random_scores, int num_top_random_scores, double observed_score)
{
	int i;
	double mean = 0.0, sum_squares = 0.0;
	for(i=0;i<num_top_random_scores;i++)
	{
		mean += top_random_scores[i];
		sum_squares += top_random_scores[i]*top_random_scores[i];
	}
	mean = mean / num_top_random_scores;
	double sd = sqrt(sum_squares / num_top_random_scores - mean*mean);
	double x = observed_score - mean;
	double z_score = x / sd;
	double p_val = gsl_cdf_gaussian_Q(x, sd);
	return p_val;
}


extern double entropy_optiamal_aligned_core;

double estimate_empirical_pval(double *top_random_scores, int num_top_random_scores, double observed_score)
{
	int i;
	double p_val = 0.0;
	for(i=0;i<num_top_random_scores;i++)
	{
		if(top_random_scores[i] >= observed_score)
		{
			p_val += 1.0;
		}	
	}
	p_val = p_val / num_top_random_scores;
	return p_val;
}


void output_motif_clusters(d_gs **arr, int num, int *c_index, int *unique, int num_unique_motifs, double sim_score_top_ensemble_pssm_to_known_pssm, 
						motif_score_sign sign, params *p, IO_params *p_IO, d_gs *expected_consensus)
{
	int i,j,k;
	FILE *fp = open_file(p_IO->out_summary, "a");
	fprintf(fp, "Motif cluster  Motifs              Score(#targets, %%targets, %%cumulative)");

	if(random_index > 0)
		fprintf(fp, "              p-value");
	fprintf(fp, "\n");

	char out_motif_occ_file[500];
	sprintf(out_motif_occ_file, "%spredicted_targets_with_offsets_%s", p_IO->output_dir, p->TF_name);
	FILE *fp1 = open_file(out_motif_occ_file, "w");
	fclose(fp1);

	char out_motif_genesets_file[500];

	sprintf(out_motif_genesets_file, "%spredicted_targets_%s", p_IO->output_dir, p->TF_name);
	FILE *fp_motif_genesets = open_file(out_motif_genesets_file, "w");
	fclose(fp_motif_genesets);

//	printf("here555555\n");fflush(stdout);

	int sign_multiplier = (sign == positive ? 1:-1);
	bool non_empty_flag = false;
//	printf("num = %i\n", num);

	int num_clusters = get_num_clusters(c_index, num);
//	printf("num_clusters = %i\n", num_clusters);

//	printf("here66666666\n");fflush(stdout);

	int top_cluster_members[num_unique_motifs];
	int ind_top = 0;
//	printf("output_motif_clusters: outputing cluster members that score better than next cluster's top member\n");
	for(i=0;i<num_unique_motifs;i++)
	{
		if(c_index[unique[i]] == ind_top)
		{
			top_cluster_members[ind_top] = unique[i];
//			printf("%s: %10.3lf\n", arr[unique[i]]->degen_oligo, arr[unique[i]]->score);
			ind_top++;
		}
	}
//	printf("ind_top = %i\n", ind_top);fflush(stdout);
//	printf("num_unique_motifs = %i\n", num_unique_motifs);fflush(stdout);

//	printf("num_clusters = %i\n\n", num_clusters);fflush(stdout);

//	FILE *fp_clust = fopen("scores_clustered_motifs", "w");

	// used in significance estimate
//	num_top_scores = num_clusters < num_top_scores ? num_clusters : num_top_scores;
//	for(i=0;i<num_top_scores;i++)
//		top_scores[i] = -arr[indexes_of_unique_motifs[i]]->score;

	// number of positions the main motif is offset from the start
	int d = 10;

//	char *out_file_res[500];
//	sprintf(out_file_res, "%s/res", p_IO->output_dir);
//	FILE *fp_sim = fopen(out_file_res, "w");

	char *out_file_res[500];
	sprintf(out_file_res, "%smax_sim_%s", p_IO->output_dir, p->TF_name);
	FILE *fp_sim = fopen(out_file_res, "w");
//	printf("p_IO->output_dir: %s\n", p_IO->output_dir);
//	printf("p->TF_name: %s\n", p->TF_name);
//	printf("out_file_res: %s\n", out_res);
//	getchar();
	double top_cluster_similarity = 0;

	int *explained_peaks_cumulative = malloc(sizeof(int) * p->d.num_seq_level_stats);
	int *explained_peaks_current = malloc(sizeof(int) * p->d.num_seq_level_stats);

	// intialize explained cumulative peaks count for all clusters
	for(i=0;i<p->d.num_seq_level_stats;i++)
		explained_peaks_cumulative[i] = 0;

	bool first_i_equal_zero = true;
	int ii;
//	int best_pssm_index = 0;
//	double best_sim = 0;
	for(ii=0;ii<num_clusters;ii++)
	{
		int top_cluster_motif_index = find_top_cluster_motif_index(c_index, unique, num_unique_motifs, ii);
		if(top_cluster_motif_index == -1)
			break;
		double top_cluster_score = arr[top_cluster_motif_index]->score;
		double score_threshold = -top_cluster_score * p->cluster_pssm_threshold_fraction;
		
//		printf("cluster %i\n", ii);fflush(stdout);
//		printf("score_threshold   = %lf\n", score_threshold);fflush(stdout);
//		printf("p->cluster_pssm_threshold_fraction   = %lf\n", p->cluster_pssm_threshold_fraction);fflush(stdout);
//		getchar();

		bool first = true;
		int extra_offset = -20;
		char *cluster_label_oligo;
		int cluster_label_num_genes;
		char dummy[10];

		double prev_fraction_cum_curr = 0;

		// intialize explained peaks count for current cluster
		int num_peaks_curr = 0;
		for(j=0;j<num_unique_motifs;j++)
		{
//			printf("%i: %s\t%lf\n", j, arr[unique[j]]->degen_oligo, -arr[unique[j]]->score);
//			printf("p_val[unique[j]]  = %lf\n", p_val[unique[j]]);fflush(stdout);

//			if(c_index[unique[j]] == ii && ( (ii+1 >= ind_top) || -arr[unique[j]]->score >= -arr[top_cluster_members[ii+1]]->score) )
			if
			(
				c_index[unique[j]] == ii			 && 
				p_val[unique[j]] <= p->hypegeom_p_val_cutoff     && 
				-arr[unique[j]]->score >= score_threshold
//				(ii + 1 >= ind_top || -arr[unique[j]]->score >= -arr[top_cluster_members[ii+1]]->score)
			)//p_val[unique[i]] <= p->hypegeom_p_val_cutoff)
//			if(c_index[unique[j]] == ii)// && -arr[unique[j]]->score >= -arr[top_cluster_members[ii]]->score * p->cluster_pssm_threshold_fraction )
			{
/*
				if(p->known_pssm != 0)
				{
//					printf("p->required_core_length = %i\n", p->required_core_length);fflush(stdout);
					double max_current_sim;
					int required_core_length = p->required_core_length; int offset_top_score;
//					double similarity11 = similarity_score_Harbison_with_overhang_no_background(arr[unique[j]]->pssm, p->known_pssm, required_core_length, &offset_top_score);
					double similarity11 = eval_PSSM_similarity(arr[unique[j]]->pssm, p->known_pssm, required_core_length, &offset_top_score);

					if(consider_rev_compl_flag == false)
					{
						max_current_sim = similarity11;
					}
					else
					{
						gsl_matrix *rev_compl_pssm = get_rev_compl_pssm(arr[unique[j]]->pssm);
//						double similarity22 = similarity_score_Harbison_with_overhang_no_background(rev_compl_pssm, p->known_pssm, required_core_length, &offset_top_score);
						double similarity22 = eval_PSSM_similarity(rev_compl_pssm, p->known_pssm, required_core_length, &offset_top_score);
						gsl_matrix_free(rev_compl_pssm);
						max_current_sim = MAX(similarity11, similarity22);
					}
					if(ii == 0)
					{
						if(best_sim < max_current_sim && p_val[unique[j]] <= p->hypegeom_p_val_cutoff)
						{
							best_sim = max_current_sim;
							best_pssm_index = j;
						}
					}
				}
				else
					best_pssm_index = -1;
*/

				if(random_index > 0)
				{
//					printf("random_index = %i\n", random_index);
					int num_top_random_scores = random_index;
					double observed_score = arr[unique[j]]->score*sign_multiplier;
//					arr[unique[j]]->p_value = estimate_empirical_pval(top_random_scores, num_top_random_scores, observed_score);
					arr[unique[j]]->p_value = estimate_Gaussian_fit_pval(top_random_scores, num_top_random_scores, observed_score);
//					printf("%s: p-val = %lf\n", arr[unique[j]]->degen_oligo, arr[unique[j]]->p_value);

					double pval_cutoff = 1.0;
					bool fails_to_reach_a_pval_cutoff = arr[unique[j]]->p_value > pval_cutoff;
					if(fails_to_reach_a_pval_cutoff == true)
					{
						continue;
					}
				}

				if(first == true)
				{	
					// intialize explained peaks count for current cluster
					for(i=0;i<p->d.num_seq_level_stats;i++)
						explained_peaks_current[i] = 0;
					num_peaks_curr = 0;

					if(arr[unique[j]]->score*sign_multiplier > 0)
					{
						// offset due to cluster number printed at the start of the cluster label line
						extra_offset = sprintf(dummy, "%i", ii+1);
						fprintf(fp, "%i", ii+1);
					}
					cluster_label_oligo = arr[unique[j]]->degen_oligo;
					cluster_label_num_genes = arr[unique[j]]->num_genes;
					first = false;
					// test the number of clusters for which to output the target sets (in a separate file)
					if(c_index[unique[j]] < p->num_clusters_to_output_motif_occurrences_for)
						output_motif_target_set(arr[unique[j]], p, out_motif_occ_file, out_motif_genesets_file);
				}
				else
				{
					bool fails_to_target_enough_seq_regions = false;//arr[unique[j]]->num_genes < 0.5 * cluster_label_num_genes;
					bool fails_to_co_occurr_in_enough_seq_regions = p_val[unique[j]] > p->hypegeom_p_val_cutoff;

					if(
						fails_to_co_occurr_in_enough_seq_regions == true
						||  fails_to_target_enough_seq_regions   == true
					  )
						continue;
				}

				int leading_edge_index, leading_edge_num_genes;
				double ks_score = ks_ES(arr[unique[j]]->genes, arr[unique[j]]->num_genes, p->d.num_seq_level_stats, &leading_edge_index, &leading_edge_num_genes);
				int total_offset = 35;
				// if score has the right sign
				if(arr[unique[j]]->score*sign_multiplier > 0)
				{
					bool first = false;
					if(num_peaks_curr == 0)
						first = true;
					for(i=0;i<arr[unique[j]]->num_genes;i++)
					{
						if(explained_peaks_current[arr[unique[j]]->genes[i]] == 0)
							num_peaks_curr+=1;
						explained_peaks_current[arr[unique[j]]->genes[i]] += 1;
						explained_peaks_cumulative[arr[unique[j]]->genes[i]] += 1;
					}
					double fraction_of_total = (1.0 * arr[unique[j]]->num_genes) / p->d.num_seq_level_stats;
					double fraction_cum_curr = (1.0 * num_peaks_curr) / p->d.num_seq_level_stats;
//					if((int)(100 * (fraction_cum_curr - prev_fraction_cum_curr)) >= 0)
					if(true)//fraction_cum_curr - prev_fraction_cum_curr > 0.0)
					{
						non_empty_flag = true;
		
						int offset, balance_offset;
						if(strlen(cluster_label_oligo) > arr[unique[j]]->oligo_length)
							offset = d + optimal_offsets[unique[j]] + extra_offset;
						else
							offset = d - optimal_offsets[unique[j]] + extra_offset;
						balance_offset = total_offset - strlen(arr[unique[j]]->degen_oligo) - offset;
		
						if(eq(arr[unique[j]]->degen_oligo, cluster_label_oligo) == true)
						{
							offset = d;
							balance_offset = total_offset - strlen(arr[unique[j]]->degen_oligo) - offset - extra_offset;
						}
						print_spaces(fp, offset);
		
						if(rev_compl_flags[unique[j]] == true)
						{
							char *rev_compl_oligo = rev_compl(arr[unique[j]]->degen_oligo);
							fprintf(fp, "%+s", rev_compl_oligo);
							FREE(rev_compl_oligo);
						}
						else
							fprintf(fp, "%+s", arr[unique[j]]->degen_oligo);
						print_spaces(fp, balance_offset);
		
						fprintf(fp, "%4.2lf(%i, %i, %i)", -arr[unique[j]]->score, arr[unique[j]]->num_genes, 
										(int)(100 * fraction_of_total), (int)(100 * fraction_cum_curr));
						if(p->known_pssm != 0)
						{
							double max_sim;
							gsl_matrix *tmp;
							if(ii < p_IO->num_motif_clusters_to_output)// && p->ensemble_pssm[ii] != 0)
							{
								tmp = arr[unique[j]]->pssm;
							}
		
							int required_core_length = p->required_core_length; int offset_top_score;
//							double similarity1 = similarity_score_Harbison_with_overhang_no_background(p->known_pssm, arr[unique[j]]->pssm, required_core_length, &offset_top_score);
							double similarity1 = eval_PSSM_similarity(arr[unique[j]]->pssm, p->known_pssm, required_core_length, &offset_top_score);
							double similarity2 = similarity1;
							if(consider_rev_compl_flag == true)
							{
								gsl_matrix *rev_compl_pssm = get_rev_compl_pssm(arr[unique[j]]->pssm);
//								similarity2 = similarity_score_Harbison_with_overhang_no_background(p->known_pssm, rev_compl_pssm, required_core_length, &offset_top_score);
								similarity2 = eval_PSSM_similarity(rev_compl_pssm, p->known_pssm, required_core_length, &offset_top_score);
								gsl_matrix_free(rev_compl_pssm);
							}
//							printf("here\n"); getchar();
							char *known = "literature: ";
							max_sim = MAX(similarity1, similarity2);
							double entropy_threshold = 1.0;
//							if(entropy_optiamal_aligned_core > entropy_threshold)
//								max_sim = 0.0;
			
							fprintf(fp ,"%+*s%-8.2lf", strlen(known) + 2, known, max_sim);
//							double default_eps = -1;
//							double ave_ent = ave_entropy(arr[unique[j]]->pssm, default_eps);
//							fprintf(fp ,"(ave entr = %lf, entropy_optiamal_aligned_core = %lf)", ave_ent, entropy_optiamal_aligned_core);
							if(entropy_optiamal_aligned_core >= entropy_threshold)
							{
//								fprintf(stdout ,"%i: %s: ave_entropy_aligned_core = %lf", ii, arr[unique[j]]->degen_oligo, entropy_optiamal_aligned_core);
//								fprintf(fp ,"(ave_entropy_aligned_core = %lf)", entropy_optiamal_aligned_core);
							}
						
							if(ii < p_IO->num_motif_clusters_to_output)// && p->ensemble_pssm[ii] != 0)
							{
								arr[unique[j]]->pssm = tmp;
								if(top_cluster_similarity < max_sim)
									top_cluster_similarity = max_sim;
							}
						}
						if(arr[unique[j]]->p_value >= 0)
							fprintf(fp, "          % 10.8lf", arr[unique[j]]->p_value);
						fprintf(fp, "\n");
						prev_fraction_cum_curr = fraction_cum_curr;
					}
					if(first == true)
					{
//						prev_fraction_cum_curr = fraction_cum_curr;
					}
				}
			}
		}
		if(non_empty_flag == true)
		{	
			int num_explained_cum = 0;
			for(i=0;i<p->d.num_seq_level_stats;i++)
			{
				if(explained_peaks_cumulative[i] > 0)
					num_explained_cum += 1;
			}
			double fraction_of_total = (1.0 * num_explained_cum) / p->d.num_seq_level_stats;
			fprintf(fp ,"cumulative \% = %i\n", (int)(100 *fraction_of_total) );fflush(fp);

			non_empty_flag = false;
		}
		fflush(fp);
		if(ii == 1)
			first_i_equal_zero = false;
	}
	free(explained_peaks_cumulative);
	free(explained_peaks_current);

	fprintf(fp_sim, "%s\t%3.2lf\n", p->TF_name, top_cluster_similarity);
//	fprintf(fp_sim, "%s\t%3.2lf\n", p->TF_name, sim_score_top_ensemble_pssm_to_known_pssm);
//	printf("here111\n");
//	getchar();

//	if(best_pssm_index != -1)
//	{
//		printf("best_sim: %s, sim = %lf\n", arr[unique[best_pssm_index]]->degen_oligo, best_sim);
//		create_logo(arr[unique[best_pssm_index]], expected_consensus, p_IO);
//	}
	fprintf(fp ,"\n\n");
	fclose(fp);
	fclose(fp_sim);
//	fclose(fp_clust);
}

void output_merged_clusters(d_gs **arr, clustered_motifs *cm, int num, int *unique, int num_unique_motifs, motif_score_sign sign, params *p, IO_params *p_IO)
{
	int i,j,k, ii;
	int num_clusters = cm->num_clusters;

	// get some data
	int sign_multiplier = (sign == positive ? 1:-1);
	bool non_empty_flag = false;

	int top_cluster_members[num_unique_motifs];
	int ind_top = 0;
	for(i=0;i<num_unique_motifs;i++)
	{
		if(cm->c_index[unique[i]] == ind_top)
		{
			top_cluster_members[ind_top] = unique[i];
			ind_top++;
		}
	}

	// write the file
	FILE *fp = open_file(p_IO->out_summary, "a");
	// print some initial info
	//fprintf(fp, "Motif cluster     Cluster score (cumulative)     Previous clusters\n");
	fprintf(fp, "Motif cluster   Top k-mer score   Cluster score\n");
	for(i=0; i<num_clusters; i++)
	{
		int top_cluster_motif_index = find_top_cluster_motif_index(cm->c_index, unique, num_unique_motifs, i);
		//fprintf(fp, "%i                 %8.2lf                        %s\n", i+1, cm->cluster_scores[i], cm->merged_clusters[i]);
		fprintf(fp, "%i             %8.2lf             %8.2lf\n", i+1, sign_multiplier*arr[top_cluster_motif_index]->score, cm->cluster_scores[i]);
	}

	fprintf(fp, "\nMotif cluster     motif(k-mer)     score(#targets, %%targets, %%cumulative)\n");

	int d = 17;

	int *explained_peaks_cumulative = malloc(sizeof(int) * p->d.num_seq_level_stats);
	int *explained_peaks_current = malloc(sizeof(int) * p->d.num_seq_level_stats);

	// intialize explained cumulative peaks count for all clusters
	for(i=0;i<p->d.num_seq_level_stats;i++)
		explained_peaks_cumulative[i] = 0;

	bool first_i_equal_zero = true;
	for(ii=0;ii<num_clusters;ii++)
	{
		int top_cluster_motif_index = find_top_cluster_motif_index(cm->c_index, unique, num_unique_motifs, ii);
		if(top_cluster_motif_index == -1)
			break;
		double top_cluster_score = arr[top_cluster_motif_index]->score;
		double score_threshold = -top_cluster_score * p->cluster_pssm_threshold_fraction;

		bool first = true;
		int extra_offset = -20;
		char *cluster_label_oligo;
		int cluster_label_num_genes;
		char dummy[10];

		double prev_fraction_cum_curr = 0;

		// intialize explained peaks count for current cluster
		int num_peaks_curr = 0;
		for(j=0;j<num_unique_motifs;j++)
		{
			if
			(
				cm->c_index[unique[j]] == ii			 &&
				p_val[unique[j]] <= p->hypegeom_p_val_cutoff     &&
				-arr[unique[j]]->score >= score_threshold
			)
			{

				if(random_index > 0)
				{
//					printf("random_index = %i\n", random_index);
					int num_top_random_scores = random_index;
					double observed_score = arr[unique[j]]->score*sign_multiplier;
					arr[unique[j]]->p_value = estimate_Gaussian_fit_pval(top_random_scores, num_top_random_scores, observed_score);

					double pval_cutoff = 1.0;
					bool fails_to_reach_a_pval_cutoff = arr[unique[j]]->p_value > pval_cutoff;
					if(fails_to_reach_a_pval_cutoff == true)
					{
						continue;
					}
				}

				if(first == true)
				{
					// intialize explained peaks count for current cluster
					for(i=0;i<p->d.num_seq_level_stats;i++)
						explained_peaks_current[i] = 0;
					num_peaks_curr = 0;

					if(arr[unique[j]]->score*sign_multiplier > 0)
					{
						// offset due to cluster number printed at the start of the cluster label line
						extra_offset = sprintf(dummy, "%i", ii+1);
						fprintf(fp, "%i", ii+1);
					}
					cluster_label_oligo = arr[unique[j]]->degen_oligo;
					cluster_label_num_genes = arr[unique[j]]->num_genes;
					first = false;
				}
				else
				{
					bool fails_to_target_enough_seq_regions = false;//arr[unique[j]]->num_genes < 0.5 * cluster_label_num_genes;
					bool fails_to_co_occurr_in_enough_seq_regions = p_val[unique[j]] > p->hypegeom_p_val_cutoff;

					if(
						fails_to_co_occurr_in_enough_seq_regions == true
						||  fails_to_target_enough_seq_regions   == true
					  )
						continue;
				}

				int leading_edge_index, leading_edge_num_genes;
				double ks_score = ks_ES(arr[unique[j]]->genes, arr[unique[j]]->num_genes, p->d.num_seq_level_stats, &leading_edge_index, &leading_edge_num_genes);
				int total_offset = 35;
				// if score has the right sign
				if(arr[unique[j]]->score*sign_multiplier > 0)
				{
					bool first = false;
					if(num_peaks_curr == 0)
						first = true;
					for(i=0;i<arr[unique[j]]->num_genes;i++)
					{
						if(explained_peaks_current[arr[unique[j]]->genes[i]] == 0)
							num_peaks_curr+=1;
						explained_peaks_current[arr[unique[j]]->genes[i]] += 1;
						explained_peaks_cumulative[arr[unique[j]]->genes[i]] += 1;
					}
					double fraction_of_total = (1.0 * arr[unique[j]]->num_genes) / p->d.num_seq_level_stats;
					double fraction_cum_curr = (1.0 * num_peaks_curr) / p->d.num_seq_level_stats;
					if(true)//fraction_cum_curr - prev_fraction_cum_curr > 0.0)
					{
						non_empty_flag = true;

						int offset, balance_offset;
						if(strlen(cluster_label_oligo) > arr[unique[j]]->oligo_length)
							offset = d + optimal_offsets[unique[j]] + extra_offset;
						else
							offset = d - optimal_offsets[unique[j]] + extra_offset;
						balance_offset = total_offset - strlen(arr[unique[j]]->degen_oligo) - offset;

						if(eq(arr[unique[j]]->degen_oligo, cluster_label_oligo) == true)
						{
							offset = d;
							balance_offset = total_offset - strlen(arr[unique[j]]->degen_oligo) - offset - extra_offset;
						}
						print_spaces(fp, offset);

						if(rev_compl_flags[unique[j]] == true)
						{
							char *rev_compl_oligo = rev_compl(arr[unique[j]]->degen_oligo);
							fprintf(fp, "%+s", rev_compl_oligo);
							FREE(rev_compl_oligo);
						}
						else
							fprintf(fp, "%+s", arr[unique[j]]->degen_oligo);
						print_spaces(fp, balance_offset);

						fprintf(fp, "%4.2lf(%i, %i, %i)", -arr[unique[j]]->score, arr[unique[j]]->num_genes,
										(int)(100 * fraction_of_total), (int)(100 * fraction_cum_curr));
						if(p->known_pssm != 0)
						{
							double max_sim;
							gsl_matrix *tmp;
							if(ii < p_IO->num_motif_clusters_to_output)// && p->ensemble_pssm[ii] != 0)
							{
								tmp = arr[unique[j]]->pssm;
							}

							int required_core_length = p->required_core_length; int offset_top_score;
							double similarity1 = eval_PSSM_similarity(arr[unique[j]]->pssm, p->known_pssm, required_core_length, &offset_top_score);
							double similarity2 = similarity1;
							if(consider_rev_compl_flag == true)
							{
								gsl_matrix *rev_compl_pssm = get_rev_compl_pssm(arr[unique[j]]->pssm);
								similarity2 = eval_PSSM_similarity(rev_compl_pssm, p->known_pssm, required_core_length, &offset_top_score);
								gsl_matrix_free(rev_compl_pssm);
							}
							char *known = "literature: ";
							max_sim = MAX(similarity1, similarity2);
							double entropy_threshold = 1.0;

							fprintf(fp ,"%+*s%- ", strlen(known) + 2, known, max_sim);
						}
						if(arr[unique[j]]->p_value >= 0)
							fprintf(fp, "          % 10.8lf", arr[unique[j]]->p_value);
						fprintf(fp, "\n");
						prev_fraction_cum_curr = fraction_cum_curr;
					}
				}
			}
		}
		if(non_empty_flag == true)
		{
			int num_explained_cum = 0;
			for(i=0;i<p->d.num_seq_level_stats;i++)
			{
				if(explained_peaks_cumulative[i] > 0)
					num_explained_cum += 1;
			}
			double fraction_of_total = (1.0 * num_explained_cum) / p->d.num_seq_level_stats;
			fprintf(fp ,"cumulative \% = %i\n", (int)(100 *fraction_of_total) );fflush(fp);

			non_empty_flag = false;
		}
		fflush(fp);
		if(ii == 1)
			first_i_equal_zero = false;
	}

	free(explained_peaks_cumulative);
	free(explained_peaks_current);
	fclose(fp);
}

// write some txt files with the results after clustering the motifs
void generate_summary_output(d_gs **arr, int num_evolved, int total_num_seq, d_gs *expected, int *c_index, 
							double sim_score_top_ensemble_pssm_to_known_pssm, motif_score_sign sign, params *p, IO_params *p_IO)
//output_motif_clusters(d_gs **arr, int num, int *c_index, d_gs *expected, int *indexes_of_unique_motifs, int num_unique_motifs, motif_score_sign sign)
//void output_full_pssm(d_gs **arr, int num_evolved, int total_num_seq, d_gs *expected, int *c_index, params *p, IO_params *p_IO)
{
	int i;

	// 'num_evolved' contains the number of non-ZERO scoring motifs
	bool any_valid_prediction;

	if(num_evolved == 0)
		any_valid_prediction = false;
	else
		any_valid_prediction = true;

	int *indexes_of_unique_motifs; int num_unique_motif_predictions;
	extract_unique_motifs(arr, num_evolved, &indexes_of_unique_motifs, &num_unique_motif_predictions);

	int num_clusters = get_num_clusters(c_index, num_evolved);

	FILE *fp = open_file(p_IO->out_summary, "a");
	if(expected != 0)
	{
		fprintf(fp, "\n");
		fclose(fp);
	}

	if(any_valid_prediction == true)
	{
		output_motif_clusters(arr, num_evolved, c_index, indexes_of_unique_motifs, num_unique_motif_predictions, sim_score_top_ensemble_pssm_to_known_pssm, sign, p, p_IO, expected);

		for(i=0;i<p_IO->num_motif_clusters_to_output;i++)
		{
//			printf("%i: 'gsl_matrix_free(ensemble_pssm);'\n", i);fflush(stdout);
			gsl_matrix_free(p->ensemble_pssm[i]);
	//		p->ensemble_pssm[i] = 0;
		}

/*
		// used in significance estimate
		int num = num_unique_motif_predictions < num_top_scores ? num_unique_motif_predictions : num_top_scores;
		num_top_scores = num;
		for(i=0;i<num_top_scores;i++)
			top_scores[i] = -arr[indexes_of_unique_motifs[i]]->score;
*/
	}
}

// write a txt file with the results of merging
void generate_merging_output(d_gs **arr, clustered_motifs *cm, int num_evolved, motif_score_sign sign, params *p, IO_params *p_IO)
{
	int i;
	int num_clusters = get_num_clusters(cm->c_index, num_evolved);
	// 'num_evolved' contains the number of non-ZERO scoring motifs
	bool any_valid_prediction;
	if(num_evolved == 0)
		any_valid_prediction = false;
	else
		any_valid_prediction = true;

	int *indexes_of_unique_motifs; int num_unique_motif_predictions;
	extract_unique_motifs(arr, num_evolved, &indexes_of_unique_motifs, &num_unique_motif_predictions);

//	FILE *fp = open_file(p_IO->out_summary, "a");
	if(any_valid_prediction == true)
	{
		output_merged_clusters(arr, cm, num_evolved, indexes_of_unique_motifs, num_unique_motif_predictions, sign, p, p_IO);
		for(i=0;i<p_IO->num_motif_clusters_to_output;i++)
		{
			gsl_matrix_free(p->ensemble_pssm[i]);
		}
	}
}

// write a txt file with all evolved motifs and their scores, #targets, %targets, cluster
void output_evolved_motifs(d_gs **arr, int num_evolved, int total_num_seq, int *c_index, motif_score_sign sign, params *p, IO_params *p_IO)
{
	int i;

	// 'num_evolved' contains the number of non-ZERO scoring motifs
	bool any_valid_prediction;

	if(num_evolved == 0)
		any_valid_prediction = false;
	else
		any_valid_prediction = true;

	int *unique_motifs; int num_unique_motifs;
	extract_unique_motifs(arr, num_evolved, &unique_motifs, &num_unique_motifs);
	printf("# unique evolved motifs = %i\n", num_unique_motifs);fflush(stdout);

	if(any_valid_prediction == true)
	{
		FILE *fp = open_file(p_IO->out_extra, "w");
		fprintf(fp, "All evolved motifs\tScore\t#targets\t%%targets\tMotif cluster\n");

		int sign_multiplier = (sign == positive ? 1:-1);
		bool non_empty_flag = false;

		for(i=0;i<num_unique_motifs;i++)
		{
			//printf("%+s  %4.2lf\n", arr[unique[i]]->degen_oligo, sign_multiplier*arr[unique[i]]->score);fflush(stdout);
			double fraction_of_total = (1.0 * arr[unique_motifs[i]]->num_genes) / p->d.num_seq_level_stats;
			fprintf(fp, "%+s \t%4.2lf \t%i \t%i\t%i\n", arr[unique_motifs[i]]->degen_oligo, sign_multiplier*arr[unique_motifs[i]]->score, arr[unique_motifs[i]]->num_genes, (int)(100 * fraction_of_total), 1+c_index[unique_motifs[i]]);
		}
		fflush(fp);
		fclose(fp);
	}
}

// append in a txt file the similarity matrix between all evolved motifs
output_similarity_matrix(d_gs **arr, int num_evolved, params *p, IO_params *p_IO)
{
	int i, j, offset_top_score;
	double sim;
	bool rev_compl;
	// unique motifs
	int *unique; int num_unique_motifs;
	extract_unique_motifs(arr, num_evolved, &unique, &num_unique_motifs);
	// open file
	FILE *fp = open_file(p_IO->out_extra, "a");
	// write first matrix line
	fprintf(fp, "\nSimilarity ");
	for(i=0;i<num_unique_motifs;i++)
		fprintf(fp, "%+s ",arr[unique[i]]->degen_oligo);
	fprintf(fp, "\n");
	// compute and write similarity
	if (num_evolved > 0)
	{
		for(i=0;i<num_unique_motifs;i++)
		{
			fprintf(fp, "%+s ",arr[unique[i]]->degen_oligo);
			for(j=0;j<=i;j++)
			{
				sim = get_similarity(arr[unique[j]], arr[unique[i]], &rev_compl, p->required_core_length, &offset_top_score);
				//printf("%+s %+s %4.2lf %i\n", arr[unique[i]]->degen_oligo, arr[unique[j]]->degen_oligo, sim, offset_top_score);
				fprintf(fp, "%4.2lf ", sim);
			}
			fprintf(fp, "\n");
		}
	}

}
