/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef MOTIF_FINDER_H_
#define MOTIF_FINDER_H_

#include <unistd.h>
#include <sys/types.h>

#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>

#include <gsl/gsl_cdf.h>
#include <gsl/gsl_randist.h>

#include "interface.h"
#include "sort.h"
#include "utilities.h"
#include "tree.h"
#include "motif_occurrences.h"
#include "parameters.h"
#include "degen_gene_set.h"
#include "PSSM.h"
#include "eigsvd.h"
#include "knn/my_knn.h"
#include "hypergeometric.h"

#define oops(s) { perror((s)); exit(EXIT_FAILURE); }
#define MALLOC(s,t) if(((s) = malloc(t)) == 0) { oops("error: malloc() ");}
#define FREE(p) {free((void *)p);p=0;}
#define MAX(x,y)        ((x) > (y) ? (x) : (y))
#define MIN(x,y)        ((x) < (y) ? (x) : (y))
#define ABS(x)   	    ((x) >  0  ? (x) : (-x))

typedef struct
{
	int *cluster_indexes;
	int num;
}
cluster;

typedef struct
{
	int *c_index, *indexes_of_unique_motifs;
	d_gs **motifs;
	int num_unique_motifs, total_num_motifs, num_clusters;
	// num_clusters = number of clusters after merging original clusters
	double sim_score_top_ensemble_pssm_to_known_pssm;
	double *cluster_scores; // scores of merged clusters
	char **merged_clusters; // text info about merged clusters: which original clusters they contain
}
clustered_motifs;


void cluster_motifs(d_gs **dg_evolved, int num_evolved, clustered_motifs *cm, params *p);


void count_ovelapping_gene_offsets(d_gs *motif1, d_gs *motif2, void **index, int num_seq, int seq_length,
					int *num_offset_specific_gene_overlaps, int *total_num_gene_overlaps);

void output_full_pssm(d_gs **arr, int num_evolved, int total_num_seq, d_gs *expected_consensus, int *c_index, params *p, IO_params *p_IO);


bool run_motif_finder(params *parameters, IO_params *p_IO);

//void extract_unique_motifs(d_gs **arr, int num, int **indexes_of_unique_motifs, int *num_unique);

//void cluster_motifs(d_gs **dg_evolved, int num_evolved, params *p);

#endif /*MOTIF_FINDER_H_*/
