/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "motif_occurrences.h"


///////////////////////////////////////////////////////////////////////////////////////////////
// internal functions used by 'degen_oligo_occurrences_init' and 'degen_oligo_occurrences_free'
///////////////////////////////////////////////////////////////////////////////////////////////

void exact_oligo_occurrences_init(exact_oligo_occ *exact_occ)
{
	exact_occ->num_locations = 0;
}

void exact_oligo_occurrences_free(exact_oligo_occ *exact_occ)
{
	FREE(exact_occ->locations);
	exact_occ->num_locations = 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// internal functions used by 'degen_oligo_occurrences_init' and 'degen_oligo_occurrences_free'
///////////////////////////////////////////////////////////////////////////////////////////////

void degen_oligo_occurrences_init_new(degen_oligo_occurrences *dg_occ, int num_instantiations, int num_seq_regions)
{
	int i;
	if(num_seq_regions <= 0 || num_instantiations <= 0)
	{
		printf("ERROR: (motif_occurrences.c) 'degen_oligo_occurrences_init', number of num_seq_regions (%i) and num_instantiations (%i) must be > 0! num = %i\n", 
							num_seq_regions, num_instantiations);
		exit(1);
	}

	MALLOC(dg_occ->match_info, sizeof(region_match_info) * num_seq_regions);
	for(i=0;i<num_seq_regions;i++)
	{
		dg_occ->match_info[i].index_start_offset = dg_occ->match_info[i].index_last_offset = -1;
		dg_occ->match_info[i].num_occ_in_region = 0;
	}

	dg_occ->num_exact = num_instantiations;
	MALLOC(dg_occ->exact_occ, sizeof(exact_oligo_occ) * dg_occ->num_exact);
	for(i=0;i<dg_occ->num_exact;i++)	
		exact_oligo_occurrences_init(&dg_occ->exact_occ[i]);
}

void degen_oligo_occurrences_init(degen_oligo_occurrences *dg_occ, int num_instantiations, int num_seq_regions)
{
	degen_oligo_occurrences_init_new(dg_occ, num_instantiations, num_seq_regions);
	return;
/*
	int i;
	if(num <= 0)
	{
		printf("ERROR: (motif_occurrences.c) 'degen_oligo_occurrences_init', number of occurrences must be > 0! num = %i\n", num);
		exit(1);
	}
	dg_occ->num_occ = num;
	MALLOC(dg_occ->occ, sizeof(occurrences) * dg_occ->num_occ);
	for(i=0;i<dg_occ->num_occ;i++)	
		occurrences_init(&dg_occ->occ[i]);
*/
}

void degen_oligo_occurrences_free_new(degen_oligo_occurrences *dg_occ)
{
	int i;
	for(i=0;i<dg_occ->num_exact;i++)	
		exact_oligo_occurrences_free(&dg_occ->exact_occ[i]);
	FREE(dg_occ->exact_occ);
	dg_occ->num_exact = 0;
	FREE(dg_occ->match_info);
}


void degen_oligo_occurrences_free(degen_oligo_occurrences *dg_occ)
{
	degen_oligo_occurrences_free_new(dg_occ);
	return;
/*	
	int i;
	for(i=0;i<dg_occ->num_occ;i++)	
		occurrences_free(&dg_occ->occ[i]);
	FREE(dg_occ->occ);
	dg_occ->num_occ = 0;

	FREE(dg_occ->region_indexes);
	FREE(dg_occ->within_region_offsets);	
	FREE(dg_occ->num_occ_per_region);
*/
}


//void get_occ_location(unsigned long **occ, unsigned long numocc, unsigned long occ_index, int seq_region_length, int *occ_region_index_out, int *occ_region_offset_out)
void get_occ_location_info(exact_oligo_occ *exact_occ, unsigned long occ_index, int seq_region_length, int *occ_region_index_out, int *occ_region_offset_out)
{
	if(exact_occ == 0)
	{
		printf("ERROR: (motif_occurrences.c) 'get_occ_location', occurrences = 0\n");
		exit(1);
	}
	if(occ_index > exact_occ->num_locations)
	{
		printf("ERROR: (motif_occurrences.c) 'get_occ_location', index out of bound! numocc = %i\tocc_index= % i\n", exact_occ->num_locations, occ_index);
		exit(1);
	}
	*occ_region_index_out = exact_occ->locations[occ_index] / seq_region_length;
	if( exact_occ->locations[occ_index] % seq_region_length < (seq_region_length / 2) )
		*occ_region_offset_out = exact_occ->locations[occ_index] % seq_region_length;	
	else
		*occ_region_offset_out = seq_region_length - exact_occ->locations[occ_index] % seq_region_length;
}


void find_degen_oligo_set_occurrences_ALL(void **index, unsigned char **oligos, int num_oligos, int seq_length, int num_sequences, degen_oligo_occurrences *dg_occ_out)
{
	int i;
	unsigned long occ_index;
	unsigned char *pattern;
	unsigned long pattern_length;
	for(i=0;i<num_oligos;i++)
	{	
		pattern = oligos[i]; pattern_length = strlen(pattern);
//		printf("%s (len = %i)\n", oligos[i], pattern_length); fflush(stdout);
		if(locate(*index, pattern, pattern_length, &dg_occ_out->exact_occ[i].locations, &dg_occ_out->exact_occ[i].num_locations) != 0)
		{
			printf("(motif_occurrences.c) 'find_degen_oligo_set_locations': couldn't find %s!\n", oligos[i]);fflush(stdout);
			continue;
		}
//		printf("o[%i]->numocc = %i, seq_length = %i\n", i, dg_occ_out->occ[i].num_locations, seq_length);

		int occ_region_index_out, occ_region_offset_out; 
		for(occ_index=0;occ_index<dg_occ_out->exact_occ[i].num_locations;occ_index++)
		{
			get_occ_location_info(&dg_occ_out->exact_occ[i], occ_index, seq_length, &occ_region_index_out, &occ_region_offset_out);
			if(
 			   dg_occ_out->match_info[occ_region_index_out].index_start_offset == -1 || 
			   dg_occ_out->match_info[occ_region_index_out].index_start_offset > occ_region_offset_out
 			   )
				dg_occ_out->match_info[occ_region_index_out].index_start_offset = occ_region_offset_out;

			if(
 			   dg_occ_out->match_info[occ_region_index_out].index_last_offset == -1 || 
			   dg_occ_out->match_info[occ_region_index_out].index_last_offset < occ_region_offset_out
 			   )
				dg_occ_out->match_info[occ_region_index_out].index_last_offset = occ_region_offset_out;
			dg_occ_out->match_info[occ_region_index_out].num_occ_in_region++;
		}
	}
}



void find_degen_oligo_set_locations(void **index, unsigned char **oligos, int num_oligos, int seq_length, int num_sequences, motif_occurrences *occ_out)
{
	MALLOC(occ_out->num_occ_per_region, sizeof(int) * num_sequences);
	int i,j;
	unsigned char *pattern;
	unsigned long pattern_length;
	 // including overlaps!
	unsigned long numocc;
	unsigned long **occ;
	MALLOC(occ, sizeof(unsigned long *));

	int indexes[num_sequences];
	int offsets[num_sequences];
	for(i=0;i<num_sequences;i++)
	{
		indexes[i] = 0;
		occ_out->num_occ_per_region[i] = 0;
	}

	for(i=0;i<num_oligos;i++)
	{
		pattern = oligos[i]; pattern_length = strlen(pattern);
//		printf("%s (len = %i)\n", oligos[i], pattern_length); fflush(stdout);
		if(locate(*index, pattern, pattern_length, occ, &numocc) != 0)
		{
			printf("(motif_occurrences.c) 'find_degen_oligo_set_locations': couldn't find %s!\n", oligos[i]);
			FREE(*occ);
			continue;
		}
//		printf("numocc = %i\n", numocc); fflush(stdout);
		for(j=0;j<numocc;j++)
		{
			int seq_index = occ[0][j] / seq_length;
			int seq_offset = occ[0][j] % (seq_length / 2);
			if(seq_index < 0 || seq_index >= num_sequences)
			{
				printf("ERROR: (motif_occurrences.c) 'find_degen_oligo_set_locations' array out of bounds! %li \n", occ[0][j]/seq_length);
				printf("occ[0][%i]=%i\t%i\n", j, occ[0][j], seq_length);
				exit(1);
			}
			indexes[seq_index] = 1;
			offsets[seq_index] = seq_offset;
			occ_out->num_occ_per_region[seq_index]++;
		}
		FREE(*occ);
	}
	FREE(occ);

	// count number of occurrences
	occ_out->num_occurrences = 0;
	for(i=0;i<num_sequences;i++)
		if(indexes[i] == 1)
			occ_out->num_occurrences++;

	occ_out->region_indexes = 0;
	occ_out->within_region_offsets = 0;
	if(occ_out->num_occurrences > 0)
	{
		// initialize sequence index & offset arrays
		MALLOC(occ_out->region_indexes,        sizeof(int) * occ_out->num_occurrences);
		MALLOC(occ_out->within_region_offsets, sizeof(int) * occ_out->num_occurrences);
		int ind = 0;	
		for(i=0;i<num_sequences;i++)
		{
			if(indexes[i] == 1)
			{
				occ_out->region_indexes[ind] = i;
				occ_out->within_region_offsets[ind] = offsets[i];
				ind++;
			}
		}
	}
}

void free_motif_occurrences(motif_occurrences *occ)
{
	FREE(occ->region_indexes);
	FREE(occ->within_region_offsets);	
	FREE(occ->num_occ_per_region);
}


void output_motif_occ_locations(char *out_file, motif_occurrences *occ)
{
	int i;
	FILE *fp = open_file(out_file, "a");
	for(i=0;i<occ->num_occurrences;i++)
	{
		fprintf(fp, "%i:%i", occ->region_indexes[i], occ->within_region_offsets[i]);
		if(i<occ->num_occurrences-1)
			fprintf(fp,",");
	}
	fprintf(fp, "\n");
	fclose(fp);
}


void output_motif_geneset(char *out_file, motif_occurrences *occ, params *p)
{
	int i;
	FILE *fp = open_file(out_file, "a");
	char *seq_id;
	for(i=0;i<occ->num_occurrences;i++)
	{
//		if(i<occ->num_occurrences-1)
//		fprintf(fp,"\t");
		seq_id = p->d.seq.seq_ids[occ->region_indexes[i]];
		fprintf(fp, "\t%s", seq_id);
	}
	fprintf(fp, "\n");
	fclose(fp);
}

void output_motif_target_set(d_gs *motif, params *p, char *out_motif_occ_file, char *out_motif_genesets_file)
{
	unsigned char **oligos = motif->instantiations;
	int num_oligos = motif->instantiations_length;
	motif_occurrences occ;
	find_degen_oligo_set_locations(p->d.seq.index, oligos, num_oligos, p->d.seq.seq_length, p->d.num_seq_level_stats, &occ);

	// output occurrence index + offset
	FILE *fp_motif_occ = fopen(out_motif_occ_file, "a");
	fprintf(fp_motif_occ, "%-20s(%-4i)", motif->degen_oligo, motif->num_genes);
	fclose(fp_motif_occ);
	output_motif_occ_locations(out_motif_occ_file, &occ);
//	fp_motif_occ = fopen(out_motif_occ_file, "a");
//	fprintf(fp_motif_occ, "\n");
//	fclose(fp_motif_occ);

	// output genset member ids
	FILE *fp_motif_genesets = fopen(out_motif_genesets_file, "a");
	fprintf(fp_motif_genesets, "%s(num_targets=%i)\t", motif->degen_oligo, motif->num_genes);
	fprintf(fp_motif_genesets, "%s(num_targets=%i)", motif->degen_oligo, motif->num_genes);
	fclose(fp_motif_genesets);
	output_motif_geneset(out_motif_genesets_file, &occ, p);
//	fp_motif_genesets = fopen(out_motif_genesets_file, "a");
//	fprintf(fp_motif_genesets, "\n");
//	fclose(fp_motif_genesets);


	free_motif_occurrences(&occ);
}


