/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "params_reader.h"

void init_significance(params *parameters)
{
	parameters->significance.evaluate_significance		   	   = false;//true;// output file name is initialized in 'analysis.c'
	parameters->significance.num_random_permutations		   = 1;
	parameters->significance.num_random_permutations_remaining = parameters->significance.num_random_permutations;
}

double p_val_calib_threshold;


species species_type_global;

int length_threshold;

void init_adaptive_step_size_control(adaptive_step_size_control *adpt_stp_sz)
{
	adpt_stp_sz->step_size                 = 5;
	adpt_stp_sz->increment_step_size       = 5;
	adpt_stp_sz->decrement_ratio_step_size = 2.0;
	adpt_stp_sz->max_step_size             = 50;
	adpt_stp_sz->ratio_lower_threshold     = 1.05;
	adpt_stp_sz->ratio_upper_threshold     = 1.15;
	adpt_stp_sz->ratio_type                = scale;
}


void init_tuning_params(params *p, IO_params *p_IO)
{

	init_adaptive_step_size_control(&p->scoring_params.step_size_control_params);

	p->select_next_greedy = select_next_motif_top_scoring;

	// random set + exact null
	p->geneset_ES = geneset_random_set_exact_null_ES;
	p->scoring_params.pre_calculate_location_and_scale = 0;//estimate_global_MEAN_and_VAR_random_set;//estimate_global_MEAN_and_VAR_KS_enrichment_score;//estimate_global_MEDIAN_and_IQR_random_set;//

	p->scoring_params.num_draws 		              = 500;
	p->scoring_params.pre_calculate_global_datasets_location_and_scale = estimate_global_MEAN_and_VAR;//estimate_global_MEDIAN_and_IQR;//

	p->num_neighbours			              = 10;

	int i;
	for(i=0;i<p_IO->num_motif_clusters_to_output;i++)
		p->ensemble_pssm[i] = 0;

//	species_type_global = human;//yeast;//arabidopsis;//
	init_background_freq(species_type_global);
}

void init_common(params *parameters, IO_params *p_IO, char *oligo_seeds_file)
{
	parameters->use_conservation = false;

	read_oligos(oligo_seeds_file, &parameters->oligo_seeds);
//	printf("oligo_seeds_file: %s\n\n", oligo_seeds_file);
	parameters->kNN_oligos_file = "oligos_size_6_only_2048";//"oligos_size_5";//"kNN_oligos";

	parameters->d.seq.delim			 = '$';

	parameters->d.seq.is_initialized	 = false;
	parameters->d.expr.delim		 = "\t";
	parameters->d.expr.is_initialized	 = false;
	parameters->d.is_initialized		 = false;
//	init_significance(parameters);
}


double pecentage_rank_cutoff_global;



void print_params(params *p, IO_params *p_IO, FILE *fp)
{
	fprintf(fp, "min_motif_length = %i\n", p->min_motif_length);
	fprintf(fp, "max_motif_length = %i\n", p->max_motif_length);

	fprintf(fp, "p->required_core_length = %i\n", p->required_core_length);
	fprintf(fp, "p->pssm_crop_threshold = %lf\n", p->pssm_crop_threshold);

	fprintf(fp, "p->cluster_pssm_threshold_fraction = %lf\n", p->cluster_pssm_threshold_fraction);
	fprintf(fp, "p->cluster_sim_threshold = %lf\n", p->cluster_sim_threshold);

	fprintf(fp, "p->fraction_of_top_sequences_to_consider = %lf\n", p->fraction_of_top_sequences_to_consider);
	fprintf(fp, "p->hypegeom_p_val_cutoff = %g\n", p->hypegeom_p_val_cutoff);

	fprintf(fp, "p_IO->num_motif_clusters_to_output = %i\n", p_IO->num_motif_clusters_to_output);
	fprintf(fp, "p->degen_threshold = %lf\n", p->degen_threshold);

	fprintf(fp, "p->fraction_degen_pos_threshold = %lf\n", p->fraction_degen_pos_threshold);
	fprintf(fp, "p->scoring_params.max_gene_set_size_percentage_threshold = %lf\n", p->scoring_params.max_gene_set_size_percentage_threshold);
	fprintf(fp, "p->scoring_params.min_gene_set_size_absolute_threshold = %i\n", p->scoring_params.min_gene_set_size_absolute_threshold);
	fprintf(fp, "p->scoring_params.min_gene_set_size_percentage_threshold = %lf\n", p->scoring_params.min_gene_set_size_percentage_threshold);
	fprintf(fp, "p->num_clusters_to_output_motif_occurrences_for = %i\n", p->num_clusters_to_output_motif_occurrences_for);

	fprintf(fp, "img_type = %s\n", p_IO->img_type);

	if(species_type_global == human)
		fprintf(fp, "species_type_global = human\n");
	else if(species_type_global == yeast)
		fprintf(fp, "species_type_global = yeast\n");
	else if(species_type_global == arabidopsis)
		fprintf(fp, "species_type_global = arabidopsis\n");
	else if(species_type_global == unspecified)
		fprintf(fp, "species_type_global = unspecified\n");
	else
	{
		fprintf(fp, "ERROR: unknown species_type_global!\n");
		exit(1);
	}
}

extern bool is_regression_initialized;

void read_input(int argc, char*argv[], params *p, IO_params *p_IO)
{
	if(argc >= 21)
	{
//		printf("is_regression_initialized = %i\n", is_regression_initialized);

		p_IO->score_file   = argv[1];
		p_IO->seq_file  = argv[2];

		char out_dir[500];
		if(argv[3][strlen(argv[3]) - 1] == '/')
			sprintf(out_dir, "%s", argv[3]);
		else
			sprintf(out_dir, "%s/", argv[3]);
		p_IO->output_dir       = out_dir;

		p_IO->input_seeds_file = argv[4];
		char *oligo_seeds_file = argv[4];

		p->min_motif_length = atoi(argv[5]);
		p->max_motif_length = atoi(argv[6]);


		// for aligning motifs during clustering (require at least 'required_core_length' overlap)
		p->required_core_length = atoi(argv[7]);
	
		// crop all flanking positions in the pssm that have <= pssm_crop_threshold * (# reads of "core")
		p->pssm_crop_threshold	= atof(argv[8]);
//		p_IO->trim_fraction     = atof(argv[8]);
	
		// (for cluster pssm construction) consider all cluster members that have score >= cluster_pssm_threshold_fraction * top_cluster_score
		p->cluster_pssm_threshold_fraction = atof(argv[9]);
	
		// required sequence similarity to assign to the same cluster (range: [0.0,1.0], identical <=> 1.0)
		p->cluster_sim_threshold = atof(argv[10]);//0.75
	
		// fraction of top sequences to use in the construction of the PSSM
		p->fraction_of_top_sequences_to_consider = atof(argv[11]);

		// for aligning motifs during clustering (require hypergeometric p-value for co-occurrence <= 'hypegeom_p_val_cutoff')
		p->hypegeom_p_val_cutoff = atof(argv[12]);//1.0e-30;
	
		// output option for the number of required top motif clusters (PSSM construction) 
		p_IO->num_motif_clusters_to_output = atoi(argv[13]);//5;

		// allowable 'degeneracy score' (range [0.0, 1.0] 0.0 => no degeneracy, 1.0 => all 'N's)
		p->degen_threshold = atof(argv[14]);//0.75;//1.0;//
	
		// allowable percentage of degenerate bases in a motif
		p->fraction_degen_pos_threshold	 = atof(argv[15]);//0.75;//1.0;//
	
//		#####################
//		# max target set size = max_gene_set_size_percentage_threshold * total_num_seq_regions)
//		#####################
		// maximum target size (as a percentage of total number of input sequence regions)
		p->scoring_params.max_gene_set_size_percentage_threshold = atof(argv[16]);//0.95;
	
//		#####################
//		# min target set size = MIN(min_gene_set_size_absolute_threshold, min_gene_set_size_percentage_threshold * total_num_seq_regions)
//		#####################
		// minimum geneset size (as an absolute number)
		p->scoring_params.min_gene_set_size_absolute_threshold   = atoi(argv[17]);//20;	
		// minimum geneset size (as a percentage of total # input sequence regions)
		p->scoring_params.min_gene_set_size_percentage_threshold = atof(argv[18]);//0.01;
	
		p->num_clusters_to_output_motif_occurrences_for = atoi(argv[19]);//3;

		char *species = argv[20];
		if(strcmp(species, "human") == 0)
			species_type_global = human;
		else if(strcmp(species, "yeast") == 0)
			species_type_global = yeast;
		else if(strcmp(species, "arabidopsis") == 0)
			species_type_global = arabidopsis;
		else
			species_type_global = unspecified;

		// image file type (eps, png, pdf)
		sprintf(p_IO->img_type, "%s", argv[31]);


		init_common(p, p_IO, oligo_seeds_file);
		init_tuning_params(p, p_IO);

/*		// print some info
		print_params(p, p_IO, stdout);
		printf("is_regression_initialized = %i\n", is_regression_initialized);
		getchar();*/

	}
	else
	{
		printf("Usage: main_out <input_params_file> -h|-y|-a -max_genset_size <upper bound on gene set size>\n");
		printf("-h = human sequence data\n-y = yeast sequence data\n-a = arabidopsis sequence data \n");
		exit(1);
	}
}

void free_parameters(params *p)
{
	int i;
	gsl_matrix_free(p->known_pssm);

	if(p->d.expr.is_initialized == true)
	{
		for(i=0;i<p->d.expr.num_genes;i++)
		{
			FREE(p->d.expr.genes[i].expr_profile);
			FREE(p->d.expr.genes[i].gene_id);
			FREE(p->d.expr.genes[i].seq);
		}
		FREE(p->d.expr.genes);
		FREE(p->d.expr.pattern_to_correlate_with);
		p->d.expr.is_initialized = false;
	}
}
