/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "preprocess.h"
/*
double sum_preprocess(double *vec, int num)
{
	int i;
	double sum = 0;	
	for(i=0;i<num;i++)
		sum += vec[i];
	return sum;
}
*/

double sum_squares(double *vec, int num)
{
	int i;
	double sum_squares = 0;
	for(i=0;i<num;i++)
		sum_squares += vec[i]*vec[i];
	return sum_squares;
}

void GSL_row_mean(gsl_matrix *mat, gsl_vector *mean_out)
{
	int i;
	gsl_vector_view row_view;
	gsl_vector_set_zero(mean_out);
	int num_rows = mat->size1;
	for(i=0;i<num_rows;i++)
	{
		row_view = gsl_matrix_row(mat, i);
		gsl_vector_add(mean_out, &row_view.vector);
	}
	gsl_vector_scale(mean_out, 1.0 / num_rows);
}

void GSL_col_mean(gsl_matrix *mat, gsl_vector *mean_out)
{
	int i;
	gsl_vector_view col_view;
	gsl_vector_set_zero(mean_out);
	int num_cols = mat->size2;
	for(i=0;i<num_cols;i++)
	{
		col_view = gsl_matrix_column(mat, i);
		gsl_vector_add(mean_out, &col_view.vector);
	}
	gsl_vector_scale(mean_out, 1.0 / num_cols);
}

void GSL_center_cols(gsl_matrix *mat)
{
	int i;
	int num_rows = mat->size1, num_cols = mat->size2;
	gsl_vector *row_mean = gsl_vector_alloc(num_cols);
	GSL_row_mean(mat, row_mean);
	gsl_vector_view row_view;
	for(i=0;i<num_rows;i++)
	{
		row_view = gsl_matrix_row(mat, i);
		gsl_vector_sub(&row_view.vector, row_mean);
	}
	gsl_vector_free(row_mean);
}

void GSL_center_rows(gsl_matrix *mat)
{
	int i;
	int num_rows = mat->size1, num_cols = mat->size2;
	gsl_vector *col_mean = gsl_vector_alloc(num_rows);
	GSL_col_mean(mat, col_mean);
	gsl_vector_view col_view;
	for(i=0;i<num_cols;i++)
	{
		col_view = gsl_matrix_column(mat, i);
		gsl_vector_sub(&col_view.vector, col_mean);
	}
	gsl_vector_free(col_mean);
}


// assume columns of matrix 'mat' have already been centered
void GSL_scale_cols_using_L2_norm(gsl_matrix *mat)
{
	int i;
	double ss;
	gsl_vector_view col_view;
	int num_cols = mat->size2;
	for(i=0;i<num_cols;i++)
	{
		col_view = gsl_matrix_column(mat, i);
		ss  = sum_squares(col_view.vector.data, col_view.vector.size);
		if(sqrt(ss) == 0)
			printf("WARNING: 'GSL_scale_cols_using_L2_norm', sqrt(sum_squares) == 0! column = %i\n", i);
		else
			gsl_vector_scale(&col_view.vector, 1.0 / sqrt(ss));
	}
}


// assume rows of matrix 'mat' have already been centered
void GSL_scale_rows_using_L2_norm(gsl_matrix *mat)
{
	int i;
	double ss;
	gsl_vector_view row_view;
	int num_rows = mat->size1;
	for(i=0;i<num_rows;i++)
	{
		row_view = gsl_matrix_row(mat, i);
		ss  = sum_squares(row_view.vector.data, row_view.vector.size);
		if(sqrt(ss) == 0)
			printf("WARNING: 'GSL_scale_rows_using_L2_norm', sqrt(sum_squares) == 0! row = %i\n", i);
		else
			gsl_vector_scale(&row_view.vector, 1.0 / sqrt(ss));
	}
}


// assume columns of matrix 'mat' have already been centered
void GSL_scale_cols_using_SD(gsl_matrix *mat)
{
//	printf("mat\n");
//	print_matrix_GSL_m_n(mat, 5, 5);

	int i, j;
	double ss, sd;
	gsl_vector_view col_view;
	int num_cols = mat->size2;

	double col[mat->size1];
	for(i=0;i<num_cols;i++)
	{
		for(j=0;j<mat->size1;j++)
		{
			col[j] = gsl_matrix_get(mat, j, i);//mat->data[j*mat->size2 + i];
//			printf("%lf, ", col[j]);
		}
//		printf("\n");

//		col_view = gsl_matrix_column(mat, i);
//		print_vector_n_GSL(&col_view.vector, 5);

//		ss = sum_squares(col_view.vector.data, col_view.vector.size);
//		sd = sqrt(ss / col_view.vector.size);
//		ss = sum_squares(col, mat->size1);
//		double sd1 = sqrt(ss / (mat->size1 - 1));
//		printf("sd1 = %lf\n", sd1);
		size_t stride = 1;
		sd = gsl_stats_sd(col, stride, mat->size1);
//		printf("sd(gsl_stats_sd) = %lf\n", sd);

//		printf("sd = %lf\n", sd);
		if(sd == 0)
			printf("WARNING: 'GSL_scale_cols_using_SD', sd == 0! column = %i\n", i);
		else
//			gsl_vector_scale(&col_view.vector, 1.0 / sd);
		{
			for(j=0;j<mat->size1;j++)
			{
//				mat->data[j*mat->size2 + i] = col[j] / sd;
				gsl_matrix_set(mat, j, i, col[j] / sd);
			}		
		}
	}
//	getchar();
}


// assume rows of matrix 'mat' have already been centered
void GSL_scale_rows_using_SD(gsl_matrix *mat)
{
	int i, j;
	double ss, sd;
	gsl_vector_view row_view;
	int num_row = mat->size1;

	double row[mat->size1];
	for(i=0;i<num_row;i++)
	{
		for(j=0;j<mat->size2;j++)
		{

			row[j] = gsl_matrix_get(mat, i, j);
//			row[j] = mat->data[i*mat->size2 + j];
//			printf("%lf, ", col[j]);
		}
		size_t stride = 1;
		sd = gsl_stats_sd(row, stride, mat->size2);

//		row_view = gsl_matrix_row(mat, i);
//		ss = sum_squares(row_view.vector.data, row_view.vector.size);
//		sd = sqrt(ss / row_view.vector.size);
		if(sd == 0)
			printf("WARNING: 'GSL_scale_rows_using_SD', sd == 0! row = %i\n", i);
		else
//			gsl_vector_scale(&row_view.vector, 1.0 / sd);
		{
			for(j=0;j<mat->size2;j++)
			{
				gsl_matrix_set(mat, i, j, row[j] / sd);
//				mat->data[i*mat->size2 + j] = row[j] / sd;
			}		
		}
	}
}

void GSL_center(gsl_matrix *mat, dimension dim)
{
	switch(dim)
	{
		case rows:
			GSL_center_rows(mat);
			break;
		case columns:
			GSL_center_cols(mat);
			break;
		default: printf("ERROR: 'GSL_center', invalid value for 'dim'!\n");exit(1);
	}
}

void GSL_scale(gsl_matrix *mat, dimension dim, scaling_type scaling)
{
	switch(dim)
	{
		case rows:
			switch(scaling)
			{
				// imeplemented 
				case StdDev: printf("StdDev\n"); GSL_scale_rows_using_SD(mat); break;
				case L2_norm: printf("L2_norm\n");GSL_scale_rows_using_L2_norm(mat); break;

				// to do
				case L1_norm: printf("L1_norm scaling NOT IMPLEMENTED YET!\n");exit(1);break;
				case L_Inf_norm: printf("L_Inf_norm scaling NOT IMPLEMENTED YET!\n");exit(1);break;
				case Frobenius: printf("Frobenius scaling NOT IMPLEMENTED YET!\n");exit(1);break;
				default: printf("ERROR: 'GSL_center', invalid value for 'scaling'!\n");exit(1);
			}
			break;
		case columns:
			switch(scaling)
			{
				// imeplemented 
				case StdDev: printf("StdDev\n"); GSL_scale_cols_using_SD(mat); break;
				case L2_norm: printf("L2_norm\n");GSL_scale_cols_using_L2_norm(mat); break;

				// to do
				case L1_norm: printf("L1_norm scaling NOT IMPLEMENTED YET!\n");exit(1);break;
				case L_Inf_norm: printf("L_Inf_norm scaling NOT IMPLEMENTED YET!\n");exit(1);break;
				case Frobenius: printf("Frobenius scaling NOT IMPLEMENTED YET!\n");exit(1);break;
				default: printf("ERROR: 'GSL_center', invalid value for 'scaling'!\n");exit(1);
			}
			break;
		default: printf("ERROR: 'GSL_center', invalid value for 'dim'!\n");exit(1);
	}
}

void GSL_center_and_scale(gsl_matrix *mat, dimension dim, scaling_type scaling)
{
	GSL_center(mat, dim);
	GSL_scale(mat, dim, scaling);
}

