/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "random_gen.h"

/*
Robert G. Brown	                       http://www.phy.duke.edu/~rgb/
Duke University Dept. of Physics, Box 90305
Durham, N.C. 27708-0305
Phone: 1-919-660-2567  Fax: 919-660-2525     email:rgb@phy.duke.edu
*/
unsigned long int random_seed()
{
	unsigned int seed;
	struct timeval tv;
	FILE *devrandom;
	
	if((devrandom = fopen("/dev/random","r")) == NULL)
	{
		gettimeofday(&tv,0);
		seed = tv.tv_sec + tv.tv_usec;
//		if(verbose == D_SEED) printf("Got seed %u from gettimeofday()\n",seed);
	} 
	else 
	{
		fread(&seed,sizeof(seed),1,devrandom);
//		if(verbose == D_SEED) printf("Got seed %u from /dev/random\n",seed);
		fclose(devrandom);
	}
	return(seed);
}


//static const gsl_rng_type *T;
gsl_rng *r1 = 0;


// call once at start of program
bool first_ran_uniform = true;
void init_ran_gen_static()
{
	if(first_ran_uniform == true)
	{
		gsl_rng_type *T = gsl_rng_gfsr4;//gsl_rng_ranlxs2;//
		r1 = gsl_rng_alloc(T);
		unsigned long int s = 12345;//time(123456);//time(NULL);//
//		unsigned long int s = random_seed();

		gsl_rng_set(r1, s);
       		first_ran_uniform = false;
//		printf("done\n");fflush(stdout); 
	}
}

gsl_rng *get_rng()
{
	init_ran_gen_static();
	return r1;
}


void clean_up_ran_gen_static()
{
	gsl_rng_free(r1);
}

/*
need to add 'rmv.c' which implements 'rwishart'
void ran_wishart(const int n,const int dof,const gsl_matrix *scale,gsl_matrix *result)
{
	init_ran_gen_static();
//	 Wishart distribution random number generator
//	
//		n	 	gives the dimension of the random matrix
//		dof	 	degrees of freedom
//		scale	scale matrix of dimension n x n
//		result	output variable with a single random matrix Wishart distribution generation
	int res = rwishart(r1,n,dof,scale,result);
}
*/

size_t draw_from_discrete_distrib(double *weights, int num)
{
	init_ran_gen_static();
	gsl_ran_discrete_t *discrete_random_num_generator = gsl_ran_discrete_preproc(num, weights);
	return gsl_ran_discrete(r1, discrete_random_num_generator);
}

// assumes no ties => all elements in 'v' are unique
void get_ranks_vector(gsl_vector *v, gsl_permutation *rank)
{
  	gsl_permutation *perm = gsl_permutation_alloc(v->size);
 	gsl_sort_vector_index(perm, v);
  	gsl_permutation_inverse(rank, perm);
	gsl_permutation_free(perm);
}

void sample_with_replacement(unsigned long *dest, unsigned long num_dest, unsigned long *src, unsigned long num_src)
{
	init_ran_gen_static();
	gsl_ran_sample(r1, dest, num_dest, src, num_src, sizeof(int));
}

void get_ranks(double *arr, size_t n, int **ranks)
{
	gsl_vector_view v = gsl_vector_view_array(arr,n);
	gsl_permutation *rank = gsl_permutation_alloc(n);
	get_ranks_vector(&v.vector, rank);
	MALLOC(*ranks, sizeof(int)*n);
	int i;
	for(i=0;i<n;i++)
		(*ranks)[i] = rank->data[i];
}

void sample_without_replacement(unsigned long *dest, unsigned long num_dest, unsigned long *src, unsigned long num_src)
{
	init_ran_gen_static();
	gsl_ran_choose(r1, dest, num_dest, src, num_src, sizeof(unsigned long));
}

void permute1(double *arr, size_t size, gsl_permutation *p)
{
	init_ran_gen_static();
//	gsl_permutation *p = gsl_permutation_alloc(size);
	gsl_permutation_init(p);
	gsl_ran_shuffle(r1, p->data, size, sizeof(size_t));
	
	size_t stride = 1;
	gsl_permute(p->data, arr, stride, size);
//	gsl_permutation_free(p);
}

void permute(double *arr, size_t size)
{
	init_ran_gen_static();
	gsl_permutation *p = gsl_permutation_alloc(size);
	gsl_permutation_init(p);
	gsl_ran_shuffle(r1, p->data, size, sizeof(size_t));
	
	size_t stride = 1;
	gsl_permute(p->data, arr, stride, size);
	gsl_permutation_free(p);
}


// generates an integer in the range [lower, upper - 1]
int sample(int lower, int upper)
{
	init_ran_gen_static();
	return lower + gsl_rng_get(r1) % (upper-lower);
}

void test_sample()
{
	int i;
	int l=3,u=9;
	for(i=0;i<10;i++)
		printf("%li\t", sample(l, u));
	getchar();
}

void ran_gaussian(double *dest, int num_dest, double mu, double sigma)
{
	init_ran_gen_static();
  	int i;
  	for(i=0;i<num_dest;i++)
   		dest[i] = gsl_ran_gaussian(r1, sigma) + mu;
}

void ran_t_dist(double *dest, int num_dest, double deg_freedom, double mean)
{
	init_ran_gen_static();
  	int i;
  	for(i=0;i<num_dest;i++)
   		dest[i] = gsl_ran_tdist(r1, deg_freedom) + mean;
}

// 'dest' needs to have allocated 'num' elements
void ran_uniform(double *dest, int num_dest, double lower, double upper)
{
	init_ran_gen_static();
  	int i;
 	for(i=0;i<num_dest;i++)
   	  	dest[i] = lower + gsl_rng_uniform(r1) * (upper-lower);
}

double ran_uniform_draw(double lower, double upper)
{
	init_ran_gen_static();
   	return lower + gsl_rng_uniform(r1) * (upper-lower);
}

unsigned int ran_bernoulli_draw(double p)
{
	init_ran_gen_static();
	return gsl_ran_bernoulli(r1, p);
}



double ran_uniform_single_value()
{
	init_ran_gen_static();
	return gsl_rng_uniform(r1);
}

void ran_uniform_integer_non_unique(unsigned long int *dest, unsigned long lower, unsigned long upper, int num)
{
	init_ran_gen_static();
   	int i;
  	for(i=0;i<num;i++)
   		dest[i]= lower + gsl_rng_get(r1) % (upper-lower);
}

// 'dest' needs to have allocated 'num' elements
void ran_beta(double *dest, int num_dest, double a, double b)
{
	init_ran_gen_static();
  	int i;
 	for(i=0;i<num_dest;i++)
   	  	dest[i] = gsl_ran_beta(r1, a, b);
}

// 'dest' needs to have allocated 'num' elements
void ran_gamma(double *dest, int num_dest, double a, double b)
{
	init_ran_gen_static();
  	int i;
 	for(i=0;i<num_dest;i++)
   	  	dest[i] = gsl_ran_gamma(r1, a, b);
}


///////////////////////////////////////////////////////
///////////////////////////////////////////////////////

// draw from an unnormalized multinomial distribution
int draw_unnormalized(double *scores, int num)
{
	int i;
	double sum = 0;
	for(i=0;i<num;i++)
		sum += scores[i];
	double draw = ran_uniform_draw(0, sum);
	for(i=0;i<num;i++)
	{
		sum -= scores[i];
		if(draw >= sum)
			return i;
	}
}

void test_draw_unnormalized()
{
	int i;
	double count_0 = 0, count_1 = 0, count_2 = 0, arr[3] = {1,1,1};
	int num_sim = 100000;
	for(i=0;i<num_sim;i++)
	{
		int draw = draw_unnormalized(arr, 3);
		switch(draw)
		{
			case 0:
				count_0 += 1;
				break;
			case 1:
				count_1 += 1;
				break;
			case 2:
				count_2 += 1;
				break;
			default:
				break;
		}
	}
	printf("%lf\t%lf\t%lf\n ", count_0, count_1, count_2);
	printf("%g\t%g\t%g\n ", count_0/num_sim, count_1/num_sim, count_2/num_sim);
	getchar();
}

void normalize(double *arr, int num)
{
	int i;
	double sum = 0;
	for(i=0;i<num;i++)
		sum += arr[i];
	for(i=0;i<num;i++)
		arr[i] = arr[i] / sum;
}


////////////////////////// TESTS ///////////////////////////////

void test_get_ranks()
{
	int n = 5;
	double arr[5] = {0,5,-2,20,3};
	int *ranks;
	get_ranks(arr, n, &ranks);
	int i=0;
	for(i=0;i<n;i++)
	{
		printf("%i: %lf\n", ranks[i], arr[i]);
	}
}

void test_sample_with_replacement()
{
	int num_src = 10, num_dest = 5;
	int a[num_dest], b[num_src];
    int i;
    for (i = 0; i < num_src; i++)
    	b[i] = i;
    sample_with_replacement (a, num_dest, b, num_src);
    for (i = 0; i < num_dest; i++)
   		printf("%i: %i\n", i, a[i]);
   	printf("done\n");
   	getchar();
}

void test_sample_without_replacement()
{
	int num_src = 10, num_dest = 5;
	int a[num_dest], b[num_src];
    int i;
    for (i = 0; i < num_src; i++)
    	b[i] = i;
    sample_without_replacement (a, num_dest, b, num_src);
    for (i = 0; i < num_dest; i++)
   		printf("%i: %i\n", i, a[i]);
   	printf("done\n");
   	getchar();
}

void test_permute()
{
	double arr[5] = {1.2,4.9,-4,0.3,2.5};
	int i;
 	   for(i=0;i<5;i++)
 	   	printf("%i: %lf\t", i, arr[i]);
	gsl_permutation *per = gsl_permutation_alloc(5);
	permute1(arr, 5, per);
	printf("\n");
	for(i=0;i<5;i++)
  	  	printf("%i: %lf\t", i, arr[i]);
	gsl_permutation_free(per);
	getchar();
}

void test_ran_gen_uniform()
{
	int i;
	int num = 10;
	double *dest;
	MALLOC(dest, sizeof(double)*num);
	ran_uniform(dest, num, 0, 1);
	for(i=0;i<num;i++)
	{
		printf("%i: %lf\n",i, dest[i]);
	}
	FREE(dest);
	getchar();
}

void test_gaussian()
{
	printf("Testing gaussian:\n");
	double mu = 0;double sigma = 2;int num = 10;
	double *ran;
	MALLOC(ran, sizeof(double)*num);
	ran_gaussian(ran, num, mu, sigma);
	int ind;
	for(ind=0;ind<10;ind++)
		printf("%f\n",ran[ind]);
	FREE(ran);
}


void test_uniform()
{
	printf("Testing uniform:\n");
	double lower = 5;double upper = 10;int num =10;
	double *ran;
	MALLOC(ran, sizeof(double)*num);
	ran_uniform(ran, num, lower, upper);
	int ind;
	for(ind=0;ind<10;ind++)
		printf("%f\n",ran[ind]);
	FREE(ran);
}

/*
void test_generate_p_vals()
{
	int num_in_gs = 5;
	int num_outside_gs = 10;
	double lower_gs_p_val = 0;
	double upper_gs_p_val = 0.1;

	double *seq_level_stat;
	sort_order so = increasing;

	p_vals *p_v;
	MALLOC(p_v, sizeof(p_vals));
	MALLOC(p_v->data, sizeof(double)*(num_in_gs+num_outside_gs));
	generate_p_vals(&p_v, num_in_gs, num_outside_gs,lower_gs_p_val, upper_gs_p_val, so);

	int i;
	printf("in gs:\n");
	for(i=0;i<num_in_gs;i++)
		printf("%i:%lf\n", p_v->in_gs_locations[i], p_v->data[p_v->in_gs_locations[i]]);
	printf("\n");

	printf("outside gs:\n");
	for(i=0;i<num_outside_gs;i++)
		printf("%lf\n", p_v->outside_gs_locations[i],  p_v->data[p_v->outside_gs_locations[i]]);

	for(i=0;i<num_in_gs + num_outside_gs;i++)
		printf("%i: %lf\n", i, seq_level_stat[i]);
}
*/

void demo1()
{
	const gsl_rng_type * T;
    gsl_rng *r;

    T = gsl_rng_ranlxs2;
    r = gsl_rng_alloc(T);
    unsigned long int s=15021980;
    gsl_rng_set(r, s);

	double data[5] = {17.2, 18.1, 16.5, 18.3, 12.6};
    double mean, variance, largest, smallest;

    mean     = gsl_stats_mean(data, 1, 5);
    variance = gsl_stats_variance(data, 1, 5);
    largest  = gsl_stats_max(data, 1, 5);
    smallest = gsl_stats_min(data, 1, 5);

    printf ("The dataset is %g, %g, %g, %g, %g\n", data[0], data[1], data[2], data[3], data[4]);
    printf ("The sample mean is %g\n", mean);
    printf ("The estimated variance is %g\n", variance);
    printf ("The largest value is %g\n", largest);
    printf ("The smallest value is %g\n", smallest);
}

void display_available_generators()
{
    const gsl_rng_type **t, **t0;
    t0 = gsl_rng_types_setup();
    printf ("Available generators:\n");
    for (t = t0; *t != 0; t++)
    {
    	printf ("%s\n", (*t)->name);
    }
}
