/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef RANDOM_GEN_H_
#define RANDOM_GEN_H_

#include <math.h>
#include <time.h>
#include <stdbool.h>

#include <stdio.h>
#include <sys/time.h>

#include <gsl/gsl_randist.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_rng.h>

#define oops(s) { perror((s)); exit(EXIT_FAILURE); }
#define MALLOC(s,t) if(((s) = malloc(t)) == 0) { oops("error: malloc() ");}
#define FREE(p) {free((void *)p);p=0;}

typedef struct
{
	double *data;
	int *in_gs_locations;      int num_in_gs;
	int *outside_gs_locations; int num_ouside_gs;
}p_vals;



double ran_uniform_draw(double lower, double upper);

gsl_rng *get_rng();
//static void init_ran_gen_static();
void clean_up_ran_gen_static();
//void test_ran_gen_uniform();


//double draw_from_discrete_distrib(double *weights, int num);
size_t draw_from_discrete_distrib(double *weights, int num);

//void permute(double *arr, size_t size);
void permute1(double *arr, size_t size, gsl_permutation *p);
void permute(double *arr, size_t size);

void sample_without_replacement(unsigned long *dest, unsigned long num_dest, unsigned long *src, unsigned long num_src);
void sample_with_replacement(unsigned long *dest, unsigned long num_dest, unsigned long *src, unsigned long num_src);

void get_ranks_vector(gsl_vector *v, gsl_permutation *rank);

unsigned int ran_bernoulli_draw(double p);

void ran_uniform_integer_non_unique(unsigned long int *dest, unsigned long lower, unsigned long upper, int num);
//unsigned long int *ran_uniform_integer(unsigned long lower, unsigned long upper, int *num);
//unsigned long int *ran_uniform_integer_non_unique(unsigned long lower, unsigned long upper, int num);
double ran_uniform_single_value();

void ran_gaussian(double *dest, int num_dest, double mu, double sigma);
void ran_uniform(double *dest, int num_dest, double lower, double upper);
void ran_beta(double *dest, int num_dest, double a, double b);
void ran_gamma(double *dest, int num_dest, double a, double b);

// to make use of this function need to add 'rmv.c' which implements 'rwishart'
//void ran_wishart(const int n, const int dof, const gsl_matrix *scale, gsl_matrix *result);

// draw from an unnormalized multinomial distribution
int draw_unnormalized(double *scores, int num);
void test_draw_unnormalized();
void normalize(double *arr, int num);

void display_available_generators();
void demo1();
//double random_set_ES(d_gs *dg, double *all_gene_level_stat, int num_all_genes);

// test routine
void test_gaussian();
void test_uniform();
void test_sample_with_replacement();
void test_sample_without_replacement();
void test_permute();
//void test_generate_p_vals();

#endif /*RANDOM_GEN_H_*/
