/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "regression.h"


void estimate_var(double *data, int num)
{
//	double var = gsl_stats_variance(data, 1, num);
 	double mean = gsl_stats_mean(data, 1, num);
	double var = gsl_stats_variance_m(data, 1, num, mean);

	double mean_sum_squared = 0;
	int i;
	for(i=0;i<num;i++)
		mean_sum_squared += (data[i])*(data[i]);
	mean_sum_squared = mean_sum_squared / (num - 1);
	double var1 = mean_sum_squared - mean*mean;
	printf("sd  = %lf\n", sqrt(var));
	printf("sd1 = %lf\n", sqrt(var1));
}


char *covariates_file;
gsl_matrix *covariates, *covariates_original, *covariates_shuffled;
//gsl_vector *response;
gsl_vector *beta;//, *S;
gsl_matrix *cov;//, *U, *V;
gsl_multifit_linear_workspace *work;


bool consider_rev_compl_flag;

void display_regression_var_names()
{
	char var_names_file[500];
	sprintf(var_names_file, "%s.names", covariates_file);
//	printf("var_names_file = %s\n", var_names_file);
	FILE *fp = fopen(var_names_file, "r");	
	if(fp == 0)
	{
		printf("ERROR: unable to open covariates var names file %s!\n", var_names_file);
		exit(1);
	}
	char var_name_loc[100];
	int index = 0;
	while(fscanf(fp, "%s ", var_name_loc) != EOF) 
	{
		printf("%s ", var_name_loc);
	}
	fclose(fp);
}


gsl_matrix *read_mat_regression(char *file)
{
	FILE *f = fopen(file, "r");
	if(f == NULL)
		return 0;

	int rows, cols;
	fscanf(f, "%i\t%i\n", &rows, &cols);
	gsl_matrix *mat = gsl_matrix_alloc(rows, cols);
	int res = gsl_matrix_fscanf(f, mat);
	fclose(f);
	if(res == GSL_EFAILED)
	{
 //		printf("failed reading pssm from file '%s'\n", file);
		return 0;
	}
	return mat;
}


void set_column(gsl_matrix *mat, int col_index, double value)
{
	int i;
	for(i=0;i<mat->size1;i++)
		gsl_matrix_set(mat, i, col_index, value);
}

void log_transform_column(gsl_matrix *mat, int col_index)
{
	int i;
	for(i=0;i<mat->size1;i++)
	{
		double val = gsl_matrix_get(mat, i, col_index);
		if(val < 1.0)
			gsl_matrix_set(mat, i, col_index, 0);
		else
			gsl_matrix_set(mat, i, col_index, log(val));	
	}
}

void center_column(gsl_matrix *mat, int col_index)
{
	int i;
	double col_mean = 0.0;
	for(i=0;i<mat->size1;i++)
		col_mean += gsl_matrix_get(mat, i, col_index);
	col_mean /= mat->size1;
	for(i=0;i<mat->size1;i++)
		gsl_matrix_set(mat, i, col_index, gsl_matrix_get(mat, i, col_index) - col_mean );
}

void scale_column_SD(gsl_matrix *mat, int col_index)
{
	int i;
	double col_mean = 0.0, ss = 0.0;
	for(i=0;i<mat->size1;i++)
	{
		double val = gsl_matrix_get(mat, i, col_index);
		col_mean += val;
		ss += val * val;
	}
	col_mean /= mat->size1;

	double col_sd = sqrt(ss / mat->size1 - col_mean * col_mean);
	if(col_sd > 0)
	{
		for(i=0;i<mat->size1;i++)
		{
			double val = gsl_matrix_get(mat, i, col_index);
			gsl_matrix_set(mat, i, col_index, val / col_sd );
		}
	}
}

void log_transfrm_and_standardize_var(gsl_matrix *mat, int col_index)
{
	log_transform_column(mat, col_index);
//	printf("covariates (log_transform_column lengths column)\n");
//	print_matrix_GSL_m_n(mat, 5, 5);
	center_column(mat, col_index);
	scale_column_SD(mat, col_index);
//	printf("covariates (transformed lengths column)\n");
//	print_matrix_GSL_m_n(mat, 5, 5);
}

void get_var_name_index(char *covariates_file, char *var_name, int *var_name_index)
{
	char var_names_file[500];
	sprintf(var_names_file, "%s.names", covariates_file);
//	printf("var_names_file = %s\n", var_names_file);
	FILE *fp = fopen(var_names_file, "r");	
	if(fp == 0)
	{
		printf("ERROR: unable to open covariates var names file %s!\n", var_names_file);
		exit(1);
	}
	char var_name_loc[100];
	int index = 0;
	while(fscanf(fp, "%s ", var_name_loc) != EOF) 
	{
		if(strcmp(var_name_loc, var_name) == 0)
		{
				*var_name_index = index;		
//				printf("%i: %s\n", *var_name_index, var_name);
		}
		index++;
	}
//	printf("done read_var_names\n");
	fclose(fp);
//	getchar();
//	char *var_name = strtok(char *str1, const char *str2 );
	
}



void select_k(gsl_vector *singular_vals, int *k)
{
	int i;
	double total = 0;
	for(i=0;i<singular_vals->size;i++)
		total += (singular_vals->data[i] * singular_vals->data[i]);

	double var_thresh = 0.99;
	double explained_var = 0;
	double prev_explained_var = (singular_vals->data[0] * singular_vals->data[0]) / total;
	double curr_explained_var;
	*k = 1;
	for(i=0;i<singular_vals->size;i++)
	{
		curr_explained_var = (singular_vals->data[i] * singular_vals->data[i]) / total;
		explained_var += curr_explained_var;
//		printf("%i: prev_explained_var = %lf\tcurr_explained_var = %lf, prev/curr = %lf\n", i+1, prev_explained_var, curr_explained_var, prev_explained_var / curr_explained_var);
		if(i > 0)
		{
			if(curr_explained_var < 0.01)// && curr_explained_var <= 0.5 * prev_explained_var)
			{
//				printf("curr_explained_var (%lf) <= 0.5 * prev_explained_var (%lf)\n", curr_explained_var, prev_explained_var);
				*k = i;
				break;
			}
			else if(explained_var >= var_thresh)
			{
//				printf("explained_var (%lf) >= var_thresh (%lf)\n", explained_var, var_thresh);
				*k = i+1;
				break;
			}
		}
		prev_explained_var = curr_explained_var;
//		if(explained_var >= var_thresh)
//			break;
	}
//	printf("k = %i\n", *k);
	double cum_sum = 0;
	for(i=0;i<singular_vals->size;i++)
	{
		cum_sum += (singular_vals->data[i] * singular_vals->data[i]) / total;
//		if((100 * singular_vals->data[i]) / total >= 1.0)
			printf("%i: %15.2lf%15.2lf\n", i+1, (100 * (singular_vals->data[i] * singular_vals->data[i])) / total, cum_sum);
		if(i == *k - 1)
			printf("---------------\n");
	}
}

double get_norm(gsl_matrix *mat, int col_index)
{
	int i;
	double norm = 0;
	for(i=0;i<mat->size1;i++)
	{
		double entry = gsl_matrix_get(mat, i, col_index);
		norm += entry * entry;
	}
	norm = sqrt(norm);
	return norm;
}

int num_PCs;

gsl_matrix *U_global;

bool is_regression_initialized = false;
bool do_PC_regression = false;	
bool add_intercept = true;


void init_regression(char *covariates_file)
{
	printf("covariates file: %s\n", covariates_file);

	is_regression_initialized = true;

	U_global = 0;

	FILE *fp = fopen(covariates_file, "r");
	if(fp == 0)
	{
		printf("cannot open covariates_file: %s\n", covariates_file);
		exit(1);
	}
	fclose(fp);
	gsl_matrix *common_covariates = read_mat_regression(covariates_file);
//	printf("covariates 'read_mat_regression'\n");
//	print_matrix_GSL_m_n(common_covariates, 5, 5);


//	char *var_name = "length";//"GC";//
	char *var_name = "+Strand";//"GC";//
	int var_name_index = -1;
	get_var_name_index(covariates_file, var_name, &var_name_index);
//	if(var_name_index != -1)
//		log_transfrm_and_standardize_var(common_covariates, var_name_index);
	int i;
	for(i=0;i<common_covariates->size2;i++)
	{
//		log_transfrm_and_standardize_var(common_covariates, i);
//		if(i != var_name_index)
//		{
			center_column(common_covariates, i);
			scale_column_SD(common_covariates, i);
//		}
	}
//	printf("covariates (centered & scaled)\n");
//	print_matrix_GSL_m_n(common_covariates, 5, 5);
//	getchar();

/*
	dimension dim = columns;//rows;
	scaling_type scaling = StdDev;
	GSL_center(common_covariates, dim);
	printf("centered covariates\n");
//	print_matrix_GSL_m_n(common_covariates, 5, 5);

//	dim = rows;//columns;//rows;
	GSL_scale(common_covariates, dim, scaling);
//	GSL_center_and_scale(common_covariates, dim, scaling);

	printf("centered + scaled covariates\n");
	print_matrix_GSL_m_n(common_covariates, 5, 5);
//	getchar();
*/

	do_PC_regression = false;

	bool combined_regression = false;
	if(do_PC_regression == true)
	{
//		printf("common_covariates: %i-by-%i\n", common_covariates->size1, common_covariates->size2);
//		getchar();
		eigs_SVD out;

//		out.U = gsl_matrix_alloc(common_covariates->size1, common_covariates->size2);
		out.U = gsl_matrix_alloc(common_covariates->size1, common_covariates->size1);

		out.S = gsl_vector_alloc(common_covariates->size2);
//		out.V = gsl_matrix_alloc(common_covariates->size2, common_covariates->size2);
		out.V = gsl_matrix_alloc(common_covariates->size1, common_covariates->size2);

		gsl_matrix *common_covariates_copy = gsl_matrix_alloc(common_covariates->size1, common_covariates->size2);
	
		gsl_matrix_memcpy(common_covariates_copy, common_covariates);
	
//		printf("common_covariates (before svd)'\n");
//		print_matrix_GSL_m_n(common_covariates_copy, 10, 10);

//		printf("svd: modified_Golub_Reinsch\n");

		QR_decomp(common_covariates_copy, out.U, out.V);

//		SVD_modified_Golub_Reinsch(common_covariates_copy, &out);
//		SVD_Jacobi(common_covariates_copy, &out);
//		SVD_Golub_Reinsch(common_covariates_copy, &out);
//		printf("done SVD\n");

//		printf("out.V\n");
//		print_matrix_GSL_m_n(out.V, 10, 10);
//		printf("out.U\n");
//		print_matrix_GSL_m_n(out.U, 10, 10);
//		getchar();

//		printf("sqrt(lambda_max / lambda_min) = %lf\n\n", out.S->data[0] / out.S->data[out.S->size - 1]);

//		select_k(out.S, &num_PCs);//getchar();
//		printf("selected top %i\n\n", num_PCs);
//		num_PCs = 5;//8;//5-8
		num_PCs = common_covariates->size2 - 1; // centered data!!!
//		num_PCs = num_PCs < common_covariates->size2 ? num_PCs : common_covariates->size2;//common_covariates->size2;
		printf("using top %i\n\n", num_PCs);

		printf("out.U\n");
		print_matrix_GSL_m_n(out.U, 5, 5);
//		getchar();


		gsl_matrix_view mview_U = gsl_matrix_submatrix(out.U, 0, 0, out.U->size1, num_PCs);
/*
		int offset = 0;
		gsl_vector_view vview_S = gsl_vector_subvector(out.S, offset, num_PCs);
		gsl_matrix *S_diag = diag(&vview_S.vector);
	
		gsl_matrix *U_times_S = gsl_matrix_alloc(out.U->size1, num_PCs);
		mat_mat_multiply(&mview_U.matrix, S_diag, U_times_S, CblasNoTrans, CblasNoTrans);
*/
//		printf("S_diag\n");
//		print_matrix_GSL_m_n(S_diag, 5, 5);

		if(combined_regression == false)
		{
			common_covariates = gsl_matrix_alloc(out.U->size1, num_PCs);
			gsl_matrix_memcpy(common_covariates, &mview_U.matrix);
//			printf("common_covariates\n");
//			print_matrix_GSL_m_n(common_covariates, 5, 5);
		}
		else
		{
			gsl_matrix_view mview_V = gsl_matrix_submatrix(out.V, 0, 0, out.V->size1, num_PCs);
	

			int offset = 0;
			gsl_vector_view vview_S = gsl_vector_subvector(out.S, offset, num_PCs);
			gsl_matrix *S_diag = diag(&vview_S.vector);
		

			gsl_matrix *U_times_S = gsl_matrix_alloc(out.U->size1, num_PCs);
			mat_mat_multiply(&mview_U.matrix, S_diag, U_times_S, CblasNoTrans, CblasNoTrans);

//			printf("U_times_S\n");
//			print_matrix_GSL_m_n(U_times_S, 5, 5);
	
			gsl_matrix *U_times_S_times_V_tr = gsl_matrix_alloc(U_times_S->size1, mview_V.matrix.size1);
			mat_mat_multiply(U_times_S, &mview_V.matrix, U_times_S_times_V_tr, CblasNoTrans, CblasTrans);
		
//			printf("U_times_S_times_V_tr\n");
//			print_matrix_GSL_m_n(U_times_S_times_V_tr, 5, 5);
		
			gsl_matrix *residual = gsl_matrix_alloc(common_covariates->size1, common_covariates->size2);
			gsl_matrix_memcpy(residual, common_covariates);
	
			gsl_matrix_sub(residual, U_times_S_times_V_tr);

			int ii;
			for(ii=0;ii<residual->size2;ii++)
			{
				double norm = get_norm(residual, ii);
				printf("norm %i = %lf\n", ii, norm);
			}
//		getchar();
//		gsl_matrix_view mview0 = gsl_matrix_submatrix(residual, 0, 0, residual->size1, num_PCs);

//		gsl_matrix *top_PCs = gsl_matrix_alloc(out.U->size1, num_PCs);
	//	gsl_matrix_memcpy(top_PCs, &mview_U.matrix);

			gsl_matrix_free(common_covariates);
			common_covariates = gsl_matrix_alloc(out.U->size1, num_PCs + residual->size2);
			
			gsl_matrix_view mview1 = gsl_matrix_submatrix(common_covariates, 0, 0, out.U->size1, num_PCs);
			gsl_matrix_memcpy(&mview1.matrix, &mview_U.matrix);
			mview1 = gsl_matrix_submatrix(common_covariates, 0, num_PCs, out.U->size1, residual->size2);
			gsl_matrix_memcpy(&mview1.matrix, residual);
		}
//		printf("common_covariates\n");
//		print_matrix_GSL_m_n(common_covariates, 5, 5);
//		getchar();
	//	gsl_matrix_free(S_diag);
		gsl_matrix_free(out.U);
		gsl_matrix_free(out.V);
		gsl_vector_free(out.S);

		gsl_matrix_free(common_covariates_copy);
//		common_covariates = top_PCs;
	}

//	printf("common_covariates\n");
//	print_matrix_GSL_m_n(common_covariates, 5, 5);

//	printf("PCs\n");
//	print_matrix_GSL_m_n(common_covariates, 5, 5);

	// 1. mean-center response => exclude intercept 
	// 2. mean-center all covariates
	// 3. 

//	printf("hhh000\n");fflush(stdout);

	// add one covariate for the motif occurrence scores
	size_t n = common_covariates->size1, p = common_covariates->size2 + 1;

	add_intercept = true;

	gsl_matrix_view mview;
	if(add_intercept == true)
	{
		p++;
		covariates_original = gsl_matrix_alloc(n, p);
		int col_index = 0;
		set_column(covariates_original, col_index, 1.0);
		mview = gsl_matrix_submatrix(covariates_original, 0, 1, covariates_original->size1, common_covariates->size2);
		gsl_matrix_memcpy(&mview.matrix, common_covariates);
	}
	else
	{
		covariates_original = gsl_matrix_alloc(n, p);
		mview = gsl_matrix_submatrix(covariates_original, 0, 0, covariates_original->size1, common_covariates->size2);
		gsl_matrix_memcpy(&mview.matrix, common_covariates);
	}
	set_column(covariates_original, p-1, 0);
	covariates_shuffled = gsl_matrix_alloc(n, p);
	gsl_matrix_memcpy(covariates_shuffled, covariates_original);

//	printf("covariates_original\n");
//	print_matrix_GSL_m_n(covariates_original, 5, 200);
//	getchar();

	work = gsl_multifit_linear_alloc(n, p);
//	response = gsl_vector_alloc(n);
	beta = gsl_vector_alloc(p);
	cov = gsl_matrix_alloc(p, p);
	gsl_matrix_free(common_covariates);
//	printf("done reading covariates\n");

//	do_PC_regression = false;
}


void free_regression()
{
	gsl_matrix_free(covariates_original);
	gsl_matrix_free(covariates_shuffled);
//	gsl_vector_free(response);
	gsl_matrix_free(cov);
	gsl_vector_free(beta);
	gsl_multifit_linear_free(work);
}

