/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include "rmvn.h"

//extern gsl_rng *r1;

/*
gsl_rng *get_rng()
{
	gsl_rng *r = gsl_rng_alloc(gsl_rng_gfsr4);
	unsigned long int s = time(NULL);//123456;//
	gsl_rng_set(r, s);
	return r;
}
*/
// generate W = tr(A) %*% A, where A is a rectangular matrix with
// 'dof' independent draws from N(0,I), where I is p-by-p
void my_rmvn(gsl_rng *r, int num_subjects, int num_SNPs, gsl_matrix *result)
{
	int m = num_subjects;
	int n = num_SNPs;

	int i,j;
	double univariate_draw;
	for(i=0;i<m;i++)
	{
		for(j=0;j<n;j++)
		{
			// generate univariate N(0,1) 
			univariate_draw = gsl_ran_gaussian_ziggurat(r, 1);
			gsl_matrix_set(result, i, j, univariate_draw);
		}
	}
}


/* multivariate normal distribution random number generator */
/*
*	n	dimension of the random vector
*	mean	vector of means of size n
*	var	variance matrix of dimension n x n
*	result	output variable with a sigle random vector normal distribution generation
*/
void rmvnorm(gsl_rng *r, int n, gsl_vector *mean, gsl_matrix *var, gsl_vector *result)
{
	int k;
	gsl_matrix *work = gsl_matrix_alloc(n,n);
	
	gsl_matrix_memcpy(work,var);
	gsl_linalg_cholesky_decomp(work);
	
	for(k=0; k<n; k++)
		gsl_vector_set( result, k, gsl_ran_ugaussian(r) );
	
	gsl_blas_dtrmv(CblasLower, CblasNoTrans, CblasNonUnit, work, result);
	gsl_vector_add(result,mean);
	
	gsl_matrix_free(work);
}

void generate_simple_Gaussian_in_place(int m, int n, gsl_matrix *G, gsl_rng *r1)
{
//	gsl_rng *r1 = get_rng();
	my_rmvn(r1, m, n, G);
}

// generate W = tr(A) %*% A, where A is a rectangular matrix with
// 'dof' independent draws from N(0,I), where I is p-by-p
void my_rmvu(gsl_rng *r, int num_subjects, int num_SNPs, gsl_matrix *result)
{
	int m = num_subjects;
	int n = num_SNPs;

	int i,j;
	double univariate_draw;
	for(i=0;i<m;i++)
	{
		for(j=0;j<n;j++)
		{
			// generate univariate U(-1,1) 
// 			double gsl_ran_flat (const gsl_rng * r, double a, double b)
			univariate_draw = gsl_ran_flat(r, -1, 1);
			gsl_matrix_set(result, i, j, univariate_draw);
		}
	}
}

void generate_simple_Uniform_in_place(int m, int n, gsl_matrix *G, gsl_rng *r1)
{
	my_rmvu(r1, m, n, G);
}


// COV = Identity
gsl_matrix *generate_simple_Gaussian(int p, int dof, gsl_rng *r1)
{
	init_ran_gen_static();

//	gsl_rng *r1 = get_rng();
	gsl_matrix *M = gsl_matrix_alloc(p, dof);
	my_rmvn(r1, p, dof, M);
	return M;	

	gsl_matrix *W = gsl_matrix_alloc(dof, p);
	int i,j;
	gsl_vector *mean   = gsl_vector_alloc(p);
	gsl_vector *result = gsl_vector_alloc(p);

	for(j=0;j<p;j++)
		gsl_vector_set(mean, j, 0.0);

	gsl_matrix *V = gsl_matrix_alloc(p, p);
	gsl_matrix_set_identity(V);

	gsl_rng *r = get_rng();

	for(i=0;i<dof;i++)
	{
		rmvnorm(r, p, mean, V, result);
		for(j=0;j<p;j++)
			gsl_matrix_set(W, i, j, gsl_vector_get(result, j));
	}
	gsl_vector_free(result);
	gsl_vector_free(mean);
	gsl_matrix_free(V);
	gsl_rng_free(r);

//	double scale = 1/sqrt(dof);
//	gsl_matrix_scale(W, scale);

	return W;
}

void test_generate_simple_Gaussian()
{
	int p = 2, dof = 3;
	gsl_rng *r = get_rng();
	gsl_matrix *res = generate_simple_Gaussian(p, dof, r);
	int i,j;
	for(i=0;i<dof;i++)
	{
		for(j=0;j<p;j++)
			printf("%lf\t", gsl_matrix_get(res, i, j));
		printf("\n");
	}

//	gsl_matrix_fprintf(stdout, res, "%lf");
	gsl_rng_free(r);
	gsl_matrix_free(res);	
	getchar();
}



