/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef SIMILARITY_METRICS_H_
#define SIMILARITY_METRICS_H_

#include "parameters.h"
#include "math.h"
#include "PSSM.h"

#define ZERO	1.0e-20

//void init_background_freq(species species_type);
//gsl_vector *get_background_freq_vec_AGCT();

//typedef enum {Kullbac_Leibler_Divergence, Log_likelihood_Ratio} similarity_metric;
//typedef enum {PCC_with_overhang, PCC_no_overhang, Harbison, Harbison_no_background} similarity_score_type;
typedef enum {Harbison, Harbison_no_background} similarity_score_type;

// dispatch method for the other similarity scores (need to set a global flag to select the type of score)
double similarity_score(gsl_matrix *pssm1, gsl_matrix *pssm2, int required_core_length, int *offset_top_score);

double comapare_pssms_Harbison(pssm1, pssm2);

double similarity_score_Harbison_with_overhang(gsl_matrix *pssm1, gsl_matrix *pssm2);
double similarity_score_Harbison_with_overhang_no_background(gsl_matrix *pssm1, gsl_matrix *pssm2, int required_core_length, int *offset_top_score);
double similarity_score_Harbison_equal_length(gsl_matrix *f, gsl_matrix *g);

double ave_entropy(gsl_matrix *pssm, double eps);
double similarity_score_Harbison_with_overhang_no_background_with_entropy_constraint
(gsl_matrix *predicted_pssm, gsl_matrix *known_pssm, int required_core_length, int *offset_top_score);

double eval_PSSM_similarity(gsl_matrix *pssm1, gsl_matrix *pssm2, int required_core_length, int *offset_top_score);


double eval_distance_DRIM(char *m1, char *m2);

#endif /*SIMILARITY_METRICS_H_*/
