/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#include <stdlib.h>

//#include "include_all.h"
#include "sort.h"

/****************************************/
// sorting routines for:
// (1) int
// (2) double
// (3) d_gs (by value of 'score')
// (4) d_gs (by absolute value of 'score')
/*****************************************/

/* sorting 'integers' */
int compar(const void *s1 ,const void *s2 )
{
	if(*((int *)s1) > *((int *)s2)) return 1;
	else if(*((int *)s1) == *((int *)s2)) return 0;
	else return -1;
}

void sort(int array[], int NumberOfElements)
{
	size_t Size = sizeof(int);
	qsort(array, NumberOfElements, Size, compar);
}

void test_sort()
{
	int a[] = {1,2,65,32,5};
	int j;
	for(j=0; j < 5; j++)
		printf("%i\n", a[j]);
	printf("\n");
	sort(a, 5);
	for(j=0; j < 5; j++)
		printf("%i\n", a[j]);
}

/* sorting 'integers'  and keep track of index */
int *cmp_int_array;
int compar_increasing_idx(const void *s1 ,const void *s2 )
{
	int is1 = *(int *)s1;
	int is2 = *(int *)s2;
	if(cmp_int_array[is1] > cmp_int_array[is2]) return 1;
	else if(cmp_int_array[is1] == cmp_int_array[is2]) return 0;
	else return -1;
}

int compar_decreasing_idx(const void *s1 ,const void *s2)
{
	int is1 = *(int *)s1;
	int is2 = *(int *)s2;
	if(cmp_int_array[is1] > cmp_int_array[is2]) return -1;
	else if(cmp_int_array[is1] == cmp_int_array[is2]) return 0;
	else return 1;
}

int* sort_idx(int array[], int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(int);
	int *index;
	int i;
	MALLOC(index, sizeof(int) * NumberOfElements);
	for(i=0;i<NumberOfElements;i++)
	{
		index[i] = i;
	}
	cmp_int_array = array;
	if(so == increasing)
		qsort(index, NumberOfElements, Size, compar_increasing_idx);
	else
		qsort(index, NumberOfElements, Size, compar_decreasing_idx);
	return index;
}

void test_sort_idx()
{
	int a[] = {1,-2,65,32,5};
	int j;
	printf("Testing sort_idx()\n");
	for(j=0; j < 5; j++)
		printf("%i\n", a[j]);
	sort_order so = decreasing;
	int *idx;
	idx = sort_idx(a, 5, so);
	printf("\n\ndata\tindex\n");
	for(j=0; j < 5; j++)
		printf("%i\t%i\n", a[idx[j]], idx[j]);
}

///////////////////////////////
///////////////////////////////

/* sorting 'double' */
int compar_double_decreasing(const void *s1 ,const void *s2)
{
	if(*((double *)s1) > *((double *)s2)) return -1;
	else if(*((double *)s1) == *((double *)s2)) return 0;
	else return 1;
}


int compar_double_increasing(const void *s1 ,const void *s2)
{
	if(*((double *)s1) > *((double *)s2)) return 1;
	else if(*((double *)s1) == *((double *)s2)) return 0;
	else return -1;
}

void sort_double(double array[], int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(double);
	if(so == increasing)
		qsort(array, NumberOfElements, Size, compar_double_increasing);
	else
		qsort(array, NumberOfElements, Size, compar_double_decreasing);
}

void test_sort_double()
{
	double a[] = {1.9,-2.3,65,32.1,5};
	int j;
	for(j=0; j < 5; j++)
		printf("%lf\n", a[j]);
	printf("\n\n");
	sort_order so = decreasing;
	sort_double(a, 5, so);
	for(j=0; j < 5; j++)
		printf("%lf\n", a[j]);
}

/* sorting 'double' and keep track of index*/
double *cmp_array;

int compar_double_increasing_idx(const void *s1 ,const void *s2)
{
	int is1 = *(int *)s1;
	int is2 = *(int *)s2;
	if(cmp_array[is1] > cmp_array[is2]) return 1;
	else if(cmp_array[is1] == cmp_array[is2]) return 0;
	else return -1;
}

int compar_double_decreasing_idx(const void *s1 ,const void *s2)
{
	int is1 = *(int *)s1;
	int is2 = *(int *)s2;
	if(cmp_array[is1] > cmp_array[is2]) return -1;
	else if(cmp_array[is1] == cmp_array[is2]) return 0;
	else return 1;
}

int* sort_double_idx(double array[], int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(int);
	int *index;
	int i;
	MALLOC(index, sizeof(int) * NumberOfElements);
	for(i=0;i<NumberOfElements;i++)
	{
		index[i] = i;
	}
	cmp_array = array;
	if(so == increasing)
		qsort(index, NumberOfElements, Size, compar_double_increasing_idx);
	else
		qsort(index, NumberOfElements, Size, compar_double_decreasing_idx);
	return index;
}

void test_sort_double_idx()
{
	double a[] = {7.9,1.9,-2.3,65,32.1};
	int j;
	for(j=0; j < 5; j++)
		printf("%lf\n", a[j]);
	sort_order so = decreasing;
	int *idx;
	idx = sort_double_idx(a, 5, so);
	printf("\n\ndata\tindex\n");
	for(j=0; j < 5; j++)
		printf("%lf\t%i\n", a[idx[j]], idx[j]);
}


///////////////////////////////////
///////////////////////////////////

/* sorting 'd_gs' by value of 'score' */
int compar_d_gs_decreasing(const void *s1 ,const void *s2)
{
//	d_gs **tmp1 = (d_gs **)s1;
//	d_gs **tmp2 = (d_gs **)s2;
//	if(s1 == 0)
//		return -1;
//	else if(s2 == 0)
//		return 1;

	if((*((d_gs **)s1))->score > ((*(d_gs **)s2))->score) return -1;
	else if((*((d_gs **)s1))->score == ((*(d_gs **)s2))->score) return 0;
	else return 1;
}

int compar_d_gs_increasing(const void *s1 ,const void *s2)
{
//	d_gs **tmp1 = (d_gs **)s1;
//	d_gs **tmp2 = (d_gs **)s2;
//	if(s1 == 0)
//		return -1;
//	else if(s2 == 0)
//		return 1;

	if((*((d_gs **)s1))->score > ((*(d_gs **)s2))->score) return 1;
	else if((*((d_gs **)s1))->score == ((*(d_gs **)s2))->score) return 0;
	else return -1;
}

void sort_d_gs(d_gs **array, int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(d_gs *);
	if(so == increasing)
		qsort(array, NumberOfElements, Size, compar_d_gs_increasing);
	else
		qsort(array, NumberOfElements, Size, compar_d_gs_decreasing);
}

void test_sort_d_gs()
{
//	int num_d_gs = 5;
//	d_gs *dg[num_d_gs];
//	
////	dg[0] = create_empty_d_gs(); dg[0]->score = 1.34;
////	dg[1] = create_empty_d_gs(); dg[1]->score = -1.34;
////	dg[2] = create_empty_d_gs(); dg[2]->score = 2;
////	dg[3] = create_empty_d_gs(); dg[3]->score = 0;
////	dg[4] = create_empty_d_gs(); dg[4]->score = -4;
//
//	int j;
//	for(j=0; j < num_d_gs; j++)
//		printf("%-3.1lf\n", dg[j]->score);
//	printf("\n");
//	sort_order so = decreasing;
//	sort_d_gs(dg, num_d_gs, so);
//	for(j=0; j < num_d_gs; j++)
//		printf("%-3.1lf\n", dg[j]->score);
////	free_d_gs_batch(dg, num_d_gs, false);
//	getchar();
}

/*
int compar_miRNA_decreasing(const void *s1 ,const void *s2)
{
	printf("score = %lf\n", (*((miRNA **)s1))->max_score);

	if((*((miRNA **)s1))->max_score > ((*(miRNA **)s2))->max_score) return -1;
	else if((*((miRNA **)s1))->max_score == ((*(miRNA **)s2))->max_score) return 0;
	else return 1;
}

int compar_miRNA_increasing(const void *s1 ,const void *s2)
{
	if((*((miRNA **)s1))->max_score > ((*(miRNA **)s2))->max_score) return 1;
	else if((*((miRNA **)s1))->max_score == ((*(miRNA **)s2))->max_score) return 0;
	else return -1;
}

void sort_miRNA(miRNA **array, int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(miRNA *);
	if(so == increasing)
		qsort(array, NumberOfElements, Size, compar_miRNA_increasing);
	else
		qsort(array, NumberOfElements, Size, compar_miRNA_decreasing);
}
*/

/*
miRNA *miRNA_arr;

int compar_miRNA_decreasing(const void *s1 ,const void *s2)
{
//	printf("score = %lf\n", (*((miRNA **)s1))->max_score);
	int *ind1 = (int *)s1;
	int *ind2 = (int *)s2;
	if(miRNA_arr[*ind1].pMax->score > miRNA_arr[*ind2].pMax->score) return -1;
	else if(miRNA_arr[*ind1].pMax->score == miRNA_arr[*ind2].pMax->score) return 0;
	else return 1;
}

int compar_miRNA_increasing(const void *s1 ,const void *s2)
{
	int *ind1 = (int *)s1;
	int *ind2 = (int *)s2;
	if(miRNA_arr[*ind1].pMax->score > miRNA_arr[*ind2].pMax->score) return 1;
	else if(miRNA_arr[*ind1].pMax->score == miRNA_arr[*ind2].pMax->score) return 0;
	else return -1;
}

void sort_miRNA(miRNA *array, int *indexes, int NumberOfElements, sort_order so)
{
	miRNA_arr = array;
	size_t Size = sizeof(int);
	if(so == increasing)
		qsort(indexes, NumberOfElements, Size, compar_miRNA_increasing);
	else
		qsort(indexes, NumberOfElements, Size, compar_miRNA_decreasing);
}
*/


///////////////////////////////////
///////////////////////////////////
/*
typedef struct{int index;double score;} score_index;

typedef struct
{
	char **raw_input;
	int *index;double *scores;
	char **seq_ids;
	score_index *sort_index;
	int num;
}gene_level_scores;
*/

/*
// sorting 'gene_level_scores' by value of 'score'
int compar_gene_level_scores_decreasing(const void *s1 ,const void *s2)
{
	if((*((score_index **)s1))->score > (*((score_index **)s1))->score) return -1;
	else if((*((score_index **)s1))->score == (*((score_index **)s1))->score) return 0;
	else return 1;
}

int compar_gene_level_scores_increasing(const void *s1 ,const void *s2)
{
	if((*((score_index **)s1))->score > (*((score_index **)s1))->score) return 1;
	else if((*((score_index **)s1))->score == (*((score_index **)s1))->score) return 0;
	else return -1;
}

// needs to be tested 
void sort_gene_level_scores(gene_level_scores *array, int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(score_index);
	if(so == increasing)
		qsort(&(array->sort_index), NumberOfElements, Size, compar_gene_level_scores_increasing);
	else
		qsort(&(array->sort_index), NumberOfElements, Size, compar_gene_level_scores_decreasing);
}
*/


// sorting 'gene_level_scores' by value of 'score' 
int compar_gene_level_scores_decreasing(const void *s1 ,const void *s2)
{
	if(((score_index *)s1)->score > ((score_index *)s2)->score) return -1;
	else if(((score_index *)s1)->score == ((score_index *)s2)->score) return 0;
	else return 1;
}

int compar_gene_level_scores_increasing(const void *s1 ,const void *s2)
{
	if(((score_index *)s1)->score > ((score_index *)s2)->score) return 1;
	else if(((score_index *)s1)->score == ((score_index *)s2)->score) return 0;
	else return -1;
}

// doesn't work!!!
void sort_gene_level_scores(gene_level_scores array[], int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(score_index);
	if(so == increasing)
		qsort(array->sort_index, NumberOfElements, Size, compar_gene_level_scores_increasing);
	else
		qsort(array->sort_index, NumberOfElements, Size, compar_gene_level_scores_decreasing);
}






///////////////////////////////
///////////////////////////////

/* sorting 'expr_dataset' by value of 'expr->seq_level_stat[i]' */

/*
int compar_d_gs_decreasing(const void *s1 ,const void *s2)
{
	if((*((d_gs **)s1))->score > ((*(d_gs **)s2))->score) return -1;
	else if((*((d_gs **)s1))->score == ((*(d_gs **)s2))->score) return 0;
	else return 1;
}

int compar_d_gs_increasing(const void *s1 ,const void *s2)
{
	if((*((d_gs **)s1))->score > ((*(d_gs **)s2))->score) return 1;
	else if((*((d_gs **)s1))->score == ((*(d_gs **)s2))->score) return 0;
	else return -1;
}

void sort_d_gs(d_gs **array, int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(d_gs *);
	if(so == increasing)
		qsort(array, NumberOfElements, Size, compar_d_gs_increasing);
	else
		qsort(array, NumberOfElements, Size, compar_d_gs_decreasing);
}

void test_sort_d_gs()
{
	int num_d_gs = 5;
	d_gs *dg[num_d_gs];
	dg[0] = create_empty_d_gs(); dg[0]->score = 1.34;
	dg[1] = create_empty_d_gs(); dg[1]->score = -1.34;
	dg[2] = create_empty_d_gs(); dg[2]->score = 2;
	dg[3] = create_empty_d_gs(); dg[3]->score = 0;
	dg[4] = create_empty_d_gs(); dg[4]->score = -4;
	int j;
	for(j=0; j < num_d_gs; j++)
		printf("%-3.1lf\n", dg[j]->score);
	printf("\n");
	sort_order so = decreasing;
	sort_d_gs(dg, num_d_gs, so);
	for(j=0; j < num_d_gs; j++)
		printf("%-3.1lf\n", dg[j]->score);
//	free_d_gs_batch(dg, num_d_gs, false);
	getchar();
}
*/

///////////////////////////////
///////////////////////////////

/* sorting 'd_gs' by absolute value of 'score' */
int compar_d_gs_decreasing_abs(const void *s1 ,const void *s2)
{
	if( fabs((*((d_gs **)s1))->score) > fabs(((*(d_gs **)s2))->score) ) return -1;
	else if( (*((d_gs **)s1))->score == ((*(d_gs **)s2))->score ) return 0;
	else return 1;
}

int compar_d_gs_increasing_abs(const void *s1, const void *s2)
{
	if( fabs((*((d_gs **)s1))->score) > fabs(((*(d_gs **)s2))->score) ) return 1;
	else if( (*((d_gs **)s1))->score == ((*(d_gs **)s2))->score ) return 0;
	else return -1;
}

void sort_d_gs_abs(d_gs **array, int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(d_gs *);
	if(so == increasing)
		qsort(array, NumberOfElements, Size, compar_d_gs_increasing_abs);
	else
		qsort(array, NumberOfElements, Size, compar_d_gs_decreasing_abs);
}

///////////////////////////////////////
///////////////////////////////////////

/* sorting 'gene_set_profile' by value of 'score' */
/*
// some problems when using it...
int compar_oligo_count_pair_decreasing(const void *s1, const void *s2)
{
	if( (*((oligo_count_pair **)s1))->count > ((*(oligo_count_pair **)s2) )->count) return -1;
	else if( (*((oligo_count_pair **)s1))->count == ((*(oligo_count_pair **)s2))->count ) return 0;
	else return 1;
}

int compar_oligo_count_pair_increasing(const void *s1, const void *s2)
{
	if( (*((oligo_count_pair **)s1))->count > (*((oligo_count_pair **)s2))->count ) return 1;
	else if( (*((oligo_count_pair **)s1))->count == (*((oligo_count_pair **)s2))->count ) return 0;
	else return -1;
}

void sort_oligo_count_pairs(oligo_count_pair **array, int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(oligo_count_pair *);
	if(so == increasing)
		qsort(array, NumberOfElements, Size, compar_oligo_count_pair_increasing);
	else
		qsort(array, NumberOfElements, Size, compar_oligo_count_pair_decreasing);
}

void test_sort_oligo_count_pairs()
{
	int j;
	oligo_count_pair *array[3];
	MALLOC(array[0], sizeof(oligo_count_pair));
	MALLOC(array[1], sizeof(oligo_count_pair));
	MALLOC(array[2], sizeof(oligo_count_pair));

	array[0]->count = 12; array[0]->oligo = "accd";
	array[1]->count = 2; array[1]->oligo = "aDSd";
	array[2]->count = 33; array[2]->oligo = "EREccd";

	for(j=0; j < 3; j++)
		printf("%s\t%i\n", array[j]->oligo, array[j]->count);
	printf("\n\n");
	sort_order so = decreasing;
	sort_oligo_count_pairs(array, 3, so);
	for(j=0; j < 3; j++)
		printf("%s\t%i\n", array[j]->oligo, array[j]->count);
	getchar();
}

*/

/* sort gene_stat_pairs by seq_level_stat*/

/* sorting 'd_gs' by value of 'score' */
/*
int compar_g_s_decreasing(const void *s1 ,const void *s2)
{
	if((*((gene_stat_pair **)s1))->seq_level_stat > ((*(gene_stat_pair **)s2))->seq_level_stat) return -1;
	else if((*((gene_stat_pair **)s1))->seq_level_stat == ((*(gene_stat_pair **)s2))->seq_level_stat) return 0;
	else return 1;
}

int compar_g_s_increasing(const void *s1 ,const void *s2)
{
	if((*((gene_stat_pair **)s1))->seq_level_stat > ((*(gene_stat_pair **)s2))->seq_level_stat) return 1;
	else if((*((gene_stat_pair **)s1))->seq_level_stat == ((*(gene_stat_pair **)s2))->seq_level_stat) return 0;
	else return -1;
}

void sort_g_s(gene_stat_pair **array, int NumberOfElements, sort_order so)
{
	size_t Size = sizeof(d_gs *);
	if(so == increasing)
		qsort(array, NumberOfElements, Size, compar_g_s_increasing);
	else
		qsort(array, NumberOfElements, Size, compar_g_s_decreasing);
}
*/
