/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef TREE_H_
#define TREE_H_

#include <stdbool.h>
#include <search.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#define oops(s) { perror((s)); exit(EXIT_FAILURE); }
#define MALLOC(s,t) if(((s) = malloc(t)) == 0) { oops("error: malloc() ");}
#define FREE(p) {free((void *)p);p=0;}

// needed for hash table implementation
// int hcreate(size_t nel);
// void hdestroy(void);
// ENTRY *hsearch(ENTRY item, ACTION action);
// also binary tree search
#define _GNU_SOURCE
#include <search.h>


//void action_initialize_reference(const void* nodep, VISIT value, int level);
//void action_find_overlap_to_reference(const void* nodep, VISIT value, int level);
//void action_fn(const void* nodep, VISIT value, int level);
//void tree_walk(const void *root, __action_fn_t action);

void tree_walk_in_sorted_order(const void *root);

void tree_walk_initialize_reference(const void *root, int n, int sequence_length);
void tree_walk_find_overlap_to_reference(const void *root, int lower_offset_, int upper_offset, void **index, int length_shorter_motif);
void tree_walk_count_gene_occurrences(const void *root, void **index, int sequence_length);
void tree_walk_find_gene_occurrences(const void *root, void **index, int sequence_length, int num_genes);

void deallocate_reference_positions();

int get_length_reference_positions();
int *get_reference_positions();
int get_total_num_overlaps();
int get_num_offset_specific_gene_overlaps();

int get_gene_count();
int *get_genes();
int get_length_genes();

int compar_int(const void *s1 ,const void *s2);
void destroy_tree(void **root);
void freefct();

#endif /*TREE_H_*/
