/* 
 *    SSMART: Sequence-structure motif identification for RNA-binding proteins *
 *    Version 1.2, 20 Jul 2017                                                 *
 *    Copyright (C) 2017  Alina Munteanu, Stoyan Georgiev, Uwe Ohler           *
 *                                                                             *
 *    SSMART is free software: you can redistribute it and/or modify           *
 *    it under the terms of the GNU General Public License as published by     *
 *    the Free Software Foundation, either version 3 of the License, or        *
 *    (at your option) any later version.                                      *
 *                                                                             *
 *    This program is distributed in the hope that it will be useful,          *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *    GNU General Public License for more details.                             *
 *                                                                             *
 *    You should have received a copy of the GNU General Public License        *
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
 *                                                                             *
 *    E-mail: alina-cristina.munteanu@mdc-berlin.de, uwe.ohler@mdc-berlin.de   *
 */

#ifndef DEGEN_GENE_SET_H_
#define DEGEN_GENE_SET_H_

#define oops(s) { perror((s)); exit(EXIT_FAILURE); }
#define MALLOC(s,t) if(((s) = malloc(t)) == 0) { oops("error: malloc() ");}
#define FREE(p) {free((void *)p);p=0;}
#define ZERO	1.0e-20
#define MAX(x,y)        ((x) > (y) ? (x) : (y))
#define MIN(x,y)        ((x) < (y) ? (x) : (y))
#define ABS(x)   	    ((x) >  0  ? (x) : (-x))

#include "IUPACalphabet.h"
#include "motif_evolution_interface.h"
#include "tree.h"
#include "parameters.h"
#include "similarity_metrics.h"
#include "geneset_scoring_interface.h"
#include <gsl/gsl_matrix.h>
#include <string.h>
#include <gsl/gsl_sf_pow_int.h>
#include <gsl/gsl_sf_exp.h>


////////////////////////////////
// used in 'motif_occurrences.h' 
////////////////////////////////



typedef struct
{
	unsigned long index_start_offset;
	unsigned long index_last_offset;
	unsigned long num_occ_in_region;
}region_match_info;

typedef struct
{
	unsigned long *locations;
	unsigned long num_locations
}exact_oligo_occ;


typedef struct
{	
	// exact instantiations of the degenerate oligo
	exact_oligo_occ *exact_occ; int num_exact;

	// the degen oligo match information for each sequence region
	region_match_info *match_info; int num_region_matches;
//	int *region_indexes, *within_region_offsets, *num_occ_per_region;
}degen_oligo_occurrences;


////////////////////////////////
// used in 'motif_occurrences.h'
////////////////////////////////

/*
typedef struct
{
	int **locations_storage, num_instantiations;
	void *locations_tree;
}motif_instances;

//motif_instances *create_motif_instances(int num_oligos);
//void destroy_motif_instances(motif_instances **root);
*/

typedef struct
{
	char *oligo;
	int count;
}oligo_count_pair;

/*
typedef struct
{
	int *genes; int num_genes;
	char **oligos; int num_oligos;
	int *oligo_counts;
}gene_set_profile;


typedef struct
{
	// entries are in the following order: A, G, C, T
	double (*pos)[4][4];
	int length;
}motif_MM1;
*/

typedef struct
{
	int *absolute_locations;
	int *region_indexes, *within_region_offsets, *num_occ_per_region;
	int *strand_specificity; // 1 -> sense (even gene index), 0 -> anti-sense (odd gene index)
	int num_occurrences;
}motif_occurrences;


//typedef void(*pt2select_next_motif)(double *scores, int num, int *selected_ind);
//void select_next_motif_sample(double *scores, int num, int *selected_ind);
//void select_next_motif_top_scoring(double *scores, int num, int *selected_ind);

typedef struct
{
	char *degen_oligo; int oligo_length;
	char **instantiations; unsigned instantiations_length;
	int *instantiations_counts;
	gsl_matrix *pssm;
	int *genes; int num_genes;
	double *gene_level_stat; int total_num_genes;
	double score, score_5percentile, score_95percentile, score_5percentile_permuted, score_95percentile_permuted;

	double p_value;
//	double empirical_p_value;
	bool conservation_used; 
	bool is_valid;
	int num_evolution_steps;
}d_gs;


typedef d_gs *(*create_geneset_fptr)(char *degen_oligo, params *p, int num_evolution_steps);
d_gs *create_geneset_applying_orth_filter(char *degen_oligo, params *p, int num_evolution_steps);
void create_empty_d_gs(params *p, d_gs **dg);
d_gs *create_d_gs(const char *oligo, params *p, int num_evolution_steps);

					    
typedef d_gs *(*evolve_motif_report_top_fptr)          (char *degen_oligo, params *p, create_geneset_fptr create_d_gs_fptr, pt2select_next_motif select_next, int num_evolution_steps);
d_gs *evolve_motif_report_top_bidirectional            (char *degen_oligo, params *p, create_geneset_fptr create_d_gs_fptr, pt2select_next_motif select_next, int num_evolution_steps);
d_gs *evolve_motif_report_top_bidirectional_orth_filter(char *degen_oligo, params *p, create_geneset_fptr create_d_gs_fptr, pt2select_next_motif select_next, int num_evolution_steps);

bool is_valid_score(double score);

int *find_degen_oligo_set_occurrences_fixed_length(void **index, unsigned char **oligos, int oligo_length, int num_oligos,
 									 int seq_length, int num_sequences, int *num_occurrences);
//int *find_degen_oligo_set_occurrences_var_length(void **index, unsigned char **oligos, int oligo_length, int num_oligos,
//												 int num_sequences, int *num_occurrences, int *locations);
//int find_degen_oligo_set_occurrence_locations_fixed_length
//					(void **index, unsigned char **oligos, int oligo_length, int num_oligos,
//					 int seq_length, int num_sequences, motif_instances **motif_instances_tree);

void initalize_candidate_set(d_gs *parent, params *p);


//gsl_matrix *build_pssm_from_counts_fixed_length(void **index, int seq_length, int num_sequences, d_gs *dg);
//void build_pssm_from_counts_fixed_length(void **index, int seq_length, int num_sequences, d_gs *dg);

//gsl_matrix *build_pssm_from_counts_var_length(void **index, int num_sequences, int *locations, d_gs *dg);

//void build_pssm_from_top_sequences_only(void **index, int seq_length, int num_sequences, d_gs *dg, int how_many_top_sequences);

void free_d_gs(d_gs *dg);
void free_d_gs_batch(d_gs **dg_arr, int num);
void sort_d_gs_genes(d_gs *dg);

// Reverse complement
char *rev_compl(char *oligo);

// Compare motifs
double compare_motifs(char *motif1, char *motif2);

//d_gs *evolve_motif_report_top_bidirectional_orth_filter(char *degen_oligo, params *p, double T);


d_gs *evolve_motif(d_gs *dg, params *p);


// Evolve motifs
d_gs *evolve_motif_greedy(d_gs *dg, params *p);//d_gs *dg, void **index, int total_num_genes, int seq_length, score_type st, params *p);
// Evolve motif with ortholog filter
d_gs *evolve_motif_greedy_with_conservation(d_gs *dg, params *p);

void init_background_freq(species species_type);


char *substr(char *str, int start, int num_chars);

// Print functions
void print_d_gs_oligo(const d_gs *dg);
void print_d_gs_genes(const d_gs *dg);
void print_d_gs_gene_set(const d_gs *dg);
void print_seq_level_stat(d_gs *dg);

void test_d_gs();
//void test_create_d_gs();

#endif /*DEGEN_GENE_SET_H_*/
